/* #define DEBUG 1 */

/*==================================================================
                  +---------------+
 			|  functions.c  |
                  +---------------+

 The set of utility routines declared in hlsfun.h . Includes also
 the rpc routines do_rpc and do_rpc_err. The input_... routines
 and the update_twiss_function are elsewhere.
 
 First test version 0.1. 		(Mark Plesko 11.12.92)

 Included read_septa_from_file() to read septa (not kickers) voltages
 from the machine file.
 Beta 0.2				(M.P. 3.2.93)

 Added existence of machine file checking and error reporting.
 field_to_strength, strength_to_field, f2s_magnets and f2s_correctors
 function type and argument structure changed.
 Beta 0.3				(M.P. 4.2.93)
 
 All the functions in hlsfun.h now implemented.
 Beta 1.0				(M.P. 15.2.93)
 
 Fully tested and debugged version (except ID routines):
 get_single_BPM corrected to read realArray type. set_correctors_slow
 modified to estimate proper time needed. Waiting is done with sleep
 now to reduce CPU load - except wait_BPM_average.
 Small errors fixed in set_current_fast, wait_BPM_average, get_single_FS,
 read_BPM_from_field, set_correctors_slow, do_rpc.
 Beta 1.1				(M.P. 19.2.93)
 
 Changed init_model_from_xxx such that no twiss_record is passed.
 New get_single_FS() now according to the specs by M. Ferianis.
 Modes "RE" and "WE" for RPC now supported.
 Added get_single_voltage to read the septum and kicker high voltages.
 Changed read_magnets_from_xxx and f2s_magnets to include also the
 septa, such that all the magnetic elements which contribute to the
 optics are handled in one routine. Deleted the now obsolete routine
 read_septa_from_file.
 Beta 1.2				(M.P. 23.2.93)
 
 Modified do_rpc to account for 1) new access_record and host_record
 structure, 2) the unavailability of new import calls once the
 routine cleanup_rpc was called, 3) the bind_lpc is called only if the
 sid is different to the previous one and 4) a new error number scheme:
 -1: rpc error; -2: import error; -3: rpc type nonexistent; -4: import
 after cleanup_rpc not allowed.
 2.0					(M.P. 15.4.93)
 
 Added write_machine_file() which mimickes the machine file, but
 overwrites the whole contents if a previous file exists. Added the
 routines read_kickers_from_xxx() for completeness.
 2.1					(M.P. 5.5.93)
 
 Modifications due to the new parameterlist of do_rpc in all routines
 which call do_rpc. The routines do_rpc and do_rpc_err have been 
 transferred to the file rpc.c . The final version of get_single_FS
 (now also separate as init_single_FS, acquire_single_FS, finish_single_FS)
 and the routines to run the scraper in automatic mode (init_scraper_auto,
 start_scraper_auto, finish_scraper_auto) are added. A new routine
 to set all magnets with WTIPOL (set_magnets_fast) and another to
 convert strength to field (s2f_magnets) is added.
 2.2					(M.P. 20.5.93)
 
 set_magnets_fast, set_correctors_fast and set_correctors_slow are
 modified such that they pass the estimated time in seconds and leave
 it to the user to do the sleep.
 2.3					(M.P. 22.5.93)
 
 write_magnets_to_file and write_correctors_to_file added for separate
 saving of elements to the machine file. read_TL_emittance added, which
 reads twiss parameters and also the emittance (needed for beam profile
 displays).
 2.4					(M.P. 1.6.93)
 
 A whole set of new and modified routines, created during the 
 commissioning of the first part of the transfer line.
 New: get_single_BPM_TL, get_single_BPM_SR, get_single_CM_TL, 
      unlock_FS,
      quit_scraper_auto, wait_scraper_auto, get_scraper_CM,  
      switch_PS_on, switch_PS_off,
      read_TL_emittance.
 Mod.:init_single_FS, acquire_single_FS, finish_single_FS, get_single_FS,
      init_scraper_auto, start_scraper_auto, finish_scraper_auto,
      get_single_current,
      read_BPM_from_field, 
      write_magnets_to_file, write_correctors_to_file, write_machine_file.
 The function set_correctors_slow is not supported anymore !
 For more explanations see hlsfun.h .
 3.0					(M.P. 23.6.93)
 
 functions which relate to either TL or SR fused together, to avoid
 modifications necessary for the programmer - instead it is determined
 from the first character of the pAcc->member string, whether the
 element is from the TL ('T') or SR ('S').
 get_single_BPM_SR is still kept, because its parameter list differs
 from get_single_BPM - it returns the while BPM_array.
 A small bug fixed in field_to_strength and strength_to_field: now
 the calibration data can contain all five coefficients.
 get_BPM_SR_from_field now calls the new bpm service.
 set_current_fast now correctly checks whether the magnets are on and
 remote, for any type of magnet, TL or SR. It uses the new routine
 get_PS_status.
 3.1					(M.P. 12.8.93)
 
 A change in field_to_strength and strength_to_field allow now for
 all calibration coefficients to be zero (for the case of bends
 whith index 0). s2f_magnets calculates bend currents from the energy
 and does not use the inverse of the index anymore.
 3.2					(M.P. 8.9.93)
 
 Added handling of big SR power supplies in switch_on and set_current_fast.
 New routines: big_SR_PS, ramping, which are public and set_fast, set_bbb, 
 which are static. 
 The default machine file directory is set to be HLSPREF (if filename
 starts with '.'). The routine write_correctors_to_file writes also
 kicker voltages.
 Optimized wait_BPM_average: if time to wait is more than a second, it
 uses sleep first.
 Handling of the anular electrode is added with init_anular, anular_range,
 get_anular, finish_anular.
 3.3					(M.P. 30.9.93)

 On request of CJB the routine get_single_BPM_current is added, which
 reads the beam current from the BPMs. For now, this works only with
 TL BPMs. A SR BPM as parameter forces a -2 return value.
 ramping() modified such that it accepts also TL magnets, although in
 that case it checks for ramping by issueing another ramp, which is
 not very elegant. 
 In set_fast the waiting is optimized: if t<1s, no sleep is done and
 the exact time is waited only.
 In switch_on the DAC is always first set to I_start, independent of the
 PS type. If the PS is already on, nothing is done.
 An inconsistent handling of HLSPREF and HLSMACHINE in read_energy_from_file
 has been fixed. Now, a file beginning with "." is looked for in
 HLSPREF, instead of HLSMACHINE.
 3.4					(M.P. 3.10.93)
 
 Added one additional RPC call in init_anular and finish_anular which
 sets and resets the position function filter to read the anular 
 electrode.
 Added DCCT handling: init_DCCT, get_DCCT
 3.5					(M.P. 7.10.93)
 
 Changed get_DCCT so that it reads the DCCT mode. Therefore the
 global variable which remembers the mode in init_DCCT has been
 removed.
 Removed "set single turn mode" in anular, added one more function
 filter call.
 Added parameter "gate" to get_status. Correct handling of gate on/off
 in switch_on.
 get_single_voltage reads the set values instead or read values, since
 those are not well defined for injection elements.
 The kickers are added to the f2s_correctors and s2f_correctors.
 The error message offset for the SR BPMs from the bpm_service
 server are taken into account (20480).
 3.6					(M.P. 12.10.93)
 
 Removed all references to clock(), as also this routine works only
 in units of seconds and not fractions of seconds.
 Modified set_correctors_fast such that it does not require the
 condition of the correctors to be APPLY. However, this is still
 used for apply_correction, so this routine now calls set_correctorsRN,
 which still contains the request for condition==APPLY.
 Added to read_BPM_SR_from_field the handling of the error code 153,
 which sets the condition to bad instead of dead.
 3.7					(M.P. 14.10.93)
 
 Corrected BPM_NOT_READY to 154 instead of 153. Corrected a spelling
 error (GATEON replaced by GATON).
 3.8					(M.P. 28.10.93)
 
 Modified handling of reading and setting of PS: the machine file
 now supports both read and set values, so does the databush. The
 following routines have been modified to accomodate this:
 read_machine_file_current, read_current_from_file, get_single_current, 
 get_single_voltage, write_magnets_to_file, write_correctors_to_file,
 read_energy_from_field, read_energy_from_file, read_magnets_from_field,
 read_magnets_from_file, read_IDgaps_from_field, read_IDgaps_from_file,
 read_correctors_from_field, read_correctors_from_file, 
 read_kickers_from_field, read_kickers_from_file.
 
 In addition, the parameter list of the following routines have 
 changed: read_machine_file_current, read_current_from_file,
 	  get_single_current, get_single_voltage,
 where one more parameter for the setting has been added.
 4.0					(M.P. 10.11.93)
 
 Added error handling in read_magnets_from_field and init_model_from_field,
 such that a warning is given if for one or more magnets the reading
 differs from the setting (as reported in get_single_current with the
 return value of -3). The warning is a return value of -3 for
 read_magnets_from_field and a return value of -5 for the function
 init_model_from_field.
 4.1					(M.P. 11.11.93)
 
 Added routine set_DAC, which ramps using WTIPOL without slow start and
 stop. It is used in set_PS_on before the PS is switched on to set the
 DAC to the proper value.
 The routine set_PS_on has been modified: it sends a RESET command
 before switching on. It now checks if the PS is on after the operation.
 4.2					(M.P. 25.11.93) 
 
 Because it takes some time for the PS switch to turn, the routine
 switch_on was modified such that it waits 1 second after switch on
 before it checks the status of the PS.
 4.3					(M.P. 21.1.94)
 
 Added get_single_BPM_LASPOS, which will be used by RN for some
 tests. This feature will be deleted in future versions !!!
 4.3.1					(M.P. 1.2.94)
 
 Added get_tune(), get_RF_frequency(), set_RF_frequency(). Routines
 will be used for dispersion and chromaticity measurements.
 4.4.1					(M.P. 3.2.94)
 
 Added full treatment of ID's: get_single_IDgap(), set_IDgaps(),
 get_ID_status(), get_ID_speed(). The dummy IDgap_to_strength() has
 been replaced by a real one written by Lidia Tosi. The routine 
 read_IDgaps_from_field() is modified to take real ID calls instead 
 of dummy ones.
 Test version: set_IDgaps() not yet functional. The rest should work.
 4.5.1-t				(M.P. 11.2.94)
 
 Modified IDgap_to_strength() to take the empirical fit.
 Still test version.
 4.6.1-t				(M.P. 25.2.94)
 
 Added the routines for the TL feedback needed by R.N.: start_feedback(),
 stop_feedback() and read_feedback(). get_single_BPM_LASPOS() has not
 been deleted yet.
 4.6.2-t				(M.P. 25.2.94)
 
 Modified switch_off() such that also for the big PS the mechanical
 switch is switched off (in addition to the gate), with the exception
 of the SR bending PS.
 4.6.3-t				(M.P. 2.3.94)
 
 Fixed bug in read_IDgaps_from_field(). Deleted get_single_BPM_LASPOS().
 Fixed bug in IDgap_to_strength(). No reference to "CARID" in the code, as
 it has been placed in the access file.
 4.7					(M.P. 10.3.94)
 
 Written routine set_IDgaps. No check is made if setting of gap is
 reasonable or possible.
 4.8					(M.P. 14.4.94)
 
 Added lifetime routine (will be placed in server later) calc_lifetime().
 Modified ramping() such that it uses the BUSY action also for TL magnets.
 4.9					(M.P. 19.4.94)
 
 Modified lifetime routine: 1) does not use the fit criteria anymore,
 only the rms to determine a change in lifetime 2) if the rate is below
 the sensitivity of the DCCT, it is set to 0, tau becomes 100 h and the
 error is 100%.
 4.10					(M.P. 6.6.94)
 
 Added get_tunefeedback_DAC() which reads the DACs of the tunefeedback
 for QF and QD. Modified get_single_current() such that for QF and QD
 the tunefeedback DAC is also taken into account.
 4.11					(M.P. 9.6.94)
 
 The ID gap is read several times, until no error occurs or the maximal
 number of attempts has been performed in get_single_IDgap().
 4.12					(M.P. 10.10.94)
 
 Added the writing and reading of the tune and the BPMs to the machine
 file. For this the routine read_machine_file_BPM() was added, the
 routine write_machine_file() was modified and the previously dummy routine
 read_BPM_SR_from_file was finally made correct.
 4.13					(M.P. 21.11.94)
 
 Added the reading of the SR profile monitor status and data in the
 routines get_SRPM_status(), get_SRPM_data(). Modified get_DCCT() such
 that it now returns the correctly calculated lifetime. init_DCCT() is
 now obsolete and prints only an error message. The routine calc_lifetime()
 is still kept for special purposes.
 4.14					(M.P. 20.12.94)
 
 Added several attempts to read injection element voltages, just like
 in the case of reading ID gaps, because it is also a serial link.
 4.15					(M.P. 22.12.94)
 
 Replaced the (tested) routines for setting the IDs: get_single_IDgap(),
 get_ID_status(), get_ID_speed(), set_ID_brake(), check_ID_loop() and
 set_IDgaps(). And fixed a stupid typing bug in get_single_voltage().
 4.16					(M.P. 16.1.95)
 
 Added set_ID_loop() to control the wigglers, added the routines for the
 photon shutter: get_PHBSH_status() and close_PHBSH() from Lidia.
 4.17					(M.P. 6.2.95)
 
 Some small modifications in init_FS: background subtraction is now
 enabled with 1; background aquisition has been added.
 4.18					(M.P. 20.2.95)
 
 Added the handling of the SR scraper: status_SR_scraper(),
 read_SR_scraper_hack() and move_SR_scraper_hack(). The last two
 routines are a hack to handle easily the vertical scraper. Later,
 this MUST be incorporated into the hls databush (type=hor/ver) !
 4.19					(M.P. 3.4.95)
 
 Added  read_SR_scraper (replaces read_SR_scraper_hack) and
 move_SR_scraper (replaces move_SR_scraper_hack) from release 5 of
 functions.c for immediate use, as they do not need the new databush.
 4.20					(M.P. 19.7.95)
 
 Added new tune measuring routines from release 5 of functions.c for
 immediate use (on request of LT).
 get_single_tune_measurement()
 start_multiple_tune_measurement()
 stop_multiple_tune_measurement().
 4.21					(M.P. 25.9.95)
 
 On request of EK, added set_voltage(), which allows to set septum and
 kicker voltages. This handled by service_inj together with delays,
 which is not always advantageous.
 4.22					(M.P. 25.9.95)
 
 Modified get_DCCT, such that it now gets the data from the beam info
 server (on tom) and not from the DCCT.
 4.23					(M.P. 21.11.95)
 
 Incorporated the modifications on the tune_measurement routines by FI
 (UNLOCKs the FS and more error notification).
 4.24					(M.P. 11.12.95)
 
 Added routines f2l_bends() and field_to_length() which obtain correct
 magnetic length for the bends.
 4.25					(L.T. 23/12/97)
 
 
 Added routine to read gap, taper & phase of sazaky type IDs:
 get_sazaki_IDgataph().
 Modified read_IDgaps_from_field() to use the new call.
 Function get_single_IDphase() is probably not working any more
 due to member "PHAM" no longer supported.
                              (C. SCafuri 20.06.2002)
  
 ==================================================================*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include <time.h>

#include "hlsvar.h"
#include "hlsrpc.h"
#include "hlsfun.h"

/* ask Rudolf about this for various correctors !!! */
#define AMPS_PER_SEC	0.1

/* time needed (in milliseconds) to go bit by bit from 0 to I_high for big SR PS */
#define FULL_TIME	10*65536

/* ask Ferianis about this for FS !!! */
#define BPM_TIMEOUT	115
#define CM_TIMEOUT	115
#define FS_TIMEOUT	115
#define BPM_SLEEP	2
#define CM_SLEEP	2
#define BPM_TRIALS	5
#define CM_TRIALS	5
#define FS_TRIALS	5
#define TUNE_SCAN_SLEEP 15

/* ask Raf about this for SR BPMs */
#define BPM_ERROR_OFFSET 20480
#define BPM_OK		0
#define BPM_SIGNAL_LOW	152
#define BPM_NOT_READY	154

/* maximal allowed change of RF f */
#define MAX_DELTA_FREQUENCY 0.001
#define MAX_TOTAL_DELTA_FREQUENCY 0.01

/* maximal number of attempts to read ID gap if an error occured */
#define	N_ID_ATTEMPT	5
/* max. number of attempts to read injection el. voltage if an error occured */
#define	N_INJ_ATTEMPT	5

#define N80	80

#define ORACLEenergy "ENERGY"


/**************************************************************************/
/* The following functions include ORACLE I/O statements. They may have to 
   be modified if a different database system is used. However, they do not
   include any reference to the HLS databush, i.e. no global variables or
   record structures.							  */
/**************************************************************************/


int read_machine_file_current (fileName, fam_mem, readback, current)

str32	fileName ;
str13	fam_mem ;
double	*readback ;
double	*current ;

{
  str13	name ;
  char 	line[N80] ;
  char	*chPtr ;
  FILE	*machine ;
  char	fullName[N80] ;
  double setvalue ;
  double curr ;
       
  /* if fileName starts with '.', take HLSPREF directory as default */
  if ( *fileName == '.' ) strcpy (fullName, getenv ( "HLSPREF" ) ) ;
  /* otherwise take $HLSMACHINE as default directory */
  else strcpy ( fullName, getenv ( "HLSMACHINE" ) ) ;
  
  strcat ( fullName, fileName ) ;
  if ( (machine = fopen (fullName, "r")) == NULL ) {
    printf ("machine file not found: %s\n", fullName) ;
    return (-1) ;
  }
  while ( (chPtr = fgets(line, N80, machine)) != NULL )
  {
     /* init curr and setvalue for machine files without setvalue stored */
     curr = 0.0 ;
     setvalue = 0.0 ;
     sscanf ( line, "%s%lf%lf", name, &curr, &setvalue ) ;
     if ( ! strcmp(fam_mem,name) ) 
     {
       *current = setvalue ;
       *readback = curr ;
       fclose (machine) ;
       return ( 0 ) ;
     }
  } 
  fclose (machine) ;

  /* end of non ORACLE related I/O */
  
  printf ("family not found in machine file: %s\n", fam_mem) ;
  return (-2) ;

}	/* read_machine_file_current */


static int read_machine_file_BPM (fileName, pAcc, posH, posV, rmsH, rmsV)

str32		fileName ;
access_record	*pAcc ;	
double		*posH ;
double		*posV ;
double		*rmsH ;
double		*rmsV ;

{
  str13	name, fam_mem ;
  char 	line[N80] ;
  char	*chPtr ;
  FILE	*machine ;
  char	fullName[N80] ;
  double in1, in2, in3, in4 ;
       
  /* if fileName starts with '.', take HLSPREF directory as default */
  if ( *fileName == '.' ) strcpy (fullName, getenv ( "HLSPREF" ) ) ;
  /* otherwise take $HLSMACHINE as default directory */
  else strcpy ( fullName, getenv ( "HLSMACHINE" ) ) ;
  
  strcat ( fullName, fileName ) ;
  if ( (machine = fopen (fullName, "r")) == NULL ) {
    printf ("machine file not found: %s\n", fullName) ;
    return (-1) ;
  }
  
  /* convert access record info to family_member string */
  strcpy (fam_mem, pAcc->family) ;
  strcat (fam_mem, "_") ;
  strcat (fam_mem, pAcc->member) ;
   
  while ( (chPtr = fgets(line, N80, machine)) != NULL )
  {
     /* init in-value for machine files without values stored */
     in1 = 0.0 ;
     in2 = 0.0 ;
     in3 = 0.0 ;
     in4 = 0.0 ;
     sscanf ( line, "%s%lf%lf%lf%lf", name, &in1, &in2, &in3, &in4 ) ;
     if ( ! strcmp(fam_mem,name) ) 
     {
       *posH = in1 ;
       *posV = in2 ;
       *rmsH = in3 ;
       *rmsV = in4 ;
       fclose (machine) ;
       return ( 0 ) ;
     }
  } 
  fclose (machine) ;

  /* end of non ORACLE related I/O */
  
  printf ("family not found in machine file: %s\n", fam_mem) ;
  return (-2) ;

}	/* read_machine_file_current */


/**************************************************************************/
/* end of ORACLE specific data                                            */
/**************************************************************************/

/**************************************************************************/
/* The following function is TEMPORARY. It is used to replace the machine
   file during TL commissioning. Later the ORACLE machine file will be
   written to. Probably also in a different way !			  */
/**************************************************************************/

int write_machine_file (fileName)

str32	fileName ;

{
  FILE	*machine ;
  str13	fam_mem ;
  int	i, plane ;
  char	*chPtr ;
  char	fullName[N80] ;
       
  /* if HLSMACHINE not defined, take current directory as default */
  if ( NULL == ( chPtr = getenv ( "HLSMACHINE" ) ) ) *fullName='\0' ;
  /* if fileName starts with '.', take current directory as default */
  else if ( *fileName == '.' ) strcpy (fullName, getenv ( "HLSPREF" ) ) ;
  /* otherwise take $HLSMACHINE as default directory */
  else strcpy ( fullName, chPtr ) ;
  
  strcat ( fullName, fileName ) ;
  if ( (machine = fopen (fullName, "w")) == NULL ) {
    printf ("problems in opening machine file: %s\n", fileName) ;
    return (-1) ;
  }

  /* write first the energy */
  strcpy (fam_mem, ORACLEenergy) ;
  fprintf ( machine, "%-20s %10.4lf %10.4lf\n", fam_mem, energy.readback, energy.E ) ;

  /* now all magnet families */
  for (i=0;i<N_family;i++)
  {
    strcpy (fam_mem, fa[i]->pAcc->family) ;
    strcat (fam_mem, "_") ;
    strcat (fam_mem, fa[i]->pAcc->member) ;
    fprintf ( machine, "%-20s %10.4lf %10.4lf\n", fam_mem, fa[i]->readback, fa[i]->current ) ;
  }
    
  /* now all septa */
  for (i=0;i<N_septum;i++)
  {
    strcpy (fam_mem, sp[i]->pAcc->family) ;
    strcat (fam_mem, "_") ;
    strcat (fam_mem, sp[i]->pAcc->member) ;
    fprintf ( machine, "%-20s %10.4lf %10.4lf\n", fam_mem, sp[i]->readback, sp[i]->voltage ) ;
  }
    
  /* now all kickers */
  for (i=0;i<N_kicker;i++)
  {
    strcpy (fam_mem, ki[i]->pAcc->family) ;
    strcat (fam_mem, "_") ;
    strcat (fam_mem, ki[i]->pAcc->member) ;
    fprintf ( machine, "%-20s %10.4lf %10.4lf\n", fam_mem, ki[i]->readback, ki[i]->voltage ) ;
  }
    
  /* now all IDs */
  for (i=0;i<N_ID;i++)
  {
    if ( (id[i]->pPar->id.type == EEW) )
    {
      strcpy (fam_mem, id[i]->pAcc[0]->family) ;
      strcat (fam_mem, "_") ;
      strcat (fam_mem, id[i]->pAcc[0]->member) ;
      fprintf ( machine, "%-20s %10.4lf %10.4lf\n", fam_mem, id[i]->readback, id[i]->gap ) ;
      strcpy (fam_mem, id[i]->pAcc[1]->family) ;
      strcat (fam_mem, "_") ;
      strcat (fam_mem, id[i]->pAcc[1]->member) ;
      fprintf ( machine, "%-20s %10.4lf %10.4lf\n", fam_mem, id[i]->readback_h, id[i]->gap_h ) ;
    }
    else
    {
      strcpy (fam_mem, id[i]->pAcc[0]->family) ;
      strcat (fam_mem, "_") ;
      strcat (fam_mem, id[i]->pAcc[0]->member) ;
      fprintf ( machine, "%-20s %10.4lf %10.4lf\n", fam_mem, id[i]->phase, id[i]->gap ) ;
    }
  }
    
  /* now all correctors */
  for (i=0;i<N_corrector;i++)
  {
    plane = (co[i]->plane==lumped) ? horizontal : co[i]->plane ;
    do {
      strcpy (fam_mem, co[i]->pAcc[plane]->family) ;
      strcat (fam_mem, "_") ;
      strcat (fam_mem, co[i]->pAcc[plane]->member) ;
      fprintf ( machine, "%-20s %10.4lf %10.4lf\n", fam_mem, co[i]->readback[plane], co[i]->current[plane] ) ;
    } while ( ++plane == vertical && co[i]->plane == lumped ) ;

  }
    
  /* now all chicanes */
  for (i=0;i<N_chicane;i++)
  {
    plane = (chi[i]->plane==lumped) ? horizontal : chi[i]->plane ;
    do {
      strcpy (fam_mem, chi[i]->pAcc[plane]->family) ;
      strcat (fam_mem, "_") ;
      strcat (fam_mem, chi[i]->pAcc[plane]->member) ;
      fprintf ( machine, "%-20s %10.4lf %10.4lf\n", fam_mem, chi[i]->readback[plane], chi[i]->current[plane] ) ;
    } while ( ++plane == vertical && chi[i]->plane == lumped ) ;

  }
    
  /* now the tunes */
  fprintf ( machine, "%-20s %10.4lf %10.4lf\n", "TUNE", 
  			model_tune[horizontal], model_tune[vertical] ) ;
    
  /* now all BPMs */
  for (i=0;i<N_BPM;i++)
  {
    strcpy (fam_mem, po[i]->pAcc->family) ;
    strcat (fam_mem, "_") ;
    strcat (fam_mem, po[i]->pAcc->member) ;
    fprintf ( machine, "%-20s %10.4lf %10.4lf %10.4lf %10.4lf\n", fam_mem, 
      po[i]->beam_position[horizontal], po[i]->beam_position[vertical], 
      po[i]->rms[horizontal], po[i]->rms[vertical] ) ;
  }
    
  fclose (machine) ;

}	/* write_machine_file */


int write_magnets_to_file (fileName)

str32	fileName ;

{
  FILE	*machine ;
  str13	fam_mem ;
  int	i, plane ;
  char	*chPtr ;
  char	fullName[N80] ;
       
  /* if HLSMACHINE not defined, take current directory as default */
  if ( NULL == ( chPtr = getenv ( "HLSMACHINE" ) ) ) *fullName='\0' ;
  /* if fileName starts with '.', take current directory as default */
  else if ( *fileName == '.' ) strcpy (fullName, getenv ( "HLSPREF" ) ) ;
  /* otherwise take $HLSMACHINE as default directory */
  else strcpy ( fullName, chPtr ) ;
  
  strcat ( fullName, fileName ) ;
       
  if ( (machine = fopen (fullName, "w")) == NULL ) {
    printf ("problems in opening machine file: %s\n", fileName) ;
    return (-1) ;
  }

  /* write first the energy */
  strcpy (fam_mem, ORACLEenergy) ;
  fprintf ( machine, "%-20s %10.4lf %10.4lf\n", fam_mem, energy.readback, energy.E ) ;

  /* now all magnet families */
  for (i=0;i<N_family;i++)
  {
    strcpy (fam_mem, fa[i]->pAcc->family) ;
    strcat (fam_mem, "_") ;
    strcat (fam_mem, fa[i]->pAcc->member) ;
    fprintf ( machine, "%-20s %10.4lf %10.4lf\n", fam_mem, fa[i]->readback, fa[i]->current ) ;
  }
    
  /* now all septa */
  for (i=0;i<N_septum;i++)
  {
    strcpy (fam_mem, sp[i]->pAcc->family) ;
    strcat (fam_mem, "_") ;
    strcat (fam_mem, sp[i]->pAcc->member) ;
    fprintf ( machine, "%-20s %10.4lf %10.4lf\n", fam_mem, sp[i]->readback, sp[i]->voltage ) ;

  }
    
  fclose (machine) ;

}	/* write_magnets_to_file */


int write_correctors_to_file (fileName)

str32	fileName ;

{
  FILE	*machine ;
  str13	fam_mem ;
  int	i, plane ;
  char	*chPtr ;
  char	fullName[N80] ;
       
  /* if HLSMACHINE not defined, take current directory as default */
  if ( NULL == ( chPtr = getenv ( "HLSMACHINE" ) ) ) *fullName='\0' ;
  /* if fileName starts with '.', take current directory as default */
  else if ( *fileName == '.' ) strcpy (fullName, getenv ( "HLSPREF" ) ) ;
  /* otherwise take $HLSMACHINE as default directory */
  else strcpy ( fullName, chPtr ) ;
  
  strcat ( fullName, fileName ) ;
       
  if ( (machine = fopen (fullName, "w")) == NULL ) {
    printf ("problems in opening machine file: %s\n", fileName) ;
    return (-1) ;
  }

  /* write first the energy */
  strcpy (fam_mem, ORACLEenergy) ;
  fprintf ( machine, "%-20s %10.4lf %10.4lf\n", fam_mem, energy.readback, energy.E ) ;

    
  /* now all correctors */
  for (i=0;i<N_corrector;i++)
  {
    plane = (co[i]->plane==lumped) ? horizontal : co[i]->plane ;
    do {
      strcpy (fam_mem, co[i]->pAcc[plane]->family) ;
      strcat (fam_mem, "_") ;
      strcat (fam_mem, co[i]->pAcc[plane]->member) ;
      fprintf ( machine, "%-20s %10.4lf %10.4lf\n", fam_mem, co[i]->readback[plane], co[i]->current[plane] ) ;
    } while ( ++plane == vertical && co[i]->plane == lumped ) ;

  }
    
  /* now all chicanes */
  for (i=0;i<N_chicane;i++)
  {
    plane = (chi[i]->plane==lumped) ? horizontal : chi[i]->plane ;
    do {
      strcpy (fam_mem, chi[i]->pAcc[plane]->family) ;
      strcat (fam_mem, "_") ;
      strcat (fam_mem, chi[i]->pAcc[plane]->member) ;
      fprintf ( machine, "%-20s %10.4lf %10.4lf\n", fam_mem, chi[i]->readback[plane], chi[i]->current[plane] ) ;
    } while ( ++plane == vertical && chi[i]->plane == lumped ) ;

  }
    
  /* now all kickers */
  for (i=0;i<N_kicker;i++)
  {
    strcpy (fam_mem, ki[i]->pAcc->family) ;
    strcat (fam_mem, "_") ;
    strcat (fam_mem, ki[i]->pAcc->member) ;
    fprintf ( machine, "%-20s %10.4lf %10.4lf\n", fam_mem, ki[i]->readback, ki[i]->voltage ) ;
  }
    
  fclose (machine) ;

}	/* write_correctors_to_file */


/*========================================================================*/
/* single point routines                                                  */
/*========================================================================*/

int	wait_BPM_average ()

{
  unsigned int	res, time, secs, nanosecs ; /* C.S. 10/11/2005 unsigned */
  RPC_list	list ;
  struct timespec req;
  struct timespec remain; /*required by API but not used by this function*/
  
  /* set input parameter list to do_rpc */
  list.bpm.time = &time ;
  
  /* data in milliseconds is assumed */
  if ( 0 > (res=do_rpc(pAccBPM_SR->pHost, service_bpm, "time", &list)) ) 
  							return(res);

  /* wait until system time passed is more than averaging time */
  secs=time/1000; /*integral part in seconds*/
  nanosecs=(time-(secs*1000))*1000000; /*fractional part in nsec*/
  req.tv_sec=secs;
  req.tv_nsec=nanosecs;
  printf ( "#time to wait is %u milliseconds (%u %u)\n", time,secs,nanosecs);
  nanosleep(&req,&remain); /* C.S. 12/01/2006  - nanosleep is standard */

  return ( 0 ) ;
}


/* this routine is kept for backward compatibility - it will be deleted later */

int	get_single_BPM_TL ( ptr, posh, posv )

monitor_record	*ptr ;
double		*posh, *posv ;

{
  printf ( "#please use \"get_single_BPM\" instead of \"get_single_BPM_TL\"\n");
  return ( get_single_BPM ( ptr, posh, posv ) ) ;
}



int	get_single_BPM_current ( ptr, beam_current )

monitor_record	*ptr ;
double		*beam_current ;

{
  RPC_list	list ;
  double	data ;
  int		i,res, trials ;
  double	currold ;


  /* set input parameter list to do_rpc */
  list.lpc.family = ptr->pAcc->family ;
  list.lpc.member = ptr->pAcc->member ;
  
  /* different procedure for TL BPM and SR BPM ! */
  if ( ptr->pAcc->member[0] == 'T' )
  
  { /* this part for TL BPMs */
  
    list.lpc.action = "LASCUR" ;
    list.lpc.datum = &currold ;

    if ( 0 > do_rpc ( ptr->pAcc->pHost, service_lpc, "RR", &list ) ) return (-1) ;

  #ifdef DEBUG
  printf ( "LASCUR 1: current=%lf\n", xold, yold ) ;
  #endif

    list.lpc.action = "CUR" ;
    list.lpc.datum = &data ;
    res = do_rpc ( ptr->pAcc->pHost, service_lpc, "RR", &list ) ;

    if ( (res < 0) && (errno == BPM_TIMEOUT) )
    {
      list.lpc.action = "LASCUR" ;
      trials = 0 ;

      /* set data such that the following loop is entered */
      data = currold ; 
      res = 0 ;
      while( (res>=0) && (currold==data) )
      {
        if ( ++trials > BPM_TRIALS ) 
        {
          printf("#max No. of sleeps exceeded for %s_%s_CUR\n",
                                  list.lpc.family,list.lpc.member);
          res = -2 ;
          break ;
        }
        sleep ( BPM_SLEEP ) ;
        res = do_rpc ( ptr->pAcc->pHost, service_lpc, "RR", &list ) ;
  #ifdef DEBUG
  printf ( "LASCUR %i: current=%lf\n", trials, data ) ;
  #endif
      }
    }

    /* OK, current read was successful */
    *beam_current = data ;

    return(res) ;
    
  } /* end of TL BPM related part */
  
  else
  
  { /* this part is for the SR BPMs */
    printf ( "#get_single_BPM_current not supported for SR BPMs %s_%s\n",
             ptr->pAcc->family, ptr->pAcc->member ) ;
    return ( -1 ) ;
  } /* end of SR BPM related part */
  
} /* get_single_BPM_current */



int	get_single_BPM ( ptr, posh, posv )

monitor_record	*ptr ;
double		*posh, *posv ;

{
  RPC_list	list ;
  realArray	data, block ;
  int		i,res, trials ;
  double	xold, yold ;


  /* set input parameter list to do_rpc */
  list.lpc.family = ptr->pAcc->family ;
  list.lpc.member = ptr->pAcc->member ;
  
  /* different procedure for TL BPM and SR BPM ! */
  if ( ptr->pAcc->member[0] == 'T' )
  
  { /* this part for TL BPMs */
  
    list.lpc.action = "LASPOS" ;
    list.lpc.datum = &data ;

    data.length = 0 ; /* initialize always ! */
    res = do_rpc ( ptr->pAcc->pHost, service_lpc, "RAR", &list ) ;

    if ( res >= 0 ) {
      xold = data.sequence[0] ;
      yold = data.sequence[1] ;
  #ifdef DEBUG
  printf ( "LASPOS 1: hor=%lf, ver=%lf\n", xold, yold ) ;
  #endif
    }
    /* free even if an RPC error occured, because the data may be already there */
    if ( data.length > 0 ) { free ( data.sequence ) ;  data.length = 0 ; }
    if ( res < 0 ) return ( -1 ) ;

    list.lpc.action = "POS" ;
    res = do_rpc ( ptr->pAcc->pHost, service_lpc, "RAR", &list ) ;

    if ( (res < 0) && (errno == BPM_TIMEOUT) )
    {
      list.lpc.action = "LASPOS" ;
      data.sequence[0] = xold;
      data.sequence[1] = yold ;

      trials = 0 ;
      res = 0 ;
      while( (res>=0) && (xold==data.sequence[0]) && (yold==data.sequence[1]) )
      {
        if ( ++trials > BPM_TRIALS ) 
        {
          printf("#max No. of sleeps exceeded for %s_%s\n",
                                  list.lpc.family,list.lpc.member);
          res = -2 ;
          break ;
        }
        if ( data.length > 0 ) { free ( data.sequence ) ;  data.length = 0 ; }
        sleep ( BPM_SLEEP ) ;
        res = do_rpc ( ptr->pAcc->pHost, service_lpc, "RAR", &list ) ;
  #ifdef DEBUG
  printf ( "LASPOS %i: hor=%lf, ver=%lf\n",trials,data.sequence[0],data.sequence[1] ) ;
  #endif
      }
    }

    *posh = data.sequence[0] ;
    *posv = data.sequence[1] ;
    if ( data.length > 0 ) free ( data.sequence ) ;

    return(res) ;
    
  } /* end of TL BPM related part */
  
  else
  
  { /* this part is for the SR BPMs */
  
    list.lpc.action = "POS" ;
    list.lpc.datum = &block ; ;

    block.length = 0 ;
    if ( 0 <= (res = do_rpc ( ptr->pAcc->pHost, service_lpc, "RAR", &list ) ) ) 
    {
      if ( block.length != AL_BPM_rpc ) /*length of array, see hlsdef.h*/
        res = -2 ;
      else 
      { 
        *posh = block.sequence[0] ; 
        *posv = block.sequence[2] ; 
        if ( block.sequence[4] < 0.0 ) res = -3 ; /* signal too low warning */
      }
    }
    if ( block.length > 0 ) free ( block.sequence ) ; /* VERY IMPORTANT */
    return ( res ) ;
    
  } /* end of SR BPM related part */
  
} /* get_single_BPM */


/* the following routine is used for test purposes only !!!! */

int	get_single_BPM_LASPOS ( ptr, posh, posv )

monitor_record	*ptr ;
double		*posh, *posv ;

{
  RPC_list	list ;
  realArray	data, block ;
  int		i,res, trials ;
  double	xold, yold ;


  /* set input parameter list to do_rpc */
  list.lpc.family = ptr->pAcc->family ;
  list.lpc.member = ptr->pAcc->member ;
  
  
  { /* this part for TL BPMs */
  
    list.lpc.action = "LASPOS" ;
    list.lpc.datum = &data ;

    data.length = 0 ; /* initialize always ! */
    res = do_rpc ( ptr->pAcc->pHost, service_lpc, "RAR", &list ) ;

    *posh = data.sequence[0] ;
    *posv = data.sequence[1] ;
    if ( data.length > 0 ) free ( data.sequence ) ;

    return(res) ;
    
  } /* end of TL BPM related part */
  
  
} /* get_single_BPM_LASPOS */


/* feedback BPM readings */

int	start_TL_feedback ()
{
  RPC_list	list ;
 unsigned char	data=1 ;

  /* set input parameter list to do_rpc */
  list.lpc.family = po[0]->pAcc->family ;
  list.lpc.member = po[0]->pAcc->member ;
  list.lpc.action = "FDBON" ;
  list.lpc.datum = &data ;

  return ( do_rpc ( po[0]->pAcc->pHost, service_lpc, "WB", &list ) ) ;
}


int	stop_TL_feedback ()
{
  RPC_list	list ;
  unsigned char	data=0 ;

  /* set input parameter list to do_rpc */
  list.lpc.family = po[0]->pAcc->family ;
  list.lpc.member = po[0]->pAcc->member ;
  list.lpc.action = "FDBON" ;
  list.lpc.datum = &data ;

  return ( do_rpc ( po[0]->pAcc->pHost, service_lpc, "WB", &list ) ) ;
}


int	read_TL_feedback ( values )
double	values[8] ;

{
  RPC_list	list ;
  int		res, i ;
  realArray	block ;

  /* set input parameter list to do_rpc */
  list.lpc.family = po[0]->pAcc->family ;
  list.lpc.member = po[0]->pAcc->member ;
  list.lpc.action = "FDB" ;
  list.lpc.datum = &block ;

  block.length = 0 ;
  if ( 0 <= (res = do_rpc ( po[0]->pAcc->pHost, service_lpc, "RAR", &list ) ) ) 
  {
    for (i=0;i<block.length;i++) values[i] = block.sequence[i] ;
  }

  if ( block.length > 0 ) free ( block.sequence ) ; /* VERY IMPORTANT */
  return ( res ) ;

} /* read_TL_feedback */


int	get_single_BPM_SR ( ptr, result )

monitor_record	*ptr ;
BPM_array	result ;

{
  RPC_list	list ;
  int		res, i ;
  realArray	block ;
  
  /* set input parameter list to do_rpc */
  list.lpc.family = ptr->pAcc->family ;
  list.lpc.member = ptr->pAcc->member ;
  list.lpc.action = "POS" ;
  list.lpc.datum = &block ;
  
  block.length = 0 ;
  if ( 0 <= (res = do_rpc ( ptr->pAcc->pHost, service_lpc, "RAR", &list ) ) ) 
  {
    for (i=0;i<block.length;i++) result[i] = block.sequence[i] ;
  }
  if ( block.length > 0 ) free ( block.sequence ) ; /* VERY IMPORTANT */
  return ( res ) ;
  
} /* get_single_BPM_SR */


/* this routine is kept for backward compatibility - it will be deleted later */

int	get_single_CM_TL ( ptr, value )

CM_record	*ptr ;
double		*value ;

{
  printf ( "#please use \"get_single_CM\" instead of \"get_single_CM_TL\"\n");
  return ( get_single_CM ( ptr, value ) ) ;
}


int	get_single_CM ( ptr, value )

CM_record	*ptr ;
double		*value ;

{
  RPC_list	list ;
  int		res, trials ;
  double	old ;
  unsigned char	data ;
  
  /* set input parameter list to do_rpc */
  list.lpc.family = ptr->pAcc->family ;
  list.lpc.member = ptr->pAcc->member ;

  /* different procedure for TL CM and SR CM ! */
  if ( ptr->pAcc->member[0] == 'T' )
  
  { /* this part for TL CMs */

    list.lpc.action = "CMGO" ;
    data = 1 ;
    list.lpc.datum = &data ; ;

    /* start low level process for CM */
    if ( 0 > do_rpc ( ptr->pAcc->pHost, service_lpc, "WB", &list ) ) return(-2);

    list.lpc.action = "LASCUR" ;
    list.lpc.datum = value ;

    res = do_rpc ( ptr->pAcc->pHost, service_lpc, "RR", &list ) ;

    if ( res >= 0 ) old = *value ;
    else goto quit ;

    list.lpc.action = "CUR" ;
    res = do_rpc ( ptr->pAcc->pHost, service_lpc, "RR", &list ) ;

    if ( (res < 0) && (errno == BPM_TIMEOUT) )
    {
      list.lpc.action = "LASCUR" ;
      *value = old ;

      trials = 0 ;
      res = 0 ;
      while( (res>=0) && (old == *value) )
      {
        if ( ++trials > CM_TRIALS ) 
        {
          printf("#max No. of sleeps exceeded for %s_%s\n",
                                  list.lpc.family,list.lpc.member);
          res = -2 ;
          break ;
        }
        sleep ( CM_SLEEP ) ;
        res = do_rpc ( ptr->pAcc->pHost, service_lpc, "RR", &list ) ;
      }
    }

  quit:
    list.lpc.action = "CMGO" ;
    data = 0 ;
    list.lpc.datum = &data ; ;

    /* stop low level process for CM */
    if ( 0 > do_rpc ( ptr->pAcc->pHost, service_lpc, "WB", &list ) ) return(-3);

    return ( res ) ;
    
  } /* end of TL CM related part */
  
  else
  
  { /* this part is for the SR CMs */
  
    printf ("SR CM handling not yet defined\n" ) ;
    return ( -1 ) ;
  }
  
}


/*****************************************************************
   this is the part for the anular electrode.
   The ranges given in anular_range correspond to the following
   currents:
              range		   current
             -------------------------------------
		0		3.00 - 12.00 mA
		1		1.50 -  6.00 mA
		2		0.75 -  3.00 mA
	 	3		0.38 -  1.50 mA
		4		0.20 -  0.80 mA
		5		0.10 -  0.40 mA

   These values, as well as the IF1, IF2 and calibration values
   were provided by R. Ursic 
*****************************************************************/

/* global variable for anular calibration purposes */

static int anular_range_SETTING = 6 ;


int	init_anular ( pAcc ) 

access_record	*pAcc ;

{
  RPC_list	list ;
  unsigned char	data ;
  str6		det_member ;
  char		*filter1 = "13,00,00,00" ;
  char		*filter2 = "0D,00,00,00" ;

  /* set input parameter list to do_rpc */
  list.lpc.family = "BPMD" ; /* detector !! */
  
  /* disable AGC first */
  list.lpc.member = pAcc->member ;
  list.lpc.datum = &data ;
  data = 0 ;
  list.lpc.action = "AGC" ;
  if ( 0 >  do_rpc(pAcc->pHost,service_lpc,"WB",&list) ) return ( -2 ) ; 

  /*now set the position function filter */
  list.lpc.action = "XDETM" ;
  list.lpc.datum = &filter1 ;
  if ( 0 >  do_rpc(pAcc->pHost,service_lpc,"WM",&list) ) return ( -3 ) ; 
  list.lpc.datum = &filter2 ;
  if ( 0 >  do_rpc(pAcc->pHost,service_lpc,"WM",&list) ) return ( -3 ) ; 

} /* init_anular */


int	anular_range ( pAcc, range ) 

access_record	*pAcc ;
int 		range ;

{
  double IF1[6] = {   0.0,  12.0,  55.0,  86.0, 108.0, 131.0 } ;
  double IF2[6] = {  2235,  3840,  3840,  3840,  3840,  3840 } ;

  RPC_list	list ;
  realArray	block ;
  double	data[3] ;

  /* return if wrong range specified */
  if ( (range < 0) || (range > 5) ) 
  {
    printf("#range: %i not existent for anular: %s_%s\n", 
                                 pAcc->family,pAcc->member);
    return ( -2 ) ;
  }

  /* set global variable which tells the range selected */
  anular_range_SETTING = range ;

  /* set input parameter list to do_rpc */
  list.lpc.family = "BPMD" ; /* detector !! */
  list.lpc.member = pAcc->member ;
  list.lpc.action = "MGC" ;
  list.lpc.datum = &block ; ;

  /* set parameters to be passed to lpc */
  data[0] = 1.0 ;
  data[1] = IF1[range] ;
  data[2] = IF2[range] ;
  block.length = 3 ;
  block.sequence = data ;
  return ( do_rpc ( pAcc->pHost, service_lpc, "WAR", &list ) )  ;
  
} /* anular_range */


int get_anular ( pAcc, current, current_rms ) 

access_record	*pAcc ;
double		*current, *current_rms ;

{
  double calib[6] = {  2.60E-2, 1.43E-2, 6.57E-3, 3.26E-3, 1.76E-3, 8.21E-4 } ;

  RPC_list	list ;
  int		res ;
  realArray	block ;

  if ( (anular_range_SETTING < 0) || (anular_range_SETTING > 5) ) 
  {
    printf("#range has not been set for anular: %s_%s\n", 
                                 pAcc->family,pAcc->member);
    return ( -2 ) ;
  }

  /* the following is copied from get_single_BPM_SR, for this is
     the way the anular electrode is read for now ! */
  
  /* set input parameter list to do_rpc */
  list.lpc.family = pAcc->family ;
  list.lpc.member = pAcc->member ;
  list.lpc.action = "POS" ;
  list.lpc.datum = &block ;
  
  block.length = 0 ;
  if ( 0 <= (res = do_rpc ( pAcc->pHost, service_lpc, "RAR", &list ) ) ) 
  {
    if ( block.length < 8 ) 
    {
      printf("#get_anular: returned array uncomplete for %s_%s\n", 
                                 pAcc->family,pAcc->member);
      res = -3 ;
    } 
    else 
    {
      *current = block.sequence[6] * calib[anular_range_SETTING] ;
      *current_rms = block.sequence[7] * calib[anular_range_SETTING] ;
    }
  }
  if ( block.length > 0 ) free ( block.sequence ) ; /* VERY IMPORTANT */
  return ( res ) ;

} /* get_anular */


int finish_anular ( pAcc ) 

access_record	*pAcc ;

{
  RPC_list	list ;
  unsigned char	data ;
  char		*filter = "13,02,00,00" ;

  /* set input parameter list to do_rpc */
  list.lpc.family = "BPMD" ; /* detector !! */
  
  /* enable AGC */
  list.lpc.member = pAcc->member ;
  list.lpc.datum = &data ;
  data = 1 ;
  list.lpc.action = "AGC" ;
  if ( 0 > do_rpc(po[0]->pAcc->pHost,service_lpc,"WB",&list) ) return ( -1 ) ;

  /* set pos. filter back to normal operation */
  list.lpc.datum = &filter ;
  list.lpc.action = "XDETM" ;
  if ( 0 >  do_rpc(po[0]->pAcc->pHost,service_lpc,"WM",&list) ) return ( -2 ) ;

}

/*****************************************************************
             end of anular part
*****************************************************************/

int	init_DCCT ( mode ) 

int		mode ;

{
  printf ( "#don't use init_DCCT anymore, it will be deleted in future versions\n" ) ;

} /* init_DCCT */


int	get_DCCT ( beam_current, lifetime, filling_rate ) 

double		*beam_current, *lifetime, *filling_rate ;

{
  RPC_list	list ;
  double	intcurr, energy ;
  int		cflag, lflag, injflag, intflag, engflag ;
  static int beaminfo_sid = -1;

  /* set input parameter list to do_rpc */
  list.info.curr = beam_current ; 
  list.info.cflag = &cflag ;
  list.info.life = lifetime ; 
  list.info.lflag = &lflag ;
  list.info.injrate = filling_rate ;
  list.info.injflag = &injflag ;

  /* dummy parameters, not used here */
  list.info.intcurr = &intcurr ;
  list.info.intflag = &intflag ;
  list.info.energy = &energy ;
  list.info.engflag = &engflag ;
  
  if(beaminfo_sid <0)
	  beaminfo_sid = import_beam_info(pAccInfo->pHost);
  
  if ( get_beam_info(list.info.curr,list.info.cflag,list.info.life,list.info.lflag,list.info.injrate,list.info.injflag,list.info.intcurr,list.info.intflag,list.info.energy,list.info.engflag)<0) return ( -1 ) ; 
  if ( ! cflag ) *beam_current = -1.0 ;
  if ( ! lflag ) *lifetime = 0.0 ;
  if ( ! injflag ) *filling_rate = 0.0 ; 
  
 /*  printf("get_DCCT: beam_current = %lf \n",*beam_current); */

  return ( 0 ) ;

} /* get_DCCT */


/* HERE COME THE SR PROFILE MONITOR ROUTINES */

/* Note that the access pointer has to be given since in principle,
   several profile monitors may exist in the ring. Another argument
   is that the profile monitor will be placed into the structure file,
   so its access record will be read in automatically */


int	get_SRPM_status ( pAcc, status )
access_record	*pAcc ;
long int	*status ;

{
  RPC_list	list ;
  
  /* set input parameter list to do_rpc */

  list.lpc.family = pAcc->family ;
  list.lpc.member = pAcc->member ;
  list.lpc.action = "STAT" ;
  list.lpc.datum = status ;
  return ( do_rpc ( pAcc->pHost, service_lpc, "RF", &list ) ) ; 
  
} /* get_SRPM_status */


int	get_SRPM_data ( pAcc, x,y, rmsX, rmsY )
access_record	*pAcc ;
double		*x, *y, *rmsX, *rmsY ;

{
  RPC_list	list ;
  realArray	result ;
  int		res ;
  
  result.length = 0 ;
  
  /* set input parameter list to do_rpc */

  list.lpc.family = pAcc->family ;
  list.lpc.member = pAcc->member ;
  list.lpc.action = "RES" ;
  list.lpc.datum = &result ;
  
  if ( 0 <= ( res = do_rpc(pAcc->pHost,service_lpc,"RAR",&list) ) )
  {
    *x = result.sequence[0] ;
    *y = result.sequence[1] ;
    *rmsX = result.sequence[2] ;
    *rmsY = result.sequence[3] ;
  }
  else res = -1 ;
  if ( result.length > 0 ) free ( result.sequence ) ; /* VERY IMPORTANT */

  return ( res ) ;
  
} /* get_SRPM_data */


int	unlock_FS ( ptr )

monitor_record	*ptr ;

{
  RPC_list	list ;
  int		res ;
  unsigned char	data ;
  host_record	*pHost ;


  pHost = ptr->pAcc->pHost ;

  /* set input parameter list to do_rpc */
  list.lpc.family = ptr->pAcc->family ;
  list.lpc.member = ptr->pAcc->member ;
  
  /* lock FS such that no one else can use it (UNLOCK BEFORE QUIT !!!) */
  data = 1 ;
  list.lpc.datum = &data ;
  list.lpc.action = "UNLOCK" ;
  if ( 0 > ( res = do_rpc(pHost,service_lpc,"WP",&list) ) ) 
    printf("#can not unlock FS %s_%s\n",list.lpc.family,list.lpc.member);

  return(res) ;
}


  
int	init_single_FS ( ptr )

monitor_record	*ptr ;

/* it returns -2 if the FS is locked by another user, -1 if there was
an error - in that case finish_single_FS must be called to unlock the FS */

{
  RPC_list	list ;
  int		res ;
  unsigned char	data ;
  long int	status ;
  double	percentage ;
  host_record	*pHost ;


  pHost = ptr->pAcc->pHost ;

  /* set input parameter list to do_rpc */
  list.lpc.family = ptr->pAcc->family ;
  list.lpc.member = ptr->pAcc->member ;
  
  /* lock FS such that no one else can use it (UNLOCK BEFORE QUIT !!!) */
  data = 1 ;
  list.lpc.datum = &data ;
  list.lpc.action = "ACT" ;
  if ( 0 > ( res = do_rpc(pHost,service_lpc,"WB",&list) ) ) 
  {
    printf("#FS %s_%s locked by another user\n",list.lpc.family,list.lpc.member);
    return(-2) ;
  }
#ifdef DEBUG
printf ( "after ACT, res=%i\n", res ) ;
scanf ( "%i", &res ) ;
#endif
  
  /* check status and set proper values if not set (CHEATING !) */
  list.lpc.datum = &status ;
  list.lpc.action = "STAT" ;
  if ( 0 > ( res = do_rpc(pHost,service_lpc,"RF",&list) ) ) return (-1) ;
  list.lpc.datum = &data ;
  if ( !(status & 2) )
  {  
    data = 1 ;
    list.lpc.datum = &data ;
    list.lpc.action = "FOCUS" ;
    if ( 0 > ( res = do_rpc(pHost,service_lpc,"WB",&list) ) ) return (-1) ;
  }
  
#ifdef DEBUG
printf ( "after FOCUS, res=%i\n", res ) ;
scanf ( "%i", &res ) ;
#endif

  
  /* set colour disply mode */
  list.lpc.member = "ALL" ;
  data = 1 ; 	/* 0 = black/white; 1 = colour */
  list.lpc.datum = &data ;
  list.lpc.action = "VID" ;
  if ( 0 > ( res = do_rpc ( pHost, service_lpc, "WB", &list ) ) ) return(-1) ;
  list.lpc.member = ptr->pAcc->member ; /* set back member to FS name */

  /* set method of analysis */
  data = 31 ; 	/* percentage method */
  list.lpc.datum = &data ;
  list.lpc.action = "MTHD" ;
  if ( 0 > ( res = do_rpc ( pHost, service_lpc, "WE", &list ) ) ) return(-1) ;

  percentage = 68.3 ; 	/* corresponds to one std. deviation */
  list.lpc.datum = &percentage ;
  list.lpc.action = "SIGPER" ;
  if ( 0 > ( res = do_rpc(pHost,service_lpc,"WR",&list) ) ) return(-1);


#ifdef DEBUG
printf ( "after SIGPER, res=%i\n", res ) ;
scanf ( "%i", &res ) ;
#endif

    /* enable background subtraction */
  data = 1 ;  
  list.lpc.datum = &data ;
  list.lpc.action = "BCKSUB" ;
  if ( 0 > ( res = do_rpc(pHost,service_lpc,"WB",&list) ) ) return(-1);

#ifdef DEBUG
printf ( "after BCKSUB, res=%i\n", res ) ;
scanf ( "%i", &res ) ;
#endif

  /* perform background acquisition */
  data = 1 ;  
  list.lpc.datum = &data ;
  list.lpc.action = "BCKACQ" ;
  if ( 0 > ( res = do_rpc(pHost,service_lpc,"WP",&list) ) ) return(-1);

#ifdef DEBUG
printf ( "after BCKACQ, res=%i\n", res ) ;
scanf ( "%i", &res ) ;
#endif

  
  /* start image analysis (also pushes piston down) */
  data = 1 ;
  list.lpc.datum = &data ;
  list.lpc.action = "IMGAN" ;
  if ( 0 > ( res = do_rpc(pHost,service_lpc,"WB",&list) ) ) return(-1);

#ifdef DEBUG
printf ( "after IMGAN start, res=%i\n", res ) ;
scanf ( "%i", &res ) ;
#endif

  

} /* init_single_FS */


int	acquire_single_FS ( ptr, posh, posv, sizh, sizv )

monitor_record	*ptr ;
double		*posh, *posv, *sizh, *sizv;

/* it returns -2 if the FS has no result yet to return, -1 if there was
an error - in both cases finish_single_FS must be called to unlock the FS */

{
  RPC_list	list ;
  int		res ;
  realArray	result ;
  host_record	*pHost ;


  pHost = ptr->pAcc->pHost ;

  /* set input parameter list to do_rpc */
  list.lpc.family = ptr->pAcc->family ;
  list.lpc.member = ptr->pAcc->member ;

  /* read values */
  result.length = 0 ;
  list.lpc.datum = &result ;
  list.lpc.action = "GETRES" ;
  if ( 0 > ( res = do_rpc(pHost,service_lpc,"RAR",&list) ) ) 
  {
    if ( result.length > 0 ) free ( result.sequence ) ; /* VERY IMPORTANT */
    if ( errno == FS_TIMEOUT ) return ( -2 ) ;
    else return ( -1 ) ;
  }
  else
  {
    if ( result.length != 4 ) printf
      ("#RPC array length not consistent in get_single_FS():%i\n",result.length);
    *posh = - result.sequence[0] ; /* change sign due to our convention */
    *posv = result.sequence[1] ;
    *sizh = result.sequence[2] ;
    *sizv = result.sequence[3] ;
  }
  if ( result.length > 0 ) free ( result.sequence ) ; /* VERY IMPORTANT */

} /* acquire_single_FS */


int	finish_single_FS ( ptr )

monitor_record	*ptr ;

{
  RPC_list	list ;
  int		res, res1 ;
  unsigned char	data ;
  host_record	*pHost ;


  pHost = ptr->pAcc->pHost ;

  /* set input parameter list to do_rpc */
  list.lpc.family = ptr->pAcc->family ;
  list.lpc.member = ptr->pAcc->member ;

  /* stop acquisition (also pushes piston up) */
  data = 0 ;
  list.lpc.datum = &data ;
  list.lpc.action = "IMGAN" ;
  res = do_rpc(pHost,service_lpc,"WB",&list) ;
  
  /* unlock FS */
  data = 0 ;
  list.lpc.datum = &data ;
  list.lpc.action = "ACT" ;
  if ( 0 > ( res1 = do_rpc(pHost,service_lpc,"WB",&list) ) ) 
  {
    printf("#can not unlock FS %s_%s !\n",list.lpc.family,list.lpc.member);
    return (-2) ;
  }
  else return(res) ;

} /* finish_single_FS */


int	get_single_FS ( ptr, posh, posv, sizh, sizv )

monitor_record	*ptr ;
double		*posh, *posv, *sizh, *sizv;

{
  int	res, trials ;
  
  if ( -2 == ( res = init_single_FS ( ptr ) ) ) return ( -3 ) ;
  
  if ( res >= 0 ) /* init successful, now wait for results */
  {
    trials = 0 ;
    while ( 0 > ( res = acquire_single_FS (ptr,posh,posv,sizh,sizv) ) )
    {
      if ( res != -2 ) 
      {
        printf ( "$error while acquire FS %s_%s\n",
        		ptr->pAcc->family, ptr->pAcc->member ) ;
        finish_single_FS ( ptr ) ;
        return ( -1 ) ;
      }
      if ( ++trials > FS_TRIALS ) 
      {
        printf("#max No. of sleeps exceeded for %s_%s\n",
        			ptr->pAcc->family,ptr->pAcc->member);
        res = -2 ;
        break ;
      }
    }
  }
  
  if ( 0 > finish_single_FS ( ptr ) ) return (-4) ;
  else return ( res ) ;

}


int	init_scraper_auto ( ptr )

scraper_record	*ptr ;

{
  RPC_list	list ;
  int		res ;
  unsigned char	data ;
  long int	status ;
  host_record	*pHost ;


  pHost = ptr->pAcc->pHost ;

  /* set input parameter list to do_rpc */
  list.lpc.family = ptr->pAcc->family ;
  list.lpc.member = ptr->pAcc->member ;
  
  /* read status */
  list.lpc.datum = &status ;
  list.lpc.action = "STAT" ;
  if ( 0 >  do_rpc(pHost,service_lpc,"RF",&list) ) return ( -1 ) ;

  /* now check whether scraper is still running */
  if ( status & 1 ) 
  {
    data = 0 ;
    list.lpc.datum = &data ;
    list.lpc.action = "SCRPGO" ;
    if ( 0 > ( res = do_rpc(pHost,service_lpc,"WB",&list) ) ) return(res);
  }

  /* start initializing of scraper */
  data = 1 ;
  list.lpc.datum = &data ;
  list.lpc.action = "SCRPGO" ;
  if ( 0 > ( res = do_rpc(pHost,service_lpc,"WB",&list) ) ) return(res);

} /* init_scraper_auto */


int	quit_scraper_auto ( ptr )

scraper_record	*ptr ;

{
  RPC_list	list ;
  unsigned char	data ;
  host_record	*pHost ;


  pHost = ptr->pAcc->pHost ;

  /* release scraper */
  data = 0 ;
  list.lpc.datum = &data ;
  list.lpc.action = "SCRPGO" ;
  return ( do_rpc(pHost,service_lpc,"WB",&list) ) ;

} /* quit_scraper_auto */


int	wait_scraper_auto ( ptr )

scraper_record	*ptr ;

/* it returns -2 if the scraper init has not finished yet,
   -1 if there was an error */

{
  RPC_list	list ;
  unsigned char	data ;
  long int	status ;
  host_record	*pHost ;


  pHost = ptr->pAcc->pHost ;

  /* set input parameter list to do_rpc */
  list.lpc.family = ptr->pAcc->family ;
  list.lpc.member = ptr->pAcc->member ;

  /* read status */
  list.lpc.datum = &status ;
  list.lpc.action = "STAT" ;
  if ( 0 >  do_rpc(pHost,service_lpc,"RF",&list) ) return ( -1 ) ;

  /* now check whether scraper is ready */
  if ( !( (status & 2) && (status & 4) && (status & 8) ) ) return ( -2 ) ;

  return ( 0 ) ;

} /* wait_scraper */


int	start_scraper_auto ( ptr, x_start, x_stop, slit, Nsteps )

scraper_record	*ptr ;
double		x_start, x_stop, slit ;
int		Nsteps ;

/* it returns -2 if the scraper init has not finished yet,
   -1 if there was an error */

{
  RPC_list	list ;
  unsigned char	data ;
  long int	status ;
  double	real_N ;
  host_record	*pHost ;


  pHost = ptr->pAcc->pHost ;

  /* set input parameter list to do_rpc */
  list.lpc.family = ptr->pAcc->family ;
  list.lpc.member = ptr->pAcc->member ;

  /* read status */
  list.lpc.datum = &status ;
  list.lpc.action = "STAT" ;
  if ( 0 >  do_rpc(pHost,service_lpc,"RF",&list) ) return ( -1 ) ;

  /* now check whether scraper is ready */
  if ( !( (status & 2) && (status & 4) && (status & 8) ) ) return ( -2 ) ;

  /* set scraper parameters */
  list.lpc.datum = &x_start ;
  list.lpc.action = "SCSTR" ;
  if ( 0 > do_rpc(pHost,service_lpc,"WR",&list) ) return ( -1 ) ;
  list.lpc.datum = &x_stop ;
  list.lpc.action = "SCSTP" ;
  if ( 0 > do_rpc(pHost,service_lpc,"WR",&list) ) return ( -1 ) ;
  list.lpc.datum = &slit ;
  list.lpc.action = "SLWID" ;
  if ( 0 > do_rpc(pHost,service_lpc,"WR",&list) ) return ( -1 ) ;
  real_N = (double) Nsteps ;
  list.lpc.datum = &real_N ;
  list.lpc.action = "NSAMP" ;
  if ( 0 > do_rpc(pHost,service_lpc,"WR",&list) ) return ( -1 ) ;

  /* all is set, now start scraper scan ... */
  data = 1 ; /* value is dummy for write pulse */
  list.lpc.datum = &data ;
  list.lpc.action = "AUTOST" ;
  if ( 0 >  do_rpc(pHost,service_lpc,"WP",&list) ) return ( -1 ) ;
  
} /* start_scraper_auto */


int	get_scraper_CM ( ptr, curr_CM1, curr_CM2 )

scraper_record	*ptr ;
double		*curr_CM1, *curr_CM2 ;

{
  RPC_list	list ;
  host_record	*pHost ;
  realArray	result ;
  int		res ;

  pHost = ptr->pAcc->pHost ;

  /* set input parameter list to do_rpc */
  list.lpc.family = ptr->pAcc->family ;
  list.lpc.member = ptr->pAcc->member ;

  /* read status */
  list.lpc.datum = &result ;
  list.lpc.action = "CMREAD" ;
  if ( 0 <= ( res = do_rpc(pHost,service_lpc,"RAR",&list) ) )
  {
    *curr_CM1 = result.sequence[0] ;
    *curr_CM2 = result.sequence[1] ;
  }
  else res = -1 ;
  if ( result.length > 0 ) free ( result.sequence ) ; /* VERY IMPORTANT */

  return ( res ) ;
  
} /* get_scraper_CM */


int	finish_scraper_auto ( ptr, Npoints, array )

scraper_record	*ptr ;
int		*Npoints ;
double		array[] ;

{
  RPC_list	list ;
  int		res, i ;
  unsigned char	data ;
  long int	status ;
  realArray	result ;
  double	coef ;
  host_record	*pHost ;


  pHost = ptr->pAcc->pHost ;

  /* set input parameter list to do_rpc */
  list.lpc.family = ptr->pAcc->family ;
  list.lpc.member = ptr->pAcc->member ;

  /* read status */
  list.lpc.datum = &status ;
  list.lpc.action = "STAT" ;
  if ( 0 >  do_rpc(pHost,service_lpc,"RF",&list) ) return ( -1 ) ;

  /* now check whether scraper is ready */
  if ( !( (status & 2) && (status & 4) && (status & 8) ) ) return ( -2 ) ;

  /* OK, scraper finally finished autoscan */

  /* read values */
  result.length = 0 ;
  list.lpc.datum = &result ;
  list.lpc.action = "GETRES" ;
  if ( 0 <= ( res = do_rpc(pHost,service_lpc,"RAR",&list) ) )
  {
    *Npoints = result.length - 2 ;
    coef = (result.sequence[1]-result.sequence[0])/(result.length-3) ;
    for ( i=0; i<*Npoints; i++ ) 
    {
      array[2*i] = result.sequence[0] + coef * ((double) i) ;
      array[2*i+1] = result.sequence[i+2] ;
    }
  }
  else res = -1 ;
  if ( result.length > 0 ) free ( result.sequence ) ; /* VERY IMPORTANT */

  return ( res ) ;

} /* finish_scraper_auto */

/************************************************************************/

/* here come the scraper calls for the SR */


int	status_SR_scraper ( ptr, ready, running, OK, error )

scraper_record	*ptr ;
int		*ready ;
int		*running ;
int		*OK ;
int		*error ;

{
  RPC_list	list ;
  int		status ;
  
  /* set input parameter list to do_rpc */
  list.lpc.family = ptr->pAcc->family ;
  list.lpc.member = ptr->pAcc->member ;

  /* read status */
  list.lpc.action = "STAT" ;
  list.lpc.datum = &status ;
  if ( 0 > do_rpc ( ptr->pAcc->pHost, service_lpc, "RF", &list ) ) 
  								return ( -1 ) ;
  *ready = status & (1) ;
  *running = status & (1<<1) ;
  *OK = status & (1<<2) ;
  *error = status & (1<<3) ;
  
  return ( 0 ) ;
} /* status_SR_scraper */


int	read_SR_scraper ( ptr, left, right )

scraper_record	*ptr ;
double		*left ;
double		*right ;

{
  RPC_list	list ;
  
  /* set input parameter list to do_rpc */
  list.lpc.family = ptr->pAcc->family ;
  list.lpc.member = ptr->pAcc->member ;

  /* read first the left/upper blade */
  if ( ptr->pAcc->pPar->sc.type == vertical ) list.lpc.action = "GETUP" ;
  else list.lpc.action = "GETL" ;
  list.lpc.datum = left ;
  if ( 0 > do_rpc ( ptr->pAcc->pHost, service_lpc, "RR", &list ) ) 
  								return ( -1 ) ;

  /* read next the right/lower blade */
  if ( ptr->pAcc->pPar->sc.type == vertical ) list.lpc.action = "GETLO" ;
  else list.lpc.action = "GETR" ;
  list.lpc.datum = right ;
  if ( 0 > do_rpc ( ptr->pAcc->pHost, service_lpc, "RR", &list ) ) 
  								return ( -2 ) ;
  return ( 0 ) ;
  
} /* read_SR_scraper */


/* mode corresponds to the direction: -1 = right/down, 1 = left/up */

int	move_SR_scraper_hack ( index, mode, blade )

int	index ;
int	mode ;
double	blade ;

{
  printf ( "#don't use move_SR_scraper_hack anymore, it will be deleted soon\n" ) ;
  return move_SR_scraper ( sc[index], mode, blade ) ;
}


int	move_SR_scraper ( ptr, mode, blade )

scraper_record	*ptr ;
int	mode ;
double	blade ;

{
  RPC_list	list ;
  
  /* set input parameter list to do_rpc */
  list.lpc.family = ptr->pAcc->family ;
  list.lpc.member = ptr->pAcc->member ;

  /* set first the left/upper blade */
  if ( mode == 1 )
  {
    if ( ptr->pAcc->pPar->sc.type == vertical ) list.lpc.action = "UPPER" ;
    else list.lpc.action = "LEFT" ;
    list.lpc.datum = &blade ;
    if ( 0 > do_rpc ( ptr->pAcc->pHost, service_lpc, "WR", &list ) ) 
  								return ( -1 ) ;
    /* now start movement */
    if ( ptr->pAcc->pPar->sc.type == vertical ) list.lpc.action = "UPMOVE" ;
    else list.lpc.action = "LMOVE" ;
    if ( 0 > do_rpc ( ptr->pAcc->pHost, service_lpc, "WP", &list ) ) 
  								return ( -2 ) ;
    return ( 0 ) ;
  }
  
  /* set next the right/lower blade */
  else if ( mode == -1 )
  {
    if ( ptr->pAcc->pPar->sc.type == vertical ) list.lpc.action = "LOWER" ;
    else list.lpc.action = "RIGHT" ;
    list.lpc.datum = &blade ;
    if ( 0 > do_rpc ( ptr->pAcc->pHost, service_lpc, "WR", &list ) ) 
  								return ( -1 ) ;
    /* now start movement */
    if ( ptr->pAcc->pPar->sc.type == vertical ) list.lpc.action = "LOMOVE" ;
    else list.lpc.action = "RMOVE" ;
    if ( 0 > do_rpc ( ptr->pAcc->pHost, service_lpc, "WP", &list ) ) 
  								return ( -2 ) ;
    return ( 0 ) ;
  }
  
  /* mode is not correct */
  else return ( -3 ) ;
    
} /* move_SR_scraper */



/************ end of SR scraper stuff *******************/




int	get_tune ( Qh, Qv ) 

double		*Qh ;
double		*Qv ;

{
  RPC_list	list ;
  
  /* set input parameter list to do_rpc */
  list.lpc.family = pAccTune->family ;
  list.lpc.member = pAccTune->member ;

  /* read first the horizontal tune */
  list.lpc.action = "HTUNE" ;
  list.lpc.datum = Qh ;
  if ( 0 > do_rpc ( pAccTune->pHost, service_lpc, "RR", &list ) ) return ( -1 ) ;
  
  /* now read the vertical tune */
  list.lpc.action = "VTUNE" ;
  list.lpc.datum = Qv ;
  if ( 0 > do_rpc ( pAccTune->pHost, service_lpc, "RR", &list ) ) return ( -2 ) ;

  return ( 0 ) ;
} /* get_tune() */


int	get_single_tune_measurement ( Qh, Qv ) 

double		*Qh ;
double		*Qv ;

{
  RPC_list	list ;
  unsigned char	data ;
  int		err ;
  
  /* set input parameter list to do_rpc */
  list.lpc.family = pAccTune->family ;
  list.lpc.member = pAccTune->member ;

  /* access tune system first */
  list.lpc.action = "ACT" ;
  data = 1 ;
  list.lpc.datum = &data ;
  if ( 0 > do_rpc ( pAccTune->pHost, service_lpc, "WB", &list ) ) return ( -1 ) ;
  
  /* now start the tune scan */
  list.lpc.action = "SEARCH" ;
  list.lpc.datum = &data ;
  if ( 0 > do_rpc ( pAccTune->pHost, service_lpc, "WP", &list ) ) err = -2 ;
  else
  {
    /* wait for scan to finish (takes about 4 seconds per plane */
    sleep ( TUNE_SCAN_SLEEP ) ;
    err = get_tune ( Qh, Qv ) ? -3 : 0 ;
  }
 
  /* quit tune system */
  list.lpc.action = "ACT" ;
  data = 0 ;
  list.lpc.datum = &data ;
  if ( 0 > do_rpc ( pAccTune->pHost, service_lpc, "WB", &list ) ) return ( -4 ) ;
 
  return ( err ) ;
  
} /* get_single_tune_measurement() */


int	stop_multiple_tune_measurement () 

{
  RPC_list	list ;
  unsigned char	data ;
  int		err ;
  
  /* set input parameter list to do_rpc */
  list.lpc.family = pAccTune->family ;
  list.lpc.member = pAccTune->member ;
  
  /* stop the tune scan */
  list.lpc.action = "STOP" ;
  list.lpc.datum = &data ;
  if ( 0 > do_rpc ( pAccTune->pHost, service_lpc, "WP", &list ) ) err = -1 ;
  else err = 0 ;
  
  /* quit tune system */
  list.lpc.action = "ACT" ;
  data = 0 ;
  list.lpc.datum = &data ;
  if ( 0 > do_rpc ( pAccTune->pHost, service_lpc, "WB", &list ) ) return ( -2 ) ;
  /* unlock tune system */
  /* Patched by G.S. 07/12/95 */
  list.lpc.action = "UNLOCK" ;
/*  data = 0 ; */
  list.lpc.datum = &data ;
  if ( 0 > do_rpc ( pAccTune->pHost, service_lpc, "WP", &list ) ) return ( -3 ) ; 
  return ( err ) ;
  
} /* stop_multiple_tune_measurement() */


int	start_multiple_tune_measurement () 

{
  RPC_list	list ;
  unsigned char	data ;
  int		err = 0 ;
  
  /* set input parameter list to do_rpc */
  list.lpc.family = pAccTune->family ;
  list.lpc.member = pAccTune->member ;

  /* access tune system first */
  list.lpc.action = "ACT" ;
  data = 1 ;
  list.lpc.datum = &data ;
  if ( 0 > do_rpc ( pAccTune->pHost, service_lpc, "WB", &list ) ) return ( -1 ) ;
  
  /* now start the tune scan */
  list.lpc.action = "SEARCH" ;
  list.lpc.datum = &data ;
  if ( 0 > do_rpc ( pAccTune->pHost, service_lpc, "WP", &list ) ) err = -2 ;
  else
  {
    sleep ( TUNE_SCAN_SLEEP ) ;
    /* define new peak values */
    list.lpc.action = "SETDEF" ;
   if ( 0 > do_rpc ( pAccTune->pHost, service_lpc, "WP", &list ) ) err = -3 ;
    else
    {
      list.lpc.action = "STRACQ" ;
      if ( 0 > do_rpc ( pAccTune->pHost, service_lpc, "WP", &list ) ) err = -4 ;
    }
  }
   if ( err < 0 ) if ( stop_multiple_tune_measurement () == -1) err = -5 ;
   if ( err < 0 ) if ( stop_multiple_tune_measurement () == -2) err = -6 ;
   if ( err < 0 ) if ( stop_multiple_tune_measurement () == -3) err = -7 ;
  
  return ( err ) ;
  
} /* start_multiple_tune_measurement() */



int	get_tunefeedback_DAC ( action, added_current ) 

char		*action ;
double		*added_current ;

{
    *added_current = 0.0;
    return 0 ;
} /* get_tunefeedback_DAC() */


int	get_RF_frequency ( f ) 

double		*f ;

{
  RPC_list	list ;
  
  /* set input parameter list to do_rpc */
  list.lpc.family = pAccRF->family ;
  list.lpc.member = pAccRF->member ;
  list.lpc.action = "FREQ" ;
  list.lpc.datum = f ;
  return ( do_rpc ( pAccRF->pHost, service_lpc, "RRI", &list ) ) ;

} /* get_RF_frequency() */


/* global variable which keeps the initial frequency of a measurement */
double	initial_RF_frequency = -1.0 ;


int	set_RF_frequency ( f ) 

double		f ;

{
  RPC_list	list ;
  double	current_frequency ;
  
  if ( f < pAccRF->pPar->rf.f_low || f > pAccRF->pPar->rf.f_high )
    return ( -1 ) ;
  
  /* read in current frequency */
  if ( 0 > get_RF_frequency ( &current_frequency ) ) return ( -2 ) ;

  /* if first setting, record initial frequency */
  if ( initial_RF_frequency < 0.0 ) initial_RF_frequency = current_frequency ;
    
  /* check if requested frequency is within limits of current freq. */
  if ( fabs( current_frequency - f ) > MAX_DELTA_FREQUENCY ) return(-3) ;

  /* check if requested frequency is within limits of initial freq. */
  if ( fabs( initial_RF_frequency - f ) > MAX_TOTAL_DELTA_FREQUENCY ) return(-4) ;

  /* set input parameter list to do_rpc */
  list.lpc.family = pAccRF->family ;
  list.lpc.member = pAccRF->member ;
  list.lpc.action = "FREQ" ;
  list.lpc.datum = &f ;
  if ( 0 > do_rpc ( pAccRF->pHost, service_lpc, "WRI", &list ) ) return (-5);
  
  return ( 0 ) ;

} /* set_RF_frequency() */





/* here follow some utility routines for the PS setting and reading;
   the problem is, that different PS need different treatment, there-
   fore I have to hardcode family_member names in the code. Sorry ! */

int big_SR_PS ( pAcc )

access_record	*pAcc ;

{
  /* all corrector PS including the ones from the SR*/
  if ( !strncmp("PSC",pAcc->family,3) ) return ( 0 ) ;
  if ( !strncmp("PSZ",pAcc->family,3) ) return ( 0 ) ;

  /* all other SR PS */
  if ( !strncmp("PS",pAcc->family,2) && !strncmp("S",pAcc->member,1) ) return ( 1 ) ;
 
  /* all other power supplies and other elements (may not even be PS) */
  return ( 0 ) ;

}


int	ramping ( pAcc, last_current ) 

access_record	*pAcc ;
double		last_current ;

{
  RPC_list	list ;
  unsigned char	data ; 

  /* set input parameter list to do_rpc */
  list.lpc.family = pAcc->family ;
  list.lpc.member = pAcc->member ;

  list.lpc.datum = &data ;
  list.lpc.action = "BUSY" ;
  if ( 0 > do_rpc ( pAcc->pHost, service_lpc, "RB", &list ) ) return ( -1 ) ;
  return ( (int) data ) ;

} /* ramping */


/* it is kept static, i.e. local, because it should not be used directly */

static int set_DAC ( pAcc, current ) 

access_record	*pAcc ;
double		current ;

{
  WTIPOL_data	data ;
  RPC_list	list ;
  int		res ;
  clock_t	t0 ;

  /* set input parameter list to do_rpc */
  list.lpc.family = pAcc->family ;
  list.lpc.member = pAcc->member ;
  list.lpc.action = "RAMP" ;
  list.lpc.datum = &data ;

  data.curr = current ;
  data.iperc1 = 0 ;
  data.iperc2 = 100 ;
  data.alpha = pAcc->pPar->ps.rmp_f ;	/* use fast slope for ramp */

  if ( 0 > (res = do_rpc(pAcc->pHost, service_lpc, "WTIPOL", &list) ) ) return ( res ) ;
   
  /* wait for ramp to finish */
  printf ( "#waiting %i seconds for system ramp of %s_%s to finish\n", 
              (int)(res/1000) + 1, pAcc->family, pAcc->member ) ;
  sleep ( (int)(res/1000) + 1 ) ;

  /* check again, to be sure that ramping is over */
  while ( ramping ( pAcc, current ) ) sleep ( 1 ) ; /* can not sleep less than 1 second */
  return ( 0 ) ;
} /* set_DAC */


static int set_fast ( pAcc, current ) 

access_record	*pAcc ;
double		current ;

{
  WTIPOL_data	data ;
  RPC_list	list ;
  int		res ;
  clock_t	t0 ;

  /* set input parameter list to do_rpc */
  list.lpc.family = pAcc->family ;
  list.lpc.member = pAcc->member ;
  list.lpc.action = "RAMP" ;
  list.lpc.datum = &data ;

  data.curr = current ;
  data.iperc1 = pAcc->pPar->ps.i1 ;
  data.iperc2 = pAcc->pPar->ps.i2 ;
  data.alpha = pAcc->pPar->ps.rmp_f ;	/* use fast slope for ramp */

  if ( 0 > (res = do_rpc(pAcc->pHost, service_lpc, "WTIPOL", &list) ) ) return ( res ) ;
   
  /* wait for ramp to finish */
  printf ( "#waiting %i seconds for system ramp of %s_%s to finish\n", 
              (int)(res/1000) + 1, pAcc->family, pAcc->member ) ;
  sleep ( (int)(res/1000) + 1 ) ;

  /* check again, to be sure that ramping is over */
  while ( ramping ( pAcc, current ) ) sleep ( 1 ) ; /* can not sleep less than 1 second */
  return ( 0 ) ;
} /* set_fast */


static int set_bbb ( pAcc, current )

access_record	*pAcc ;
double		current ;

{
  RPC_list	list ;
  
  /* set input parameter list to do_rpc */
  list.lpc.family = pAcc->family ;
  list.lpc.member = pAcc->member ;
  list.lpc.action = "CURR" ;
  list.lpc.datum = &current ;

  return ( do_rpc ( pAcc->pHost, service_lpc, "WRI", &list ) ) ;
}


/* this routine checks the status longword of a PS. It returns ok=TRUE,
   if the PS is on and set in remote, ok=FALSE otherwise. The parameter
   gate is meaningfull only for the big SR PS. It is TRUE for gate on
   FALSE for gate off. */


int	get_PS_status ( pAcc, ok, gateON )

access_record	*pAcc ;
int		*ok ;
int		*gateON ;

{
  RPC_list	list ;
  long int	status ;
  
  /* set input parameter list to do_rpc */
  list.lpc.family = pAcc->family ;
  list.lpc.member = pAcc->member ;

  /* read status longword*/
  list.lpc.datum = &status ;
  list.lpc.action = "STAT" ;
  if ( 0 >  do_rpc(pAcc->pHost,service_lpc,"RF",&list) ) return ( -1 ) ;

  /* now check whether PS is on and remote*/
  
  /* different type of PS have different status bits !? */
  /* I don't like this, but I have to hard-code element names, sorry. */

  /* first check on the gate */
  if  ( big_SR_PS ( pAcc ) ) *gateON = !(status & (1<<3) ) ;
  else *gateON = 1 ; /* gate is always on for nongated PS */

  /* all corrector PS */
  if ( !strncmp("PSC",pAcc->family,3) )
  {
    *ok = !( status & (1<<6) ) && !( status & (1<<7) ) ;
    return ( 0 ) ;
  }

  /* all zigzag PS */
  if ( !strncmp("PSZ",pAcc->family,3) )
  {
    *ok = !( status & (1<<6) ) && !( status & (1<<7) ) ;
    return ( 0 ) ;
  }

  /* all bends and all SR PS (except the correctors) */
  if ( !strncmp("PSB",pAcc->family,3) || !strncmp("S",pAcc->member,1) )
  {
    *ok = !( status & (1) ) && !( status & (1<<2) ) ;
    return ( 0 ) ;
  }

  /* left are only the TL quads (no if, because they have divers names) */
  *ok = !(status & (1<<6) ) && !(status & (1<<7) ) ;
  return ( 0 ) ;

} /* get_PS_status */


int	switch_PS_on ( pAcc ) 

access_record	*pAcc ;

{
  RPC_list	list ;
  unsigned char	data ;
  int		on_and_remote, gateON ;
  
  /* if the PS is already on, return immediately */
  if ( 0 >  get_PS_status (pAcc,&on_and_remote,&gateON) ) 
  {
    printf ( "#%s_%s status can not be read !\n", pAcc->family,pAcc->member ) ;
    return ( -3 ) ;
  }
  if ( on_and_remote && gateON ) return ( 1 ) ;

  /* first set DAC to I_start */
  if ( 0 > set_DAC ( pAcc, pAcc->pPar->ps.I_start ) ) return ( -2 ) ;

  /* set input parameter list to do_rpc */
  list.lpc.family = pAcc->family ;
  list.lpc.member = pAcc->member ;
  list.lpc.datum = &data ;
  
  /* first reset PS in case there was a not ready signal */
  list.lpc.action = "RESET" ;
  data = 1 ;
  if ( 0 > do_rpc ( pAcc->pHost, service_lpc, "WP", &list ) ) return ( -1 ) ;

  /* enable PS */
  list.lpc.action = "OFF" ;
  data = 1 ;
  if ( 0 > do_rpc ( pAcc->pHost, service_lpc, "WB", &list ) ) return ( -1 ) ;

  if ( big_SR_PS ( pAcc ) ) /* special treatment of big SR power supplies */
  {
    /* make first GATE OFF to be sure */
    list.lpc.action = "GATOFF" ; 
    list.lpc.datum = &data ;
    if ( 0 > do_rpc ( pAcc->pHost, service_lpc, "WP", &list ) ) return ( -3 ) ;
  }
  
  /* now switch on PS */  
  list.lpc.action = "ON" ;
  list.lpc.datum = &data ;
  if ( 0 > do_rpc ( pAcc->pHost, service_lpc, "WP", &list ) ) return ( -4 ) ;

  if ( big_SR_PS ( pAcc ) ) /* special treatment of big SR power supplies */
  {
    /* switch now GATE ON */
    list.lpc.action = "GATON" ; 
    list.lpc.datum = &data ;
    if ( 0 > do_rpc ( pAcc->pHost, service_lpc, "WP", &list ) ) return ( -4 ) ;
  }

  /* check status again, if PS crashed during switch on */
  sleep ( 1 ) ; /* wait for switch to close */
  if ( 0 >  get_PS_status (pAcc,&on_and_remote,&gateON) ) 
  {
    printf ( "#%s_%s status can not be read !\n", pAcc->family,pAcc->member ) ;
    return ( -3 ) ;
  }
  if ( on_and_remote && gateON ) return ( 0 ) ;
  else return ( -5 ) ; /* still off after switch on */

} /* switch_on */


int	switch_PS_off ( pAcc ) 

access_record	*pAcc ;

{
  RPC_list	list ;
  unsigned char	data ;
  char		mode[3] ;

  /* set DAC to I_start first */
  if ( 0 > set_fast ( pAcc, pAcc->pPar->ps.I_start ) ) return ( -2 ) ;

  /* set input parameter list to do_rpc */
  list.lpc.family = pAcc->family ;
  list.lpc.member = pAcc->member ;
  
  /* disable PS */
  if ( big_SR_PS ( pAcc ) ) /* special treatment of big SR power supplies */
  {
    /* switch off gate first, but then also switch, unless element is a bend*/
    list.lpc.action = "GATOFF" ; 
    strcpy ( mode, "WP" ) ;
    if ( 0 > do_rpc ( pAcc->pHost, service_lpc, mode, &list ) ) return ( -3 ) ;
  }
  
  /* switch PS off (unless PSB) */
  if ( 0 != strcmp(pAcc->family,"PSB") || 0 != strcmp(pAcc->member,"S") )
  {
    list.lpc.action = "OFF" ;
    strcpy ( mode, "WB" ) ;
  }
  list.lpc.datum = &data ;
  data = 0 ;
  if ( 0 > do_rpc ( pAcc->pHost, service_lpc, mode, &list ) ) return ( -1 ) ;
  return ( 0 ) ;
} /* switch_off */


int	get_single_current ( pAcc, readback, current ) 

access_record	*pAcc ;
double		*readback ;
double		*current ;

{
  RPC_list	list ;
  double	extra_current ;
  
  /* set input parameter list to do_rpc */
  list.lpc.family = pAcc->family ;
  list.lpc.member = pAcc->member ;

  /* read first the readback */
  list.lpc.action = "CURR" ;
  list.lpc.datum = readback ;
  if ( 0 > do_rpc ( pAcc->pHost, service_lpc, "RR", &list ) ) return ( -1 ) ;
  
  /* now read the setting */
  list.lpc.action = "CURSET" ;
  list.lpc.datum = current ;
  if ( 0 > do_rpc ( pAcc->pHost, service_lpc, "RR", &list ) ) return ( -2 ) ;

  /* apologies again: I have to hardcode the family names of QF and QD.
     I really hate this, since the code is less and less portable, but
     these are the famous special cases !                               */

  if ( !strcmp ( pAcc->family, "PSQF" ) ) 
  {
    if ( 0 <= get_tunefeedback_DAC ( "QFDAC", &extra_current ) )
      *current += extra_current ;
    else 
    {
      printf ( "\n #Warning: cannot read tunefeedback DAC of QF\n" ) ;
      return ( -4 ) ;
    }
  }
  else if ( !strcmp ( pAcc->family, "PSQD" ) ) 
  {
    if ( 0 <= get_tunefeedback_DAC ( "QDDAC", &extra_current ) )
      *current += extra_current ;
    else 
    {
      printf ( "\n #Warning: cannot read tunefeedback DAC of QD\n" ) ;
      return ( -4 ) ;
    }
  } 

  /* crude check: give error if reading and setting differ by more than 1 Amp*/
  if ( fabs ( *readback - *current ) > 1.0 ) 
  {
    printf ( "\n #Warning: read and set values differ for %s_%s\n",
    pAcc->family, pAcc->member ) ;
    return ( -3 ) ;
  }
  else return ( 0 ) ;
}


int	get_single_voltage ( pAcc, readback, voltage ) 

access_record	*pAcc ;
double		*readback ;
double		*voltage ;

{
  RPC_list	list ;
  int		i,err ;
  
  /* set input parameter list to do_rpc */
  list.lpc.family = pAcc->family ;
  list.lpc.member = pAcc->member ;


  list.lpc.action = "PSHV" ;
  list.lpc.datum = readback ;

  /* repeat at most N times if an error has occured */
  for ( i=0;i<N_INJ_ATTEMPT; i++ )
  {
    err =  do_rpc ( pAcc->pHost, service_lpc, "RRI", &list ) ;
    if ( err >= 0 ) break ;
  }
  if ( 0 > err ) return ( -1 ) ;

  list.lpc.action = "PSVSET" ;
  list.lpc.datum = voltage ;

  /* repeat at most N times if an error has occured */
  for ( i=0;i<N_INJ_ATTEMPT; i++ )
  {
    err =  do_rpc ( pAcc->pHost, service_lpc, "RRI", &list ) ;
    if ( err >= 0 ) break ;
  }
  if ( 0 > err ) return ( -2 ) ;

  return ( 0 ) ;
} /* get_single_voltage */


int	set_voltage ( pAcc, voltage ) 

access_record	*pAcc ;
double		voltage ;

{
  RPC_list	list ;
  int		i,err ;
  
  /* set input parameter list to do_rpc */
  list.lpc.family = pAcc->family ;
  list.lpc.member = pAcc->member ;


  list.lpc.action = "PSHV" ;
  list.lpc.datum = &voltage ;

  /* repeat at most N times if an error has occured */
  for ( i=0;i<N_INJ_ATTEMPT; i++ )
  {
    err =  do_rpc ( pAcc->pHost, service_lpc, "WRI", &list ) ;
    if ( err >= 0 ) break ;
  }
  if ( 0 > err ) return ( -1 ) ;

  return ( 0 ) ;
} /* set_voltage */
  

int	read_current_from_file ( fileName, pAcc, readback, current ) 

str32		fileName ;		
access_record	*pAcc ;
double		*readback ;
double		*current ;

{
  str13	fam_mem ;
  
  strcpy (fam_mem, pAcc->family) ;
  strcat (fam_mem, "_") ;
  strcat (fam_mem, pAcc->member) ;
  return ( read_machine_file_current (fileName, fam_mem, readback, current) ) ;
}


int	set_current_fast ( pAcc, current ) 

access_record	*pAcc ;
double		current ;

{
  WTIPOL_data	data ;
  RPC_list	list ;
  int		on_and_remote, gateON ;
  int		res ;
  double	old_current ;  
  double	dummy ;  
  
  /* read status first*/
  if ( 0 >  get_PS_status (pAcc,&on_and_remote,&gateON) ) 
  {
    printf ( "#%s_%s status can not be read !\n", pAcc->family,pAcc->member ) ;
    return ( -3 ) ;
  }

  /* now check whether PS in on */
  if ( !(on_and_remote && gateON) ) 
  {
    printf ( "#%s_%s is off or local, can not set !\n", pAcc->family,pAcc->member ) ;
    return ( -2 ) ;
  }

  /* set input parameter list to do_rpc */
  list.lpc.family = pAcc->family ;
  list.lpc.member = pAcc->member ;

  /* check on type of PS */
  if ( big_SR_PS ( pAcc ) )
  {
    /* special procedure for the big ones if reading below I_start (defined by R.Richter) */
    if ( current < pAcc->pPar->ps.I_start )
    {
      res = get_single_current ( pAcc, &old_current, &dummy ) ;
      /* quit if RPC error, but not if reading differs from setting (res=-3) */
      if ( res < 0 && res != -3 ) return ( -2 ) ;
      if ( current < old_current )
      {
        if ( old_current > pAcc->pPar->ps.I_start )
          /* go down fast, but only until I_start */
          if ( 0 > set_fast ( pAcc, pAcc->pPar->ps.I_start ) ) return ( -3 ) ;
        /* going down, below I_start => go bit by bit ! */
        if ( 0 > set_bbb ( pAcc, current ) ) return ( -4 ) ;
        /* return the estimated bbb setting time in milliseconds */
        /* we assume that I_high corresponds to the maximum hex setting */
        return ( (int) ( (old_current-current)/pAcc->pPar->ps.I_high * FULL_TIME ) ) ;
      }
    }
  }
  /* all other cases, do a fast ramp ! */
  list.lpc.action = "RAMP" ;
  list.lpc.datum = &data ;

  data.curr = current ;
  data.iperc1 = pAcc->pPar->ps.i1 ;
  data.iperc2 = pAcc->pPar->ps.i2 ;
  data.alpha = pAcc->pPar->ps.rmp_f ;	/* use fast slope for ramp */

  return ( do_rpc ( pAcc->pHost, service_lpc, "WTIPOL", &list ) ) ;
}


int	set_current_slow ( pAcc, current )

access_record	*pAcc ;
double		current ;

{
  RPC_list	list ;
  int		on_and_remote, gateON ;  
  
  /* read status first*/
  if ( 0 >  get_PS_status (pAcc,&on_and_remote,&gateON) ) 
  {
    printf ( "#%s_%s status can not be read !\n", pAcc->family,pAcc->member ) ;
    return ( -3 ) ;
  }

  /* now check whether PS is on */
  if ( !(on_and_remote && gateON) ) 
  {
    printf ( "#%s_%s is off or local, can not set !\n", pAcc->family,pAcc->member ) ;
    return ( -2 ) ;
  }
  
  /* set input parameter list to do_rpc */
  list.lpc.family = pAcc->family ;
  list.lpc.member = pAcc->member ;
  list.lpc.action = "CURR" ;
  list.lpc.datum = &current ;

  return ( do_rpc ( pAcc->pHost, service_lpc, "WRI", &list ) ) ;
}
  

/*========================================================================*/
/* ID read and set gaps                                                   */
/*========================================================================*/

int	get_single_IDgap ( pAcc, readback, gap ) 

access_record	*pAcc ;
double		*readback ;
double		*gap ;

{
  RPC_list	list ;
  int		err, i ;
  
  /* set input parameter list to do_rpc */
  list.lpc.family = pAcc->family ;
  list.lpc.member = pAcc->member ;

  /* read first the readback */
  *readback = -1.0 ; /* can not read the readback, sorry */
  
  /* read now the setting of the gap */
  list.lpc.action = "GAP" ;
  list.lpc.datum = gap ;
  
  /* repeat at most N times if an error has occured */
  for ( i=0;i<N_ID_ATTEMPT; i++ )
  {
    err =  do_rpc ( pAcc->pHost, service_lpc, "RRI", &list ) ;
    if ( err >= 0 ) break ;
  }
  
  return ( err )  ;
  
} /* get_single_IDgap */


int	get_single_IDphase ( pAcc, phase ) 

access_record	*pAcc ;
double		*phase ;

{
  RPC_list	list ;
  int		err, i ;
  
  /* set input parameter list to do_rpc */
  list.lpc.family = pAcc->family ;
  list.lpc.member = pAcc->member ;

  
  /* read now the setting of the gap */
  list.lpc.action = "PHAM" ;
  list.lpc.datum = phase ;
  
  /* repeat at most N times if an error has occured */
  for ( i=0;i<N_ID_ATTEMPT; i++ )
  {
    err =  do_rpc ( pAcc->pHost, service_lpc, "RRI", &list ) ;
    if ( err >= 0 ) break ;
  }
  
  return ( err )  ;
  
} /* get_single_IDphase */

/* =============================================================*/
/* F. Iazzourene = 6 july 2016
 * get phase of apu type IDs
 */
int	get_apu_IDphase ( pAcc, phase ) 

access_record	*pAcc ;
double		*phase ;

{
RPC_list	list ;
  int		err, i ;
  
  /* set input parameter list to do_rpc */
  list.lpc.family = pAcc->family ;
  list.lpc.member = pAcc->member ; 

  /* read now the setting of the phase */
  list.lpc.action = "PHA" ;
  list.lpc.datum = phase ;
  
  /* repeat at most N times if an error has occured */
  for ( i=0;i<N_ID_ATTEMPT; i++ )
  {
    err =  do_rpc ( pAcc->pHost, service_lpc, "RR", &list ) ;
    if ( err >= 0 ) break ;
  }
  
  return ( err )  ; 
  
/*   return 0; */
  
} /* get_apu_IDphase */

/* -------------------------------------------------------------*/

/* =============================================================*/
/* C. Scafuri - 19 jun. 2002
 * get gap, taper and phase of sazaki type IDs
 */
 
 int	get_sazaki_IDgataph ( pAcc, gap,taper,phase ) 

access_record	*pAcc ;
double          *gap;
double          *taper;
double		*phase ;

{
  RPC_list	list ;
  int		err, i ;
  realArray     gataph;
  
  /* set input parameter list to do_rpc */
  list.lpc.family = pAcc->family ;
  list.lpc.member = pAcc->member ;

  
  /* read now the setting of the gap */
  list.lpc.action = "GATAPH" ;
  list.lpc.datum = &gataph ;
  gataph.length=0;
  /* repeat at most N times if an error has occured */
  for ( i=0;i<N_ID_ATTEMPT; i++ )
  {
    err =  do_rpc ( pAcc->pHost, service_lpc, "RAR", &list ) ;
    if ( err >= 0 ) break ;
  }
  if(gataph.length!=0){
    *gap=gataph.sequence[0];
    *taper=gataph.sequence[1];
    *phase=gataph.sequence[2];
    free(gataph.sequence);
  }
  
  
  return ( err )  ;
  
} /* get_sazaki_IDgataph */


int	get_single_IDsazaki ( pAcc, gap, phase ) 

access_record	*pAcc ;
double		*gap ;
double		*phase ;

{
  RPC_list	list ;
  realArray	data ;
  int		err, i ;
  
  /* set input parameter list to do_rpc */
  list.lpc.family = pAcc->family ;
  list.lpc.member = pAcc->member ;
  list.lpc.datum = &data ;
  list.lpc.action = "GATAPH" ;
  data.length=0;
  data.sequence=0;

  
  
  /* repeat at most N times if an error has occured */
  for ( i=0;i<N_ID_ATTEMPT; i++ )
  {
    err =  do_rpc ( pAcc->pHost, service_lpc, "RAR", &list ) ;
    if ( err >= 0 ) break ;
  }
  if ( err >=0 )
  {
    *gap=data.sequence[0];
    *phase=data.sequence[2];
  }
  if ( data.length != 0 && data.sequence !=0) 
  {
    free ( data.sequence ) ;
    data.length = 0 ;
  }
  return ( err )  ;
  
} /* get_single_IDsazaki */

/* -------------------------------------------------------------*/

int	get_ID_status ( pAcc, powerON, enabled, brakeOff, overlimit,
			ready, moving, error, local ) 

access_record	*pAcc ;
int		*powerON, *enabled, *brakeOff, *overlimit,
		*ready, *moving, *error, *local ;

{
  RPC_list	list ;
  int		status ;
  
  /* set input parameter list to do_rpc */
  list.lpc.family = pAcc->family ;
  list.lpc.member = pAcc->member ;

  /* read first the readback */
  list.lpc.action = "STAT" ;
  list.lpc.datum = &status ;
  if ( 0 > do_rpc ( pAcc->pHost, service_lpc, "RFI", &list ) ) return ( -1 ) ;
  
  *powerON   = status & 1 ;
  *enabled   = status & (1<<1) ;
  *brakeOff  = status & (1<<2) ;
  *overlimit = status & (1<<3) ;
  *ready     = status & (1<<4) ;
  *moving    = status & (1<<5) ;
  *error     = status & (1<<6) ;
  *local     = status & (1<<7) ;
  
  return ( 0 ) ;
  
} /* get_ID_status */


int	get_ID_speed ( pAcc, speed ) 

access_record	*pAcc ;
double		*speed ;

{
  RPC_list	list ;
  
  /* set input parameter list to do_rpc */
  list.lpc.family = pAcc->family ;
  list.lpc.member = pAcc->member ;

  /* read first the readback */
  list.lpc.action = "SPEED" ;
  list.lpc.datum = speed ;
  return ( do_rpc ( pAcc->pHost, service_lpc, "RRI", &list ) ) ; 
  
} /* get_ID_speed */


int	set_ID_brake ( pAcc, on ) 

access_record	*pAcc ;
int		on ;

{
  RPC_list	list ;
  int		dummy ;
  
  /* set input parameter list to do_rpc */
  list.lpc.family = pAcc->family ;
  list.lpc.member = pAcc->member ;
  if ( on ) list.lpc.action = "BRON" ;
  else list.lpc.action = "BROFF" ;
  list.lpc.datum = &dummy ;
  return ( do_rpc ( pAcc->pHost, service_lpc, "WPI", &list ) ) ; 
  
} /* set_ID_brake */


int	check_ID_loop ( pAcc, loop ) 

access_record	*pAcc ;
int		*loop ;

{
  RPC_list	list ;
  int		status, err ;
  
  /* This command reads the block of a wiggler only, not an undulator */
  /* Warning: no check is made whether this family exists for an ID */
  
  /* set input parameter list to do_rpc */
  list.lpc.family = "IDBLK" ;
  list.lpc.member = pAcc->member ;
  list.lpc.action = "GLSTAT" ;
  list.lpc.datum = &status ;
  *loop = status & (1<<7) ;

  if ( 0 <= ( err = do_rpc ( pAcc->pHost, service_lpc, "RFI", &list ) ) )
    *loop = status & (1<<7) ;
    
  return ( err ) ;
  
} /* check_ID_loop */




int	set_ID_loop ( pAcc, on_off ) 

access_record	*pAcc ;
int		on_off ;

{
  RPC_list	list ;
  unsigned char	datum ;
  
  /* This command sets the loop of a wiggler only, not an undulator */
  /* Warning: no check is made whether this family exists for an ID */
  
  /* set input parameter list to do_rpc */
  list.lpc.family = "IDBLK" ;
  list.lpc.member = pAcc->member ;
  list.lpc.action = "LOOPON" ;
  list.lpc.datum = &datum ;
  datum = (unsigned char) on_off ;

  return ( do_rpc ( pAcc->pHost, service_lpc, "WBI", &list )  ) ;
  
} /* set_ID_loop */



int	set_IDgaps ( N, id1, id2, id3, gaps )

int		N ;
ID_record	*id1, *id2, *id3 ;
double		gaps[] ;

{
  RPC_list	list ;
  realArray	block ;
  double	data[3] ;
  str6		family, member, memall, memdummy ;
  ID_record	*idlist[3] ;
  int		idx, IDindex[3] ;
  access_record	*pAcc ;
  int		i,j, pos ;
  int		powerON, enabled, brakeOff, overlimit,
		ready, moving, error, local, loop ;
  
  /* copy id pointers into one array for easier processing */
  idlist[0] = id1 ;
  idlist[1] = id2 ;
  idlist[2] = id3 ;  
  
  /* initialize data array for rpc */
  for ( i=0;i<3;i++ ) data[i] = 0.0 ;

  /* get generic member name of all three IDs */
  memset ( memall, 0, sizeof ( memall ) ) ;
  strncpy ( memall, idlist[0]->pAcc[0]->member, strcspn(idlist[0]->pAcc[0]->member,".") ) ;

  strcpy ( family, idlist[0]->pAcc[0]->family ) ;

  /* first loop over all requested carriages of the ID, just to check names */
  for (i=0;i<N;i++) 
  {
    /* get right access pointer */
    pAcc = idlist[i]->pAcc[0] ;

    /* get right member name */
    strcpy ( member, pAcc->member ) ;
    
    /* check whether expected member name is correct */
    memset ( memdummy, 0, sizeof ( memdummy ) ) ;
    pos = strcspn(member,".") ;
    if ( strcmp ( memall, strncpy ( memdummy, member, pos ) ) ) 
      return ( -10*(i+1)-1 ) ; 
      
    /* find index of ID in the section - ID appears twice in parameter list ? */
    idx = member[pos+1] - '1' ;
    for (j=0;j<i;j++) if ( idx == IDindex[j] )  return ( -10*(i+1)-1 ) ; 
    IDindex[i] = idx ;
  }

  /* now the full loop over all requested carriages of the ID */
  for (i=0;i<N;i++) 
  {
    /* get right access pointer */
    pAcc = idlist[i]->pAcc[0] ;

    /* get right member name */
    strcpy ( member, pAcc->member ) ;
        
    /* fill array with proper conventions (0 gap means no move) */
    /* place gap into proper order as expected by the rpc routine */
    if ( gaps[i] < 0.0 ) { data[IDindex[i]] = 0.0 ; break ; }
    else data[IDindex[i]] = gaps[i] ;
   
    /* check status first */
    if ( 0 > get_ID_status ( pAcc, &powerON, &enabled, &brakeOff, &overlimit,
		&ready, &moving, &error, &local ) ) return ( -10*(i+1)-2 ) ;
		
    if ( !powerON || !enabled || overlimit ||
         !ready || moving || local ) return ( -10*(i+1)-3 ) ;
         
    /* turn brakes off if still on */
    if ( !brakeOff )
    {
      if ( 0 > set_ID_brake ( pAcc, 0 ) ) return ( -10*(i+1)-4 ) ;
    }

    /* if the type is a wiggler, check also its loop status */
    if ( idlist[i]->pPar->id.type == hybrid )
    {
      if ( 0 > check_ID_loop ( pAcc, &loop ) ) return ( -10*(i+1)-5 ) ;
      if ( !loop ) return ( -10*(i+1)-6 ) ;
    }
  }

  /* set input parameter list to do_rpc */
  list.lpc.family = "ID" ;
  list.lpc.member = memall ;
    
  /* set rpc pointers */
  block.length = 3 ;
  block.sequence = data ;
  list.lpc.action = "GAP" ;
  list.lpc.datum = &block ;
  
  return ( do_rpc ( pAcc->pHost, service_lpc, "WAIR", &list ) ) ;

}

  
int	get_PHBSH_status (pAcc, condition, remote )
  access_record	*pAcc ;
  int *condition, *remote;
{
  RPC_list	list ;
  long int	status ;

    
    /* set input parameter list to do_rpc */
    list.lpc.family = pAcc->family ;
    list.lpc.member = pAcc->member ;

    /* read status longword*/
    list.lpc.datum = &status ;
    list.lpc.action = "STAT" ;
    if ( 0 >  do_rpc(pAcc->pHost,service_lpc,"RF",&list) ) return(-1) ;

    /* now check whether PHBSH is open and remote*/
  
    else
    {
      *condition=!(status & 1 ) ;
      *remote=!(status & (1<<4));
    }
      
  return ( 0 ) ;

} /* get_PHBSH_status */

int	close_PHBSH (pAcc )
  access_record	*pAcc ;
{
  RPC_list	list ;
  int   	close ;


    close=1;
    /* set input parameter list to do_rpc */
    list.lpc.family = pAcc->family ;
    list.lpc.member = pAcc->member ;

    /* read status longword*/
    list.lpc.datum = &close ;
    list.lpc.action = "CLOSE" ;
    if ( 0 >  do_rpc(pAcc->pHost,service_lpc,"WB",&list) ) return(-1) ;

    return ( 0 ) ;

} /* close_PHBSH */

/*========================================================================*/
/* multiple point routines                                                */
/*========================================================================*/

int	read_TL_twiss ( twiss )

twiss_record	*twiss ;

{
  FILE	*f ;
  char	*chPtr ;
  char	fullName[N80] ;
       
  if ( NULL == ( chPtr = getenv ( "HLSETC" ) ) ) *fullName='\0' ;
  else strcpy ( fullName, chPtr ) ;
  strcat ( fullName, "TL_twiss_default" ) ;

  if ( (f = fopen (fullName, "r")) == NULL ) {
    printf ("#file with TL twiss data not found: %s\n", fullName) ;
    return (-1) ;
  }
  fscanf ( f, "%lf %lf %lf %lf %lf %lf %lf %lf", 
  	&twiss->beta[horizontal], 	&twiss->beta[vertical],
  	&twiss->alpha[horizontal], 	&twiss->alpha[vertical],
  	&twiss->D[horizontal], 		&twiss->D[vertical],
  	&twiss->Dprime[horizontal], 	&twiss->Dprime[vertical] ) ;
  fclose (f) ;
  return( 0 ) ;
}
  

int	read_TL_emittance ( twiss, emittance )

twiss_record	*twiss ;
double		emittance[2] ;
{
  FILE	*f ;
  char	*chPtr ;
  char	fullName[N80] ;
       
  if ( NULL == ( chPtr = getenv ( "HLSETC" ) ) ) *fullName='\0' ;
  else strcpy ( fullName, chPtr ) ;
  strcat ( fullName, "TL_twiss_default" ) ;

  if ( (f = fopen (fullName, "r")) == NULL ) {
    printf ("#file with TL twiss data not found: %s\n", fullName) ;
    return (-1) ;
  }
  fscanf ( f, "%lf %lf %lf %lf %lf %lf %lf %lf", 
  	&twiss->beta[horizontal], 	&twiss->beta[vertical],
  	&twiss->alpha[horizontal], 	&twiss->alpha[vertical],
  	&twiss->D[horizontal], 		&twiss->D[vertical],
  	&twiss->Dprime[horizontal], 	&twiss->Dprime[vertical] ) ;
  fscanf ( f, "%lf %lf", &emittance[0], &emittance[1] ) ;
  fclose (f) ;
  return( 0 ) ;
}


int	field_to_energy ( where, current, E )

const int	where ;
double		current ;
double		*E ;

{
  int i ;
  calibration_record *ptr ;
  
  if ( where != TL && where != SR ) return ( -1 ) ;

  ptr = energy.pAcc[where-TL]->pCal1 ;
  if (ptr->Nc<1 || ptr->Nc>AL_calibration) return ( -2 ) ;   
  *E = ptr->c[ptr->Nc-1] ;
  for ( i=ptr->Nc-2;i>=0;i-- ) *E = *E * current + ptr->c[i] ;

  return ( 0 ) ;
}

  
int	read_energy_from_field ( where )

const int	where ;

{
  int i, res ;
  double readback, current ;
  calibration_record *ptr ;
  
  if ( where != TL && where != SR ) return ( -1 ) ;

  res = get_single_current ( energy.pAcc[where-TL], &readback, &current ) ;

  /* quit if RPC error, but not if reading differs from setting (res=-3) */
  if ( res < 0 && res != -3 ) return ( -2 ) ;

  res = field_to_energy ( where, readback, &(energy.readback) ) ;
  res += field_to_energy ( where, current, &(energy.E) ) ;
  
  /* res = -2 means that one field_to_energy was unsuccessful (see above)*/
  if ( res == -2 ) return ( -3 ) ;
  else return ( res ) ;
}


int	read_energy_from_file ( fileName )

str32		fileName ;		

{
  str13	fam_mem ;
  
  strcpy (fam_mem, ORACLEenergy) ;
  return ( read_machine_file_current (fileName, fam_mem,
  				 &energy.readback, &energy.E) ) ;
}


int	read_magnets_from_field ()

{
  int i, res, warning ;
  int res1 = 0;
 
  warning = 0 ; 
  for (i=0;i<N_family;i++)
  {
    res = get_single_current ( fa[i]->pAcc, &fa[i]->readback, &fa[i]->current ) ;
    /* set warning flag if one or more elements have different read/set */
    if ( -3 == res ) warning = -3 ; 
    res1 = res1 + res;
    //else if ( 0 > res ) return ( res ) ; 
  }

  for (i=0;i<N_septum;i++)
    if ( 0 > ( res = get_single_voltage ( sp[i]->pAcc, &sp[i]->readback, 
    					&sp[i]->voltage ) ) ) return ( res ) ; 

return 0; /* as test */
/*  if(res1 < 0) return (res1) ;
  return ( warning ) ; */
}


int	read_magnets_from_file ( fileName ) 

str32	fileName ;

{
  int 		i, err=0 ;

  for (i=0;i<N_family;i++)
  {
    err += read_current_from_file (fileName, fa[i]->pAcc, &(fa[i]->readback),
    			 &(fa[i]->current)) ;
    if ( i==0 ) if ( err==-1 ) return ( err ) ; /*return if file not found*/
  } /* loop over all families */

  for (i=0;i<N_septum;i++)
  {
    err += read_current_from_file (fileName, sp[i]->pAcc, &(sp[i]->readback),
    			 &(sp[i]->voltage)) ;
    if ( i==0 ) if ( err==-1 ) return ( err ) ; /*return if file not found*/
  } /* loop over all septa */
    
  return ( err ) ;
} /* read_magnets_from_file */


int	read_IDgaps_from_field () 

{
  int i, res , res1;
  RPC_list	list ;
  
  res = 0;
  res1 = 0;
  for (i=0;i<N_ID;i++)
  {
    if ( id[i]->pPar->id.type == EEW )
    {
      if ( 0 > (res=get_single_current(id[i]->pAcc[0], &(id[i]->readback), &(id[i]->gap))) )
      {  res1-- ;
        id[i]->closed=0;
      }
      else 
      {
        id[i]->closed=1;
      }
      if ( 0 > (res=get_single_current(id[i]->pAcc[1], &(id[i]->readback_h), &(id[i]->gap_h))) )
      {
        res1-- ;
        id[i]->closed=0;
      }
      else
      {
        id[i]->closed=1;
      }
    }
    if ( id[i]->pPar->id.type == modulator )
    {
      if ( 0 > (res=get_single_current(id[i]->pAcc[0], &(id[i]->readback), &(id[i]->gap))) )
        res1-- ;
      if (!res) id[i]->closed=1;
    }
    if ( (id[i]->pPar->id.type == REC) || (id[i]->pPar->id.type == hybrid))
    {
      if ( 0 > (res=get_single_IDgap(id[i]->pAcc[0], &(id[i]->readback), &(id[i]->gap))) )
      {
        res1-- ;
        id[i]->closed=0;
      }
      else
      {
       id[i]->closed=1;
       if (id[i]->gap > 100.0)
       {
         id[i]->gap = 100.0;
       }
      }
    }
    if ( id[i]->pPar->id.type == sazaki )
    {
    /* local variables for calling dedicated sazaki function */
    /* C.Scafuri 20 jun 2002 */
      double szgap=0.0;
      double sztaper=0.0;
      double szphase=0.0;
      if ( 0 > (res=get_sazaki_IDgataph(id[i]->pAcc[0], &szgap,&sztaper,&szphase)) )
      {
        res1-- ;
        id[i]->closed=0;
      }
      else
      {
      /*assign newly read values to old structure for proper handling */
        id[i]->gap=szgap;
	id[i]->phase=szphase;
	/* sztaper not yet used */
        if (id[i]->gap >100.0 )
        {
          id[i]->gap = 100.0;
        }
        if ( id[i]->gap < id[i]->pPar->id.g_low )
        {
          id[i]->gap=id[i]->pPar->id.g_low;
        }
        id[i]->closed=1;
      /*
      }
      if ( 0 > (res=get_single_IDphase(id[i]->pAcc[0], &(id[i]->phase))) )
      {
        res1-- ;
      }
      else
      {
      */
        id[i]->phase=id[i]->phase/1000.0;
        if (id[i]->phase < (id[i]->pPar->id.gh_low/1000.0) )
        {
          id[i]->phase = id[i]->pPar->id.gh_low/1000.0;
        }
        if (id[i]->phase > (id[i]->pPar->id.gh_high/1000.0) )
        {
          id[i]->phase = id[i]->pPar->id.gh_high/1000.0;
        }
      }
    } /*end sazaki*/
    
   /*  printf("\n %s_%s : %d \t %d\n",id[i]->pAcc[0]->family, id[i]->pAcc[0]->member,id[i]->pPar->id.type,id[i]->closed); */
  }
  return ( res1 ) ;
}


int	read_IDgaps_from_file ( fileName ) 

str32	fileName ;

{
  int 		i, err=0, res ;

  for (i=0;i<N_ID;i++)
  {
    if ( id[i]->pPar->id.type == EEW)
    {
      err += (res=read_current_from_file(fileName,id[i]->pAcc[0],&(id[i]->readback),
    			&(id[i]->gap)) );
      if ( i==0 ) if ( err==-1 ) return ( err ) ; /*return if file not found*/
     /* consider an ID which was not found to be open, i.e. no effect on beam */
      if ( res < 0 ) id[i]->closed = 0 ;
      else id[i]->closed = 1 ;
      err += (res=read_current_from_file(fileName,id[i]->pAcc[1],&(id[i]->readback_h),
    			&(id[i]->gap_h)) );
      if ( i==0 ) if ( err==-1 ) return ( err ) ; /*return if file not found*/
     /* consider an ID which was not found to be open, i.e. no effect on beam */
      if ( res < 0 ) id[i]->closed = 0 ;
      else id[i]->closed = 1 ;
    }
    else
    {
      err += (res=read_current_from_file(fileName,id[i]->pAcc[0],&(id[i]->phase),
    			&(id[i]->gap)) );
      if ( i==0 ) if ( err==-1 ) return ( err ) ; /*return if file not found*/
     /* consider an ID which was not found to be open, i.e. no effect on beam */
      if ( res < 0 ) id[i]->closed = 0 ;
      else id[i]->closed = 1 ;
    }
  } /* loop over all IDs */
    
  return ( err ) ;
} /* read_IDgaps_from_file */


int	read_correctors_from_field ()

{
  int i, res, plane, warning ;
  
  warning = 0 ;
  for (i=0;i<N_corrector;i++) 
  {
    plane = (co[i]->plane==lumped) ? horizontal : co[i]->plane ;
    do {
      res = get_single_current ( co[i]->pAcc[plane],
		&co[i]->readback[plane], &co[i]->current[plane] ) ; 
      /* set warning flag if one or more elements have different read/set */
      if ( -3 == res ) warning = -3 ; 
      else if ( 0 > res ) return ( res ) ; 
    } while ( ++plane == vertical && co[i]->plane == lumped ) ;

  } /* loop over all correctors */

  for (i=0;i<N_chicane;i++) 
  {
    plane = (chi[i]->plane==lumped) ? horizontal : chi[i]->plane ;
    do {
      res = get_single_current ( chi[i]->pAcc[plane],
		&chi[i]->readback[plane], &chi[i]->current[plane] ) ; 
      /* set warning flag if one or more elements have different read/set */
      if ( -3 == res ) warning = -3 ; 
      else if ( 0 > res ) return ( res ) ; 
    } while ( ++plane == vertical && chi[i]->plane == lumped ) ;

  } /* loop over all chicanes */

  return ( warning ) ;
}


int	read_correctors_from_file ( fileName ) 

str32	fileName ;

{
  double	current ;
  int 		i, err=0 ;
  int		plane ;

  for (i=0;i<N_corrector;i++)
  {
    /* set plane for loop if lumped, otherwise copy state of corrector */
    plane = (co[i]->plane==lumped) ? horizontal : co[i]->plane ;
    do {
      err += read_current_from_file(fileName, co[i]->pAcc[plane], 
                       &(co[i]->readback[plane]), &(co[i]->current[plane]));
      if ( i==0 ) if ( err==-1 ) return(-1); /*return if file not found*/
    } while ( ++plane == vertical && co[i]->plane == lumped ) ;
    /* complicated condition above ensures that loop is performed twice if
       corrector is lumped and once otherwise */

  } /* loop over all correctors */

  for (i=0;i<N_chicane;i++)
  {
    /* set plane for loop if lumped, otherwise copy state of corrector */
    plane = (chi[i]->plane==lumped) ? horizontal : chi[i]->plane ;
    do {
      err += read_current_from_file(fileName, chi[i]->pAcc[plane], 
                       &(chi[i]->readback[plane]), &(chi[i]->current[plane]));
      if ( i==0 ) if ( err==-1 ) return(-1); /*return if file not found*/
    } while ( ++plane == vertical && chi[i]->plane == lumped ) ;
    /* complicated condition above ensures that loop is performed twice if
       chicane is lumped and once otherwise */

  } /* loop over all chicanes */

  return ( err ) ;
} /* read_correctors_from_file */


int	read_kickers_from_field ()

{
  int i, res ;
  
  for (i=0;i<N_kicker;i++)
    if ( 0 > ( res = get_single_voltage ( ki[i]->pAcc, &ki[i]->readback, 
      					&ki[i]->voltage ) ) ) return ( res ) ; 

  return ( 0 ) ;
}


int	read_kickers_from_file ( fileName ) 

str32	fileName ;

{
  int 		i, err=0 ;

  for (i=0;i<N_kicker;i++)
  {
    err += read_current_from_file (fileName, ki[i]->pAcc, &(ki[i]->readback), 
    				&(ki[i]->voltage)) ;
    if ( i==0 ) if ( err==-1 ) return ( err ) ; /*return if file not found*/
  } 
    
  return ( err ) ;
} 


int	read_FS_TL_from_field ( N_OK ) 

int	*N_OK ;

{
  int i, res ;
  
  for (i=0;i<N_FS;i++)
    if ( 0 > ( res = get_single_FS ( fs[i], 
    	&fs[i]->beam_position[horizontal], &fs[i]->beam_position[vertical], 
    	&fs[i]->rms[horizontal],  &fs[i]->rms[vertical] ) ) )
    {
      *N_OK = i ;
      return ( res ) ; 
    }
  *N_OK = N_FS ;
  return ( 0 ) ;
}


int	read_BPM_from_field (where) 

int where ;

{
  if ( where == SR ) 
  {
    printf ( "#please use \"read_BPM_SR_from_field()\" instead of \"read_BPM_from_field(SR)\"\n");
    return ( read_BPM_SR_from_field () ) ;
  }
  else
  {
    printf ( "#read_BPM_from_field() works only with SR. exit program\n" ) ;
    exit ( 0 ) ;
  }
}


int	read_BPM_SR_from_field () 

{
  RPC_list	list ;
  double	x[AL_po], xrms[AL_po], y[AL_po], yrms[AL_po], s[AL_po] ;
  int		errn[AL_po] ;
  int		res ;
  int		first, i, iLonza, plane ;
  
  /* set input parameter list to do_rpc */
  list.bpm.x = x ;
  list.bpm.xrms = xrms ;
  list.bpm.y = y ;
  list.bpm.yrms = yrms ;
  list.bpm.s = s ;
  list.bpm.errn = errn ;

  if ( 0 <= ( res=do_rpc(pAccBPM_SR->pHost,service_bpm,"all",&list) ) )
  {
    /* find first BPM in the structure which is in section 1 ! */
    for ( first=0; po[first]->pGen->section != 1; first++ ) ;
    iLonza = 0 ;
    for ( i=first; i<N_BPM+first; i++ ) 
    {
      po[i%N_BPM]->beam_position[horizontal] = x[iLonza] ;
      po[i%N_BPM]->beam_position[vertical] = y[iLonza] ;
      po[i%N_BPM]->rms[horizontal] = xrms[iLonza] ;
      po[i%N_BPM]->rms[vertical] = yrms[iLonza] ;
      /* translate error numbers into hls conditions */
      switch ( errn[iLonza] - BPM_ERROR_OFFSET )
      {
        case BPM_OK:         po[i%N_BPM]->condition = GOOD ; break ;
        case BPM_SIGNAL_LOW: po[i%N_BPM]->condition = BAD  ; break ;
        case BPM_NOT_READY:  po[i%N_BPM]->condition = BAD  ; break ;
        default:             po[i%N_BPM]->condition = DEAD ;
      }
      iLonza++ ;
    }
  }  /* end of if ( 0 <= res ) */
     
  return ( res ) ;
} /* read_BPM_SR_from_field */


int	read_BPM_SR_from_file ( fileName ) 

str32	fileName ;

{
  int 		i, err=0 ;

  for (i=0;i<N_BPM;i++)
  {
    err += read_machine_file_BPM (fileName, po[i]->pAcc, 
    &(po[i]->beam_position[horizontal]), &(po[i]->beam_position[vertical]),
    &(po[i]->rms[horizontal]), &(po[i]->rms[vertical]) ) ;

    if ( i==0 ) if ( err==-1 ) return ( err ) ; /*return if file not found*/
  } 
    
  return ( err ) ;
} 



int	set_magnets_fast (t0)
int	*t0 ;
{
  int i, err=0, res, resmax = 0 ;
  
  for ( i=0; i<N_family; i++ )
  {
    if ( 0 > (res = set_current_fast(fa[i]->pAcc,fa[i]->current)) ) err += res;
    if ( resmax < res ) resmax = res ;
  }
  *t0 = resmax / (int) 1000 + 1 ;
  printf ( "$time to wait is %i seconds\n", *t0 ) ;

  return ( err ) ;
}


int	set_correctors_fast ( plane, t0 )

const int	plane ;
int		*t0 ;

{
  int i, err=0, res, resmax = 0 ;
  
  for ( i=0; i<N_corrector; i++ )
    if ( co[i]->plane == plane || co[i]->plane == lumped )
      {
        if ( 0 > ( res = set_current_fast(co[i]->pAcc[plane],
        					co[i]->current[plane]) ) )
 	  err += res ;
        if ( resmax < res ) resmax = res ;
      }

  /* data in milliseconds is assumed */
  if ( resmax != 0 )
  {
    *t0 = resmax / (int) 1000 + 1 ;
    printf ( "$time to wait is %i seconds\n", *t0 ) ;
  }
  else *t0 = 0 ;

  return ( err ) ;
} /* set_correctors_fast */


int	set_correctors_fastRN ( plane, t0 )

const int	plane ;
int		*t0 ;

{
  int i, err=0, res, resmax = 0 ;
  
  for ( i=0; i<N_corrector; i++ )
    if ( co[i]->plane == plane || co[i]->plane == lumped )
      if ( co[i]->condition[plane] == APPLY )
      {
        if ( 0 <= ( res = set_current_fast(co[i]->pAcc[plane],
        					co[i]->current[plane]) ) )
        	co[i]->condition[plane] = GOOD ;
        else err += res ;
        if ( resmax < res ) resmax = res ;
      }

  /* data in milliseconds is assumed */
  if ( resmax != 0 )
  {
    *t0 = resmax / (int) 1000 + 1 ;
    printf ( "$time to wait is %i seconds\n", *t0 ) ;
  }
  else 
  {
    *t0 = 0 ;
    printf ( "No corrector had condition APPLY\n", *t0 ) ;
  }

  return ( err ) ;
} /* set_correctors_fastRN */


int	set_correctors_slowRN ( plane, t0 )

const int	plane ;
int		*t0 ;

{
  int i, err=0, res ;
  double old_current, dummy, diff, diff_max = 0.0 ;
  
  printf ("the command set_correctors_slow is not supported, sorry\n" ) ;
  exit ( 0 ) ;


  for ( i=0; i<N_corrector; i++ )
    if ( co[i]->plane == plane || co[i]->plane == lumped )
      if ( co[i]->condition[plane] == APPLY )
      {
        if ( 0 <= ( res = get_single_current(co[i]->pAcc[plane],
        			&old_current, &dummy) ) ) 
        {
          if ( 0 <= ( res = set_current_slow(co[i]->pAcc[plane],
        					co[i]->current[plane]) ) )
          {
            /* here I assume that the ramp is not started if an RPC error
               is received !!! */
            co[i]->condition[plane] = GOOD ;
            diff = fabs ( co[i]->current[plane] - old_current ) ;
            if ( diff > diff_max ) diff_max = diff ;
          }
        }
        else err += res ;
      }

  /* AMPS_PER_SEC is valid only for correctors !!! */
  *t0 = (int)(diff_max / AMPS_PER_SEC) + 1 ;
  printf ( "$time to wait is %i seconds\n", *t0 ) ;

  return ( err ) ;
}


/*========================================================================*/
/* conversion routines                                                    */
/*========================================================================*/

int field_to_length (ptr, value, strength) 

calibration_record	*ptr ;
double			value ;
double			*strength ;

{
  int i ;

  if (ptr->Nc<1) { *strength=0.0; return ( 0 ) ; }  
  if (ptr->Nc>AL_calibration) { *strength=0.0; return ( -1 ) ; }  
  if ( energy.E == 0 ) { *strength=0.0; return ( -2 ) ; } 
   
  *strength = ptr->c[ptr->Nc-1] ;
  for ( i=ptr->Nc-2;i>=0;i-- ) *strength = *strength * fabs(value) + ptr->c[i] ;

  return ( 0 ) ;
} /* field_to_length */

int field_to_strength (ptr, value, strength) 

calibration_record	*ptr ;
double			value ;
double			*strength ;

{
  int i ;

  if (ptr->Nc<1) 
  { 
  *strength=0.0;
  return ( -1 ) ; 
  }  
  if (ptr->Nc>AL_calibration) 
  { 
  *strength=0.0; 
  return ( -1 ) ; 
  }  
  if ( energy.E <= 0 ) 
  { 
  *strength=0.0; 
  /* return ( -2 ) ;  */
  return (0) ;
  } 
   
  *strength = ptr->c[ptr->Nc-1] ;
  for ( i=ptr->Nc-2;i>=0;i-- )*strength = *strength * fabs(value) + ptr->c[i] ;

  *strength *= 0.29979245 / energy.E ;
  if ( value < 0.0 ) *strength = - *strength ;

  return ( 0 ) ;
} /* field_to_strength */


int strength_to_field (ptr, value, field) 

calibration_record	*ptr ;
double			value ;
double			*field ;

{
  int i ;
  
  if (ptr->Ncr<1) { *field=0.0; return ( 0 ); }
  if (ptr->Ncr>AL_calibration) { *field=0.0; return (-1); }
  if ( energy.E == 0 ) { *field=0.0; return ( -2 ) ; } 
   
  value /=  0.29979245 / energy.E ;
  *field = ptr->cr[ptr->Ncr-1] ;
  for ( i=ptr->Ncr-2;i>=0;i-- ) *field = *field * value + ptr->cr[i] ;

  return ( 0 ) ;
} /* strength_to_field */


/* the following routine is by L. Tosi */
int IDgap_to_kay(pID, gap, phase, kx, ky, kx1, ky1 )
ID_record *pID;
double gap, phase;
double *kx, *ky, *kx1, *ky1;
{
  int ig, ip, igh, igl, ipl, iph;
  int i, j, k;
  int ifoundg, ifoundp;
  double gapl, gaph;
  double phasel, phaseh;
  double low, high;
  double kxlow, kxhigh;
  double kx1low, kx1high;
  double kylow, kyhigh;
  double ky1low, ky1high;
  
/*  printf("id  gap, phase:  %s  %lf  %lf\n",pID->pGen->name, gap,phase);*/
  
  if ( (gap > pID->pKal->gap[pID->pKal->Nc-1]) || (phase < pID->pKal->phase[0])
        || (phase > pID->pKal->phase[pID->pKal->Nc-1]) )
  {
    *kx=0.0;
    *ky=0.0;
    *kx1=0.0;
    *ky1=0.0;
    return(0);
  }
  ifoundg=ifoundp=0;
  for (i=0; i< pID->pKal->Nc; i++)
  {
    if (pID->pKal->gap[i] == gap )
    {
      ifoundg=1;
      igl=i;
      break;
    }
  }
  if ( ifoundg )
  {
    for (i=igl; i<pID->pKal->Nc; i++)
    {
      if (pID->pKal->phase[i] == phase)
      {
        ifoundp=1;
        ipl=i;
        break;
      }
    }
    if ( ifoundp )
    {
      *kx1=pID->pKal->kx1[ipl];
      *kx=pID->pKal->kx[ipl];
      *ky1=pID->pKal->ky1[ipl];
      *ky=pID->pKal->ky[ipl];
      return(0);
    }
    if ( !ifoundp )
    {
      for( i=igl; i<(pID->pKal->Nc-1); i++)
      {
        phasel=pID->pKal->phase[i];
        phaseh=pID->pKal->phase[i+1];
        if ( (phasel <= phase) && (phaseh > phase) )
        {
          iph=i+1;
          ipl=i;
          break;
        }
      }
      phasel=pID->pKal->phase[ipl];
      phaseh=pID->pKal->phase[iph];

      low=pID->pKal->kx[ipl];
      high=pID->pKal->kx[iph];
      *kx=((high-low)/(phaseh-phasel))*(phase-phasel)+low;

      low=pID->pKal->ky[ipl];
      high=pID->pKal->ky[iph];
      *ky=((high-low)/(phaseh-phasel))*(phase-phasel)+low;

      low=pID->pKal->kx1[ipl];
      high=pID->pKal->kx1[iph];
      *kx1=((high-low)/(phaseh-phasel))*(phase-phasel)+low;

      low=pID->pKal->ky1[ipl];
      high=pID->pKal->ky1[iph];
      *ky1=((high-low)/(phaseh-phasel))*(phase-phasel)+low;
      return(0);
    }
  }
  if ( !ifoundg )
  {
    for (ig=0; ig<(pID->pKal->Nc-1); ig++)
    {
      gapl=pID->pKal->gap[ig];
      gaph=pID->pKal->gap[ig+1];
      if ( (gapl <= gap) && (gaph > gap) )
      {
        igh=ig+1;
        break;
      }
    }
    igl=0;
    for (i=igh; i>-1; i--)
    {
      if (pID->pKal->gap[i] < gapl )
      {
        igl=i+1;
        break;
      }
    }
    for (i=igl; i<igh; i++)
    {
      if (pID->pKal->phase[i] == phase )
      {
        ifoundp=1;
        ipl=i;
        iph=igh+(ipl-igl);
        break;
      }
    }
    if ( ifoundp )
    {
      phasel=pID->pKal->gap[ipl];
      phaseh=pID->pKal->gap[iph];


      low=pID->pKal->kx[ipl];
      high=pID->pKal->kx[iph];
      *kx=((high-low)/(phaseh-phasel))*(gap-phasel)+low;

      low=pID->pKal->ky[ipl];
      high=pID->pKal->ky[iph];
      *ky=((high-low)/(phaseh-phasel))*(gap-phasel)+low;

      low=pID->pKal->kx1[ipl];
      high=pID->pKal->kx1[iph];
      *kx1=((high-low)/(phaseh-phasel))*(gap-phasel)+low;

      low=pID->pKal->ky1[ipl];
      high=pID->pKal->ky1[iph];
      *ky1=((high-low)/(phaseh-phasel))*(gap-phasel)+low;

      return(0);
      
    }
    if (!ifoundp)
    {
      for(i=igl; i<(igh-1); i++)
      {
        if ( (pID->pKal->phase[i]<phase) && (pID->pKal->phase[i+1] > phase))
        {
          ipl=i;
          iph=i+1;
          break;
        }
      }
      phasel=pID->pKal->phase[ipl];
      phaseh=pID->pKal->phase[iph];

      low=pID->pKal->kx[ipl];
      high=pID->pKal->kx[iph];
      kxlow=((high-low)/(phaseh-phasel))*(phase-phasel)+low;

      low=pID->pKal->ky[ipl];
      high=pID->pKal->ky[iph];
      kylow=((high-low)/(phaseh-phasel))*(phase-phasel)+low;

      low=pID->pKal->kx1[ipl];
      high=pID->pKal->kx1[iph];
      kx1low=((high-low)/(phaseh-phasel))*(phase-phasel)+low;

      low=pID->pKal->ky1[ipl];
      high=pID->pKal->ky1[iph];
      ky1low=((high-low)/(phaseh-phasel))*(phase-phasel)+low;
      
      for(i=igh; i<(pID->pKal->Nc-1); i++)
      {
        if ( (pID->pKal->phase[i]<phase) && (pID->pKal->phase[i+1] > phase) )
        {
          ipl=i;
          iph=i+1;
          break;
        }
      }
 
      phasel=pID->pKal->phase[ipl];
      phaseh=pID->pKal->phase[iph];

      low=pID->pKal->kx[ipl];
      high=pID->pKal->kx[iph];
      kxhigh=((high-low)/(phaseh-phasel))*(phase-phasel)+low;

      low=pID->pKal->ky[ipl];
      high=pID->pKal->ky[iph];
      kyhigh=((high-low)/(phaseh-phasel))*(phase-phasel)+low;

      low=pID->pKal->kx1[ipl];
      high=pID->pKal->kx1[iph];
      kx1high=((high-low)/(phaseh-phasel))*(phase-phasel)+low;

      low=pID->pKal->ky1[ipl];
      high=pID->pKal->ky1[iph];
      ky1high=((high-low)/(phaseh-phasel))*(phase-phasel)+low;
/*******/      
      phasel=pID->pKal->gap[igl];
      phaseh=pID->pKal->gap[igh];

      low=kxlow;
      high=kxhigh;
      *kx=((high-low)/(phaseh-phasel))*(gap-phasel)+low;

      low=kylow;
      high=kyhigh;
      *ky=((high-low)/(phaseh-phasel))*(gap-phasel)+low;

      low=kx1low;
      high=kx1high;
      *kx1=((high-low)/(phaseh-phasel))*(gap-phasel)+low;

      low=ky1low;
      high=ky1high;
      *ky1=((high-low)/(phaseh-phasel))*(gap-phasel)+low;
      
      return(0);
      
    }
    
  }
  
}

int IDgap_to_strength(pID, gap, curv, curvh )
ID_record *pID;
double gap, *curv, *curvh;
{
  int i;
  double bh, pi, gap1;
  double h[5], he;
  
  if (energy.E == 0.0 ) return(-1);
  if ( gap == 0.0 || pID->pPar->id.period==0.0 ) return(-2);
  
  pi = acos(-1.0);
  gap1 = gap/1000.0;
  h[0] = 1.0;
  he = gap1/pID->pPar->id.period;
  bh  = 0.0;

  for ( i=1; i<pID->pAcc[0]->pCal1->Nc; i++)
    h[i] = he*h[i-1];
  
  for ( i=1; i<pID->pAcc[0]->pCal1->Nc; i++)
    bh += pID->pAcc[0]->pCal1->c[i]*h[i];
  
  bh = pID->pAcc[0]->pCal1->c[0] * exp(-bh)/10000.0;
  
  *curv = bh * 0.29979245/energy.E;
  if (pID->pPar->id.type == sazaki)
  {
  *curv = cos(pi*pID->phase/pID->pPar->id.period)*bh * 0.29979245/energy.E;
    bh  = 0.0;
    h[0] = 1.0;

    for ( i=1; i<pID->pAcc[0]->pCal2->Nc; i++)
      h[i] = he*h[i-1];
  
    for ( i=1; i<pID->pAcc[0]->pCal2->Nc; i++)
      bh += pID->pAcc[0]->pCal2->c[i]*h[i];
  
    bh = pID->pAcc[0]->pCal2->c[0] * exp(-bh)/10000.0;
  
    *curvh = sin(pi*pID->phase/pID->pPar->id.period)*bh * 0.29979245/energy.E;
    
  }
    
      return(0);

} /* IDgap_to_strength */

int f2l_bends () 

{
  int		i,j, err=0 ;
  magnet_record	*ptr ;
  
  for (i=0;i<N_bend;i++) 
  {
    ptr = be[i] ;
    err += field_to_length ( ptr->pCalL, ptr->pFam->current, &ptr->magnetic_length ) ;
    if ( !err )
    {
      for ( j=ptr->back_index; j>0; j-- )
      {
        if ( ge[j]->type == drift_type )
        {
          dr[ge[j]->index]->length+=
          (ptr->pPar->ps.length - ptr->magnetic_length)/2.0;
          break;
        }
      }
      for ( j=ptr->back_index; j<N_general; j++ )
      {
        if ( ge[j]->type == drift_type )
        {
          dr[ge[j]->index]->length+=(ptr->pPar->ps.length - ptr->magnetic_length)/2.0;
          break;
        }
      }
      
    }
  }
  if ( !err )
  {
    for (i=0;i<N_bend;i++) 
    {
      be[i]->pPar->ps.length = be[i]->magnetic_length;
    }
    
  }
  return ( err ) ;
} /* f2l_bends */

int f2s_magnets () 

{
  int		i,j, err=0 ;
  magnet_record	*ptr ;
  
  for (i=0;i<N_family;i++) for (j=0;j<fa[i]->N_member;j++) 
  {
    ptr = fa[i]->pMember[j] ;
    err += field_to_strength ( ptr->pCalB, fa[i]->current, &ptr->strength ) ;
  }

  for (i=0;i<N_septum;i++)
    err += field_to_strength ( sp[i]->pAcc->pCal1, sp[i]->voltage, 
    							&sp[i]->curvature ) ;
  return ( err ) ;
} /* f2s_magnets */


int s2f_magnets () 

{
  int			i,j, err=0 ;
  magnet_record		*ptr ;
  calibration_record	*pCal ;
  
  for (i=0;i<N_family;i++) 
  {
    /* take only first member of family */
    ptr = fa[i]->pMember[0] ;
    if ( ptr->pGen->type == bend_type ) 
    {
      /* this makes sense only for bends, where energy defines current */
      pCal = fa[i]->pAcc->pCal1 ;
      if (pCal->Ncr<1 || pCal->Ncr>AL_calibration) err+=-1 ;
      else
      {
        fa[i]->current = pCal->cr[pCal->Ncr-1] ;
        for (j=pCal->Ncr-2;j>=0;j--)
          fa[i]->current = fa[i]->current * energy.E + pCal->cr[j] ;
      }
    }
    else 
      err += strength_to_field ( ptr->pCalB, ptr->strength, &fa[i]->current ) ;
  }

  /* for septum, voltages is obtained from the energy directly ! */
  for (i=0;i<N_septum;i++)
  {
    pCal = sp[i]->pAcc->pCal1 ;
    if (pCal->Ncr<1 || pCal->Ncr>AL_calibration) err+=-1 ;
    else
    {
      sp[i]->voltage = pCal->cr[pCal->Ncr-1] ;
      for (j=pCal->Ncr-2;j>=0;j--)
        sp[i]->voltage = sp[i]->voltage * energy.E + pCal->cr[j] ;
    }
  }

  return ( err ) ;
} /* s2f_magnets */


int f2s_IDs () 

{
  int		i, err=0 ;
  ID_record	*ptr ;
  
  for (i=0;i<N_ID;i++) 
  {
    ptr = id[i] ;
    if ( ptr->pPar->id.type == EEW  )
    {
      err += field_to_strength(ptr->pAcc[0]->pCal1, ptr->gap, &ptr->curvature);
      err += field_to_strength(ptr->pAcc[1]->pCal1, ptr->gap_h, &ptr->curvature_h);
    }
    if ( ptr->pPar->id.type == modulator  )
    {
      err += field_to_strength(ptr->pAcc[0]->pCal1, ptr->gap, &ptr->curvature);
    }
    if ( (ptr->pPar->id.type == REC) || (ptr->pPar->id.type == hybrid) || (ptr->pPar->id.type == sazaki) )
    {
        err += IDgap_to_strength (ptr, ptr->gap, &ptr->curvature, &ptr->curvature_h) ;
    }
    if ( ptr->pPar->id.type == sazaki )
    {
        err += IDgap_to_kay (ptr, ptr->gap, ptr->phase*1000, &ptr->k_x, &ptr->k_y, &ptr->kh_x, &ptr->kh_y) ;
    }
  }
  return ( err ) ;
} /* f2s_IDs */


int f2s_correctors () 

{
  int		i,j, err=0 ;
  int		plane ;
  
  for (i=0;i<N_corrector;i++)  
  {
    plane = (co[i]->plane==lumped) ? horizontal : co[i]->plane ;
    do err += field_to_strength (co[i]->pAcc[plane]->pCal1,
      			co[i]->current[plane], &(co[i]->theta[plane]) ) ;
    while ( ++plane == vertical && co[i]->plane == lumped ) ;
  }

  for (i=0;i<N_chicane;i++)  
  {
    plane = (chi[i]->plane==lumped) ? horizontal : chi[i]->plane ;
    do err += field_to_strength (chi[i]->pAcc[plane]->pCal1,
      			chi[i]->current[plane], &(chi[i]->theta[plane]) ) ;
    while ( ++plane == vertical && chi[i]->plane == lumped ) ;
  }

  for (i=0;i<N_kicker;i++)  
  {
    err += field_to_strength (ki[i]->pAcc->pCal1,
      			ki[i]->voltage, &(ki[i]->theta) ) ;
  }

  return ( err ) ;
} /* f2s_correctors */


int s2f_correctors () 

{
  int		i,j, err=0 ;
  int		plane ;
  
  for (i=0;i<N_corrector;i++)  
  {
    plane = (co[i]->plane==lumped) ? horizontal : co[i]->plane ;
    do err += strength_to_field (co[i]->pAcc[plane]->pCal1,
      			co[i]->theta[plane], &co[i]->current[plane] ) ;
    while ( ++plane == vertical && co[i]->plane == lumped ) ;
  }

  for (i=0;i<N_chicane;i++)  
  {
    plane = (chi[i]->plane==lumped) ? horizontal : chi[i]->plane ;
    do err += strength_to_field (chi[i]->pAcc[plane]->pCal1,
      			chi[i]->theta[plane], &chi[i]->current[plane] ) ;
    while ( ++plane == vertical && chi[i]->plane == lumped ) ;
  }

  for (i=0;i<N_kicker;i++)  
  {
    err += strength_to_field (ki[i]->pAcc->pCal1,
      			ki[i]->theta, &(ki[i]->voltage) ) ;
  }

  return ( err ) ;
} /* s2f_correctors */


/*========================================================================*/
/* lifetime routines                                                      */
/*========================================================================*/

#define DCCT_SIGMA 2.E-3  
#define DCCT_SIGMA2 DCCT_SIGMA * DCCT_SIGMA  
#define DCCT_THRESHOLD 0.05
#define MIN_STEP 10
/* the lowest rate that is reasonable at current DCCT error */
#define LOWEST_RATE DCCT_SIGMA / 600. / sqrt ( 600. ) 


int	get_DCCT_array ( buffer ) 
double buffer[] ;

{
  RPC_list	list ;
  realArray	data ;
  int		i, err ;

  /* set input parameter list to do_rpc */
  /* the access pointer is taken from the global variables */
  list.lpc.family = pAccDCCT->family ; 
  list.lpc.member = pAccDCCT->member ; 

  list.lpc.datum = &data ;
  list.lpc.action = "CBUFF" ;
  if ( 0 <= (err=do_rpc(pAccDCCT->pHost,service_lpc,"RAR",&list)) )   
  {
    for (i=3;i<data.length;i++) buffer[i-3] = ( data.sequence[i] ) ;
  }
  
  if ( data.length != 0 ) 
  {
    free ( data.sequence ) ;
    data.length = 0 ;
  }
  
  return ( err ) ;

} /* get_DCCT_array */


void calc_lifetime ( p, I, tau, rate, err, t )
double	p ;
double	*I ;
double	*tau ;
double	*rate ;
double	*err ;
int	*t ;

{
  double	buffer[600] ;
  double	limit ;
  double	sum_y, sum_iy;
  double	A, B, Anew, Bnew ;
  double	rms, rmsnew, fit ;
  int		i, j ;
  double	iii ;
  

  if ( 0 > get_DCCT_array ( buffer ) )
  {
    *I = -1.0 ;
    *tau = 0.0 ;
    *rate = 0.0 ;
    *err = 0.0 ;
    *t = 0 ;
    return ;
  }

  *I =  buffer[0]  ; 
  
  if ( *I < DCCT_THRESHOLD ) 
  {
    *I = 0.0 ;
    *tau = 0.0 ;
    *rate = 0.0 ;
    *err = 0.0 ;
    *t = 0 ;
    return ;
  }

   
  limit = p*p / 12.0 ;
  sum_y = buffer[0] ;
  sum_iy = 0.0 ;
  i = 0 ;

  while ( i < 596 )
  {
    i++ ;
    iii = ( double) ( i*(i+1)*(i+2) ) ;
    sum_y += buffer[i] ;
    sum_iy += buffer[i] * i ;
    Anew = (double)i * 6./iii* ( (2.*i+1.)/3.*sum_y - sum_iy ) ;
    Bnew = 12./iii* ( sum_iy - i/2.*sum_y ) ;

    rmsnew = 0.0 ;
    for ( j=0;j<=i;j++ ) 
      rmsnew += ( buffer[j] - (Anew+Bnew*j) ) * ( buffer[j] - (Anew+Bnew*j) ) ;
    rmsnew = rmsnew / (i+1) ;
    
   if ( i>MIN_STEP && ( rmsnew > DCCT_SIGMA2 ) )
    {
      i-- ;
      iii = ( double) ( i*(i+1)*(i+2) ) ;
      break ;
    }
    A = Anew ;
    B = Bnew ;
    rms = rmsnew ;
    if ( i>=MIN_STEP && rms < limit * B*B * iii ) break ;
  } /* while */
  
  if ( fabs ( B ) < LOWEST_RATE )
  {
    *tau = 100. * 3600. ;
    *rate = 0.0 ;
    *err = 1.0 ;
  }
  else
  {
    *tau = A / B ;
    *rate = -B ;
    *err = sqrt ( rms * 12. / iii ) / fabs ( B )  ;  
  }

  *t = i ;
  
}



/*========================================================================*/
/* orbit correction utility routines                                      */
/*========================================================================*/

int	apply_correction ( plane )

const int 	plane ;

{
  int err, t0 ;
  
  if ( 0 > ( err = s2f_correctors() ) ) return ( 1 ) ;
  err = set_correctors_fastRN ( plane, &t0 ) ;
  sleep ( t0 ) ;
  
  return ( err ) ;
}


int	acquire_correction ()

{
  int err ;

  if ( 0 > ( err = read_correctors_from_field() ) ) return ( err ) ;
  if ( 0 > ( err = f2s_correctors() ) ) return ( 1 ) ;
  
  return ( 0 ) ;
}


/*========================================================================*/
/* utility routines for completely setting model in one command           */
/*========================================================================*/

int	init_model_from_field ( where )

const int	where ;

{
twiss_record	twiss ;
int		res, warning ;

  warning = 0 ;
  if ( 0 > read_energy_from_field(where) ) return ( -1 ) ;
  res = read_magnets_from_field() ;
    /* set warning flag if one or more elements have different read/set */
    if ( -3 == res ) warning = -5 ; 
    else if ( 0 > res ) return ( -2 ) ; 
  if ( 0 > read_IDgaps_from_field() ) return ( -2 ) ;
  if ( 0 > f2l_bends() ) return ( -3 ) ;
  if ( 0 > f2s_magnets() ) return ( -3 ) ;
  if ( 0 > f2s_IDs() ) return ( -3 ) ;
  if ( where == TL ) if ( 0 > read_TL_twiss(&twiss) ) return ( -4 ) ;
  if ( 0 > update_twiss_functions(where, &twiss) ) return ( -4 ) ;
  return ( warning ) ;
}


int	init_model_from_file ( fileName, where )

str32		fileName ;
const int	where ;

{
twiss_record	twiss ;

  if ( 0 > read_energy_from_file(fileName) ) return ( -1 ) ;
  if ( 0 > read_magnets_from_file(fileName) ) return ( -2 ) ;
  if ( 0 > read_IDgaps_from_file(fileName) ) return ( -2 ) ;
  if ( 0 > f2l_bends() ) return ( -3 ) ;
  if ( 0 > f2s_magnets() ) return ( -3 ) ;
  if ( 0 > f2s_IDs() ) return ( -3 ) ;
  if ( where == TL ) if ( 0 > read_TL_twiss(&twiss) ) return ( -4 ) ;
  if ( 0 > update_twiss_functions(where, &twiss) ) return ( -4 ) ;
  return ( 0 ) ;
}

/*==================================================================
  ********************** end of functions.c **********************
==================================================================*/
