/*==================================================================
                     	+------------------+
 			|   histogram.c    |
                     	+------------------+

 The following routines form the histogramming utility of the hls,
 which is meant for monitoring purposes.
 The general histogrammig routines are: init_histogram(), clear_histogram(),
 add_to_histogram() and write_histogram().
 Added are special purpose routines for PS histogramming: init_PShistogram(),
 clear_PS_histogram(), add_to_PShistogram().

 First public test version.
 Beta 0.1				(Mark Plesko 22.12.94)
 
 ==================================================================*/

#include <stdio.h>
#include <math.h>
#include <time.h>

#include "hlsdef.h"
#include "histogram.h"
#include "hlsrpc.h"
#include "maglist.h"

/* the following is needed for the PS histogramming routine */
/* the time to wait for a PS to stabilize */
#define T_STABLE  10.0

/***********************************************************************/
/**********          general histogramming routines                  ***/
/***********************************************************************/


int  init_histogram ( pH, Nel, start, step, Nbins )
histogram_record	**pH ;
int			Nel, start, step, Nbins ;

{
  histogram_record	*H ;
  int i,j ;

  /* set default values if Nbins is set to zero */
  if ( Nbins == 0 )
  {
    Nbins = 16 ;
    start = -8 * step ;
  }

  /* check validity of input parameters */
  if ( Nel <= 0 || Nbins < 0 ) 
  {
    printf ("Negative No. of elements or No. of bins not allowed !\n") ;
    return(-1) ;
  }

  /* allocate space for histogram record itself */
  if ( (H=(histogram_record*)malloc(sizeof(histogram_record))) == NULL )
  {
    printf ("hist allocation error: not enough memory space - aborting\n") ;
    return(-2) ;
  }

  H->max_elements = Nel ;
  H->threshold = start ;
  H->step = step ;
  H->Nbins = Nbins ; 

  /* allocate space for reference vector */
  if ( (H->reference=(double*)malloc(Nel*sizeof(double))) == NULL )
  {
    printf ("hist allocation error: not enough memory space - aborting\n") ;
    return(-2) ;
  }

  /* allocate space for scale vector */
  if ( (H->scale=(double*)malloc(Nel*sizeof(double))) == NULL )
  {
    printf ("hist allocation error: not enough memory space - aborting\n") ;
    return(-2) ;
  }

  /* allocating space for histograming variables */
  if ( (H->N=(int*)malloc(H->max_elements*sizeof(int))) == NULL )
  {
    printf ("hist allocation error: not enough memory space - aborting\n") ;
    return(-2) ;
  }
  if ( (H->min=(int*)malloc(H->max_elements*sizeof(int))) == NULL )
  {
    printf ("hist allocation error: not enough memory space - aborting\n") ;
    return(-2) ;
  }
  if ( (H->max=(int*)malloc(H->max_elements*sizeof(int))) == NULL )
  {
    printf ("hist allocation error: not enough memory space - aborting\n") ;
    return(-2) ;
  }
  if ( (H->hist=(int*)malloc(H->max_elements*Nbins*sizeof(int))) == NULL )
  {
    printf ("hist allocation error: not enough memory space - aborting\n") ;
    return(-2) ;
  }
  if ( (H->name=(str13*)malloc(H->max_elements*sizeof(str13))) == NULL )
  {
    printf ("hist allocation error: not enough memory space - aborting\n") ;
    return(-2) ;
  }
  
  /* initialize histogram record */
  for ( i=0;i<H->max_elements;i++ )
  {
    /* default: don't subtract reference values, don't scale */
    H->reference[i] = 0.0 ;
    H->scale[i] = 1.0 ;
    
    /* set all counters to zero */
    H->N[i] = 0 ;
    H->min[i] = 0 ;
    H->max[i] = 0 ;
    for (j=0;j<Nbins;j++) H->hist[Nbins*i+j] = 0 ;
  }

  /* pass H to outer world */
  *pH = H ;


  return ( 0 ) ;

} /* init_histogram */


void  clear_histogram ( H )
histogram_record	*H ;

{
  /* deallocate space */
  
  free ( H->reference ) ;
  free ( H->scale ) ;
  free ( H->N ) ;
  free ( H->min ) ;
  free ( H->max ) ;
  free ( H->hist ) ;
  
  free ( H ) ;

} /* clear_histogram */


void add_to_histogram ( H, i, datum )
histogram_record	*H ;
int			i ;
double			datum ;

{
  int x ;
  double difference ;
  
  /* first check whether elementNumber is a reasonable value */
  
  if ( i < 0 || i >= H->max_elements ) 
  {
    printf ("element number %i out of bounds: [0,%i)\n", i, H->max_elements ) ;
  }
  else
  {
    difference = ( datum - H->reference[i] ) / H->scale[i] ; 
                    
    /* calculate and increment bin in histogram */
    x = (int) floor ( ( difference - (double)H->threshold ) / (double)H->step );
    H->N[i]++ ;
    if ( x < 0  ) 
      { if ( H->min[i] < 100000 ) H->min[i]++ ; }
    else if ( x >= H->Nbins ) 
      { if ( H->max[i] < 100000 ) H->max[i]++ ; }
    else { if ( H->hist[H->Nbins*i+x] < 100000 ) H->hist[H->Nbins*i+x]++ ; }
    
  } /* element Number is OK */
    
} /* add_to_histogram */


int write_histogram ( H, filename ) 
histogram_record	*H ;
char			*filename ;

{
  FILE	*fp ;
  int i, l ;
  int index, plane ;
  int max_elements ;
  struct struct_fame	*names ;
  char	fullName[15] ;

  fp = fopen ( filename, "w" ) ;
  if ( fp == NULL ) 
      {  printf ( "\nfile %s can not be opened\n", filename ) ; return (-1) ; }


  for ( i=0;i<H->max_elements;i++ ) 
  {
    /* print title every 50 elements */
    if ( i % 50 == 0 )
    {
      fprintf ( fp, "%-10s %6s %5s", "NAME", "Ntot", "UFLOW" ) ;
      for (l=0;l<H->Nbins;l++) fprintf ( fp, " %5i", 
      				H->threshold+l*H->step ) ;
      fprintf ( fp, " %5s\n", "OFLOW" ) ;
      for (l=0;l<29+6*H->Nbins;l++) fprintf ( fp, "=" ) ;
      fprintf ( fp, "\n" ) ;
    }
    
    fprintf ( fp, "%-10s %6i", H->name[i], H->N[i] ) ;
    if ( H->min[i] > 0 ) fprintf ( fp, " %5i", H->min[i] ) ;
    else fprintf ( fp, "      " ) ;
    for (l=0;l<H->Nbins;l++) 
      if ( H->hist[H->Nbins*i+l] > 0 ) 
        fprintf ( fp, " %5i", H->hist[H->Nbins*i+l] ) ;
      else fprintf ( fp, "      " ) ;
    if ( H->max[i] > 0 ) fprintf ( fp, " %5i\n", H->max[i] ) ;
    else fprintf ( fp, "      \n" ) ;

  } /* end for all elements */
  
  fclose ( fp ) ;

} /* write_histogram */


/***********************************************************************/
/**********              PS histogramming routines                   ***/
/***********************************************************************/


int	init_PShistogram ( pPS, mode, stepsize ) 
PSreference_record	**pPS ;
char			*mode ;
int			stepsize ;

{
  int	i ;
  PSreference_record	*PS ;
  histogram_record *H ;    /* new variable, just for less typing */

  /* set variables max_elements and names according to the mode */
  
  /* allocate space for PS reference record itself */
  if ( (PS=(PSreference_record*)malloc(sizeof(PSreference_record))) == NULL )
  {
    printf ("PS reference allocation error: not enough memory space - aborting\n") ;
    return(-2) ;
  }

  if ( !strcmp("bqt_r",mode) ) 
  { 
    if ( 0 > init_histogram ( &H, 23, 0, stepsize, 0 ) ) return ( -1 ) ;
    
    for (i=0;i<H->max_elements;i++) 
    {
      H->scale[i] = 125. * 1.0e-6 ;
      strcpy ( H->name[i], bq_t[i].fam ) ;
      strcat ( H->name[i], "_" ) ;
      strcat ( H->name[i], bq_t[i].mem ) ;
    }
    H->scale[0] = 1080. * 1.0e-6 ;
    H->scale[1] = 1080. * 1.0e-6 ;
  }
  else if(!strcmp("ct_r",mode) ) 
  {
    init_histogram ( &H, 34, 0, stepsize, 0 ) ;
    
    for (i=0;i<H->max_elements;i++) 
    {
      H->scale[i] = 27. * 1.0e-6 ;
      strcpy ( H->name[i], c_t[i].fam ) ;
      strcat ( H->name[i], "_" ) ;
      strcat ( H->name[i], c_t[i].mem ) ;
    }
  }
  else if(!strcmp("bqss_r",mode) ) 
  {
    init_histogram ( &H, 42, 0, stepsize, 0 ) ;
    
    for (i=0;i<H->max_elements;i++) 
    {
      H->scale[i] = 300. * 1.0e-6 ;
      strcpy ( H->name[i], bqs_s[i].fam ) ;
      strcat ( H->name[i], "_" ) ;
      strcat ( H->name[i], bqs_s[i].mem ) ;
    }
    H->scale[0] = 2000. * 1.0e-6 ;
  }
  else if(!strcmp("cs_r",mode) ) 
  { 
    init_histogram ( &H, 164, 0, stepsize, 0 ) ;
    
    for (i=0;i<H->max_elements;i++) 
    {
      H->scale[i] = 16. * 1.0e-6 ;
      strcpy ( H->name[i], c_s[i].fam ) ;
      strcat ( H->name[i], "_" ) ;
      strcat ( H->name[i], c_s[i].mem ) ;
    }
  }
  else return ( -2 ) ;

  /* put H where it belongs */
    PS->H = H ;

  /* now fill PS record */

  strcpy ( PS->mode, mode ) ;

  if ( (PS->lastChange=(time_t*)malloc(H->max_elements*sizeof(time_t)))==NULL )
  {
    printf ("PSref allocation error: not enough memory space - aborting\n") ;
    return(-1) ;
  }
  if ( (PS->Iset=(double*)malloc(H->max_elements*sizeof(double))) == NULL )
  {
    printf ("PSref allocation error: not enough memory space - aborting\n") ;
    return(-1) ;
  }

  /* initialize PSreference record */
  for ( i=0;i<H->max_elements;i++ )
  {
    /* set initial readback to impossible current (as guard) */
    H->reference[i] = 10000.0 ;
    PS->Iset[i] = 0.0 ;
    PS->lastChange[i] = 0 ;
  }

  *pPS = PS ;

  return ( 0 ) ;

} /* init_PShistogram */


void  clear_PShistogram ( PS )
PSreference_record	*PS ;

{
  /* deallocate space */

  clear_histogram ( PS->H ) ;
   
  free ( PS->lastChange ) ;
  free ( PS->Iset ) ;
  
  free ( PS ) ;

} /* clear_PShistogram */


void add_to_PShistogram ( PS, threshold, nseconds, curr, curset )
PSreference_record	*PS ;
int			threshold ;
time_t			nseconds ;
double			curr[], curset[];

{
  int i, err ;
  int x ;
  double difference ;
  histogram_record *H ;
   
  /* new variable, just for jess typing */
  H = PS->H ;

  for ( i=0;i<H->max_elements;i++) 
  {
    /* check if PS setting has been changed */
    if ( curset[i] != PS->Iset[i] )
    {
      /* don't write the warning the first time ! */
      if ( PS->lastChange[i] != 0 )
      {
        printf ( "\nWarning: %s setting has changed to %8.4lf at %s",
          H->name[i], curset[i], ctime (&nseconds) ) ;
        printf ( "Old setting was %8.4lf last changed at %s",
          PS->Iset[i], ctime (&PS->lastChange[i]) ) ;
      }

      PS->lastChange[i] = nseconds ;
      PS->Iset[i] = curset[i] ;
      H->reference[i] = 10000.0 ;
      continue ;
    }
    
    /* check if corrector setting is stable for T_STABLE seconds or more */
    if ( difftime ( nseconds, PS->lastChange[i] ) < T_STABLE ) continue ;

    /* if setting has changed, define new reference readback */
    if ( H->reference[i] > 9999.0 )
    {
      H->reference[i] = curr[i] ;
      continue ;
    }

    /* submit alarm if threshold exceeded */
    difference = ( curr[i] - H->reference[i] ) / H->scale[i] ; 
    if ( fabs (difference) > threshold )       
    {
      printf ( "\nAlarm: %s I - Iref is %5.0lf ppm at %s",
        H->name[i], difference, ctime (&nseconds) ) ;
      printf ( "Iset=%8.4lf I=%8.4lf, Iref=%8.4lf\n",
        PS->Iset[i], curr[i], H->reference[i] ) ;
    }       
    
    add_to_histogram ( H, i, curr[i] ) ;
    
  } /* loop over all elements */
    
} /* add_to_PShistogram */

