/*==================================================================
                     	+---------------+
 			|    input.c    |
                     	+---------------+

 The set of routines which read the ORACLE HLS databases and build
 up the HLS data bush. All global variables and functions within this
 file are static, with the exception of:

 void	input_open_files ()
 int	input_structure ( const int )
 void	input_special( const int, str13, access_record ** ) 
 void	input_close_files ()
 void	input_all ( str32 ) 
 void	cleanup_rpc ()
 
 First official version. All 0.x versions are outdated !
 ASCII input instead of ORACLE. Calibration, parameter and access
 file are searched sequentially for each record.
 Beta 1.0				(Mark Plesko 10.11.92)
 
 Changed ASCII file reading to accept 132 characters per line (define N80).
 Typing error while creating septum_record fixed.
 Beta 1.1		 		(M.P. 1.12.92)
 
 Default magnetic_length of magnet_records set to ps.length, because
 this is sufficient for most purposes.
 Changed reading of calibration file to contain only reference to
 AL_calibration and to comply with Fatimas new convention ( name +
 five constants + five reverse constants).
 Beta 1.2				(M.P. 11.12.92)
 
 The parameter to input_structure() has been changed to a string (str32)
 which represents the file name of the structure file. input_structure()
 now returns -1 if the file could not be opened and 0 if it was OK.
 Beta 1.3				(M.P. 18.12.92)
 
 read_calibration_file() was modified to read new "cal" file by F.I.,
 which contains two lines per record. A minor bug in the routine
 find_access_record() was fixed. Names of special elements changed
 to upper case. Edge elements get magnetic length from parent sector.
 Beta 1.4				(M.P. 14.1.93)

 reading of ASCII files changed. Now cal, par and acc file are read
 only once and stored in the heap (routines copy_RA_files and 
 clear_RA_files) then the searching is done in memory with the
 modified read_xxx_file() routines.
 A lethal bug in add_general_record (use of variable i instead of plane)
 was corrected.
 Beta 1.5				(M.P. 2.2.93)
 
 input_special added for CJB. freed pointer casted to type void *.
 input_all() added. ID status initialized to open ( i.e. closed = 0 ).
 Default directory taken from environmental variable HLSETC (or it is
 the current diretory, if HLSETC is missing). If the name of the
 structure file starts with '.', then the default directory is
 always the current directory.
 Beta 1.6				(M.P. 15.2.93)
 
 stupid error fixed: d_max from parameter_record now read in correctly.
 Beta 1.7				(M.P. 8.3.93)
 
 Modified to account for change in access_record and host_record:
 The routine find_host_record gets now the pair host,port which
 uniquely define the RPC channel. A global array channel and a
 routine cleanup_rpc() to allow for performing free_lpc() is added.
 2.0					(M.P. 15.4.93)
 
 cleanup_rpc now supports different services (lpc,cyc,bpm,bump,inj).
 2.1					(M.P. 19.5.93)
 
 Reads the new scraper parameters N_low N_high. For now, only the lpc
 and cyc services are supported.
 2.2					(M.P. 22.5.93)
 
 The input files (acc,cal,par and calibration files) may now contain
 comment lines (a hash - # - in the first column).
 2.3					(M.P. 27.5.93)
 
 The special elements list was changed to comply with hlsvar.h V1.3
 All RPC services now supported.
 2.4					(M.P. 11.8.93)
 
 The access record for bend PS gets a calibration pointer set to
 the record which allows energy scaling of the PS currents. Note
 that this is not a problem for quads, since there the relation
 I <-> k is unique.
 2.5					(M.P. 8.9.93)
 
 The reading of ps.I_start of the parameter_record for power supplies
 has been added. This assumes that the par files is properly filled.
 2.6					(M.P. 16.9.93)
 
 Reading of structure file changed such that if the name starts with
 a '.', the directory $HLSPREF is chosen instead of the current
 directory.
 2.7					(M.P. 28.9.93)
 
 Added free_sare() to cleanup_rpc().
 2.8					(M.P. 26.10.93)
 
 Modified add_ID_record such that the gap is initialized to the maximum
 gap, rather than to 0.
 2.9					(M.P. 14.4.94)
 
 Added the input of pAccInfo, for the beam info server.
 2.10					(M.P. 21.11.95)
 
 Fixed a bug: I_start for correctors was not initialized
 2.11					(M.P. 26.2.96)
 ==================================================================*/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "string.h"

/* The only place where hlsinit.h is included ! Everywhere else use hlsvar.h */

#include "hlsinit.h"
#include "hlsrpc.h"

#define TRUE	1
#define FALSE	0

/* the number of characters per input line, usually 80, therefore the name */

#define N80	132

/* global variables used within these input routines - no outside access ! */

typedef struct record_counter record_counter ;
struct record_counter {
	char 		ident[AL_str32+2*sizeof(int)] ;/* extra space for port*/
	void 		*ptr ;
	family_record 	*pFam ;/*only for use of the access chain with magnets*/
	record_counter	*next ;
} ;

static record_counter *host_chain = NULL ;
static record_counter *access_chain = NULL ;
static record_counter *parameter_chain = NULL ;
static record_counter *calibration_chain = NULL ;

/* the global array channel, which keeps pointers to all host_records */

static	host_record	*channel[AL_channel] ;
static	int		N_channel = 0 ;


/**************************************************************************/
/* The following functions include ORACLE I/O statements. They may have to 
   be modified if a different databse system is used. However, they do not
   include any reference to the HLS databush, i.e. no global variables or
   record structures.							  */
/**************************************************************************/

/* global file name (change for ORACLE version) */
static FILE *structure_file;

/* path where the files are located (change for ORACLE version) */
static char *path="" ;	/* the default path is the current directory */

/* structure and pointers to the copies of files in the heap */
typedef struct file_record file_record ;
struct file_record {
	char 		line[N80] ;
	file_record	*next ;
} ;
static file_record *acc_chain = NULL ;
static file_record *par_chain = NULL ;
static file_record *cal_chain = NULL ;


/* routine which reads one line and skip it if it is a comment */
char *get_line ( line, Nchar, fp )
FILE *fp ;
int  Nchar ;
char *line ;
{
  do  
  { 
    if ( NULL == fgets(line, Nchar, fp) ) return ( NULL ) ;
  } while ( *line == '#' ) ; /* skip comments */
  return ( line ) ;
}


/* this functions copies all random access files into the heap */

static void copy_RA_files ()
{
  file_record	*ptr ;
  FILE 	*RA_file;
  char 	line[N80], line2[N80] ;
  char	*chPtr ;
  char	fullName[N80] ;


  /* finds the environmental variable which tells the path for the HLS files */
  if ( NULL != ( chPtr = getenv ( "HLSETC" ) ) ) path = chPtr ;

  
  strcpy ( fullName, path ) ;
  strcat ( fullName, "cal" ) ;
  if ( NULL == ( RA_file = fopen (fullName,"r") ) ) 
  {
    printf ( "calibration file not found in %s\n", fullName ) ;
    exit ( -1 ) ;
  }
  while ( get_line(line, N80, RA_file) != NULL )
  {
    /* read second line cal file, and put in heap before first line,
       because the chain is read backwards ! */
    get_line(line2, N80, RA_file) ;
    if ( (ptr=malloc(sizeof(file_record))) == NULL )
    {
      printf ("cal! allocation error: not enough memory space - aborting\n") ;
      exit(0) ;
    }
    strcpy ( ptr->line, line2 ) ; ptr->next = cal_chain ; cal_chain = ptr ;
    if ( (ptr=malloc(sizeof(file_record))) == NULL )
    {
      printf ("cal! allocation error: not enough memory space - aborting\n") ;
      exit(0) ;
    }
    strcpy ( ptr->line, line ) ; ptr->next = cal_chain ; cal_chain = ptr ;
  }
  fclose (RA_file) ;

  strcpy ( fullName, path ) ;
  strcat ( fullName, "par" ) ;
  if ( NULL == ( RA_file = fopen (fullName,"r") ) ) 
  {
    printf ( "parameter file not found in %s\n", fullName ) ;
    exit ( -1 ) ;
  }
  while ( get_line(line, N80, RA_file) != NULL )
  {
    if ( (ptr=malloc(sizeof(file_record))) == NULL )
    {
      printf ("par! allocation error: not enough memory space - aborting\n") ;
      exit(0) ;
    }
    strcpy ( ptr->line, line ) ; ptr->next = par_chain ; par_chain = ptr ;
  }
  fclose (RA_file) ;

  strcpy ( fullName, path ) ;
  strcat ( fullName, "acc" ) ;
  if ( NULL == ( RA_file = fopen (fullName,"r") ) ) 
  {
    printf ( "access file not found in %s\n", fullName ) ;
    exit ( -1 ) ;
  }
  while ( get_line(line, N80, RA_file) != NULL )
  {
    if ( (ptr=malloc(sizeof(file_record))) == NULL )
    {
      printf ("acc! allocation error: not enough memory space - aborting\n") ;
      exit(0) ;
    }
    strcpy ( ptr->line, line ) ; ptr->next = acc_chain ; acc_chain = ptr ;
  }
  fclose (RA_file) ;

} /* copy_RA_files () */


static void clear_RA_files ()
{
  file_record	*root_new ;

  while ( acc_chain != NULL ) 
  { root_new = acc_chain->next ; free ( acc_chain ) ; acc_chain = root_new ; }
  while ( par_chain != NULL ) 
  { root_new = par_chain->next ; free ( par_chain ) ; par_chain = root_new ; }
  while ( cal_chain != NULL ) 
  { root_new = cal_chain->next ; free ( cal_chain ) ; cal_chain = root_new ; }

} /* clear_RA_files() */


/* the following three functions find one record in the databases */

static void read_calibration_file (ident, c, cr)

str13	ident ;	/* input data, key to be searched for */
double 	c[], cr[] ;

{
  int 	i ;
  str13	Ckey ;
  file_record *ptr = cal_chain ;

  /* here starts non ORACLE reading, to be replaced later */
  while ( ptr != NULL )
  {
    sprintf(Ckey,"");
    sscanf(ptr->line,"%s", Ckey) ;
    if ( ! strcmp(ident,Ckey) )
    {
      sscanf(ptr->line,"%s%lf%lf%lf%lf%lf", Ckey, c, c+1, c+2, c+3, c+4 ) ;
      ptr = ptr->next ;
      sscanf(ptr->line,"%lf%lf%lf%lf%lf", cr, cr+1, cr+2, cr+3, cr+4) ;

/* debugging 
      printf ( "%s ", Ckey ) ;
      for ( i=0;i<AL_calibration;i++ ) printf("%lf  ", c[i]) ;
      for ( i=0;i<AL_calibration;i++ ) printf("%lf  ", cr[i]) ;
      printf ( "\n" ) ;
end debug */

      return ;
    }
    ptr = ptr->next->next ; /* jump two lines in cal file ! */
  }

  /* here no matching record has been found */
  printf ("Key not found in calibration file: %s\n", ident) ;
  exit (0) ;
    
} /* read_calibration_file */


static void read_parameter_file (
      ident,I_low,I_high,rmp_s,rmp_f,i1,i2,freq_low,freq_high,d_max)

str13	ident ;
double 	*I_low, *I_high, *rmp_s, *rmp_f, *d_max;
int 	*i1, *i2 ;
double 	*freq_low, *freq_high ;

{
  str13	Pkey ;
  file_record *ptr = par_chain ;

  /* here starts non ORACLE reading */
  while ( ptr != NULL )
  {
    sscanf(ptr->line,"%s", Pkey);
    if ( ! strcmp(ident,Pkey) ) 
    {
      sscanf(ptr->line,"%s%lf%lf%lf%lf%i%i%lf%lf%lf",
          Pkey,I_low,I_high,rmp_s,rmp_f,i1,i2,freq_low,freq_high,d_max);
      return ;
    }
    ptr = ptr->next ;
  } ;
  printf ("Key not found in parameter file: %s\n", ident) ;
  exit (0) ;
  
} /* read_parameter_file */


static void read_access_file (ident, ident1, host, port, Pident, Cident1, Cident2)

str13 	ident;
str32 	host ;
int 	*port ;
str13 	ident1, Pident, Cident1, Cident2 ;

{
  str13	Akey ;
  file_record *ptr = acc_chain ;

  /* here starts non ORACLE reading */
  while ( ptr != NULL )
  {
    sscanf(ptr->line,"%s", Akey) ;
    if ( ! strcmp(ident,Akey) )
    {
      ident1[0] = '\0' ; Pident[0] = '\0' ; Cident1[0] = '\0' ; Cident2[0] = '\0' ;
      sscanf(ptr->line,"%s%s%s%i%s%s%s",Akey,ident1,host,port,Pident,Cident1,Cident2);
      return ;
    }
    ptr = ptr->next ;
  } ;
  printf ("Key not found in access file: %s\n", ident) ;
  exit (0) ;

} /* read_access_file */


/* the following three functions deal with the structure file: read a
   record, open file and close file */


static int read_structure_file ( name, section, type, Aident,
                		  Cident_B, Cident_L, sub_type )

str13 	name, Aident, Cident_B, Cident_L ;
int 	*section, *type, *sub_type ;

{
  char 	line[N80] ;
  char 	*chPtr ;

  /* here starts non ORACLE reading */
  if ( (chPtr = get_line(line, N80, structure_file)) != NULL )
  {
/*debug     printf ( "%s", line ) ;  */
     sscanf ( line, "%s%i%i%s%s%s%i", name, section, type, Aident,
                  Cident_B, Cident_L, sub_type ) ;
     return ( TRUE ) ;
  } 
  else { return (FALSE) ; } ;

} /* read_structure_file */


static int open_structure_file ( file_name )

str32	file_name ;

{
  char	fullName[N80] ;
  
  /* if file_name starts with '.', take HLSPREF directory as default */
  if ( *file_name == '.' ) strcpy ( fullName, getenv ( "HLSPREF" ) ) ;
  else strcpy ( fullName, path ) ;
  strcat ( fullName, file_name ) ;
  if ( NULL == ( structure_file = fopen (fullName,"r") ) ) 
  {
    printf ( "structure file not found in %s\n", fullName ) ;
    return ( -1 ) ;
  }
  else return ( 0 ) ;
}


static void close_structure_file () 

{
  fclose ( structure_file ) ;
}


/**************************************************************************/
/* end of ORACLE specific data                                            */
/**************************************************************************/


/**************************************************************************/
/* The following routines create and build the complete data bush         */
/**************************************************************************/



/* three utility functions: to allocate dynamic memory for the records of
   the data bush, to build and search the chain of pointers to the
   records of the lower layers of the data bush and finally to deallocate
   the chain, which is used only during creation of the data bush */


static void *alloc_record ( size )
unsigned int size ;
{
  void 	*ptr ;
  if ( (ptr=malloc(size)) == NULL )
  {
    printf ("allocation error: not enough memory space - aborting\n") ;
    exit(0) ;
  }
  return ptr ;
}


static void *find_link ( ident, size, start_chain )

char 		*ident ;
unsigned int 	size ;
record_counter 	*(*start_chain);

{
  record_counter *chain = *start_chain ;

  /* search access_chain for ident */
  while ( chain != NULL ) 
  {
    /* if ident found already in records, return pointer to record */
    if ( ! strcmp (ident, chain->ident) ) return (chain->ptr) ;
    chain = chain->next ;
  }

  /* if ident not found, create new record_counter then insert into chain*/
  chain = (record_counter *) alloc_record ( sizeof(record_counter) ) ;
  chain->ptr = alloc_record ( size ) ;
  strcpy ( chain->ident, ident ) ;
  chain->next = *start_chain ;
  *start_chain = chain ;

  return ( NULL ) ;

}


static void free_chain(root)

record_counter 		*root ;

{
  record_counter	*root_new ;

  while ( root != NULL ) 
  {
    root_new = root->next ;
    free ( (void *) root ) ;
    root = root_new ;
  }
} /* free_chain */


/* the next four functions return the pointer to the record identified
   by ident (either existing, or newly created and read from the databases) */


static host_record *find_host_record (host, port)

str32	host ;
int	port ;

{
  host_record	*ptrRec ;
  char		ident[AL_str32+2*sizeof(int)] ;
  char		*pp, *pi ;
  int		i ;

  /* convert host and port into one single string for find_link */
  strcpy ( ident, host ) ;
  pp = (char *) &port ;
  pi = ident + strlen(ident) ;
  for(i=0;i<sizeof(port);i++) {
    if (*(pp) == '\0') { *(pi++) = '\1' ; *(pi++) = '\1' ; } /* convert \0 */
    else *(pi++) = *(pp) ; 
    pp++ ;
  }
  *pi = '\0' ; /* null terminate appended string ! */

  /* if ident found already in host_records, return pointer to record */
  if ( NULL != ( ptrRec = (host_record *)
             find_link (ident, (unsigned int)sizeof(*ptrRec), &host_chain) ) )
    return ( ptrRec ) ;
    
  /* otherwise create new entry in host_chain and return new pointer */
  ptrRec = (host_record *) host_chain->ptr ;
  if ( N_channel >= AL_channel ) {
    printf ( "Access file: too many different host/port combinations.\
     		\nIncrease AL_channel in hlsdef.h\n" ) ;
    exit(0) ; 
  }
  channel[N_channel++] = ptrRec ;
  strcpy ( ptrRec->host, host ) ;
  ptrRec->port = port ;
  ptrRec->sid = -1 ;
  return ( ptrRec ) ;
  
} /* find_host_record */


static calibration_record *find_calibration_record (ident)

str13	ident ;

{
  calibration_record 	*ptrRec ;
  int 			i ;

  /* if ident found already in calibration_records, return pointer to record */
  if ( NULL != ( ptrRec = (calibration_record *)
         find_link (ident, (unsigned int)sizeof(*ptrRec), &calibration_chain) ) )
    return ( ptrRec ) ;
    
  ptrRec = (calibration_record *) calibration_chain->ptr ;

  /* here ORACLE is read, keyword=ident */
  read_calibration_file (ident, ptrRec->c, ptrRec->cr) ;
  
  /* determine number of significant coefficients */
  ptrRec->Nc = 0 ;
  ptrRec->Ncr = 0 ;
  for (i=AL_calibration;i>0;i--) 
    if ( ptrRec->c[i-1] != (double) 0.0 ) { ptrRec->Nc = i ; break ; }
  for (i=AL_calibration;i>0;i--) 
    if ( ptrRec->cr[i-1] != (double) 0.0 ) { ptrRec->Ncr = i ; break ; }

/* debugging 
printf ( 
  "Cid=%s, Nc=%i, c=%lf,%lf,%lf,%lf,%lf,  Ncr=%i, cr=%lf,%lf,%lf,%lf,%lf\n",
  ident, ptrRec->Nc, 
  ptrRec->c[0],ptrRec->c[1],ptrRec->c[2],ptrRec->c[3],ptrRec->c[4],
  ptrRec->Ncr, 
  ptrRec->cr[0],ptrRec->cr[1],ptrRec->cr[2],ptrRec->cr[3],ptrRec->cr[4] ) ;
end debug */
    
  return ( ptrRec ) ;
  
} /* find_calibration_record */


static parameter_record *find_parameter_record (type, ident)

int 	type ;
str13 	ident ;

{
  parameter_record 	*ptrRec ;
  double		I_low, I_high, rmp_s, rmp_f, d_max ;
  int 			i1, i2 ;
  double 		freq_low, freq_high ;

  /* if ident found already in parameter_records, return pointer to record */
  if ( NULL != ( ptrRec = (parameter_record *)
           find_link (ident, (unsigned int)sizeof(*ptrRec), &parameter_chain) ) )
    return ( ptrRec ) ;
    
  ptrRec = (parameter_record *) parameter_chain->ptr ;

  /* here ORACLE database is read, keyword=ident */
  read_parameter_file (ident, &I_low, &I_high, &rmp_s, &rmp_f,
                                  &i1, &i2, &freq_low, &freq_high, &d_max) ;
  switch (type)
  {
    case bend_type:
      ptrRec->ps.angle = freq_high ;
    case quad_type:
    case sextupole_type: 
    case corrector_type:
      ptrRec->ps.I_start = d_max ; 
      ptrRec->ps.I_low = I_low ;
      ptrRec->ps.I_high = I_high ;
      ptrRec->ps.rmp_s = rmp_s ;
      ptrRec->ps.rmp_f = rmp_f ;
      ptrRec->ps.i1 = (short) i1 ;
      ptrRec->ps.i2 = (short) i2 ;
      ptrRec->ps.length = freq_low ;
    break ;  
    case ID_type:
      ptrRec->id.gh_low = I_low ;
      ptrRec->id.gh_high = I_high ;
      ptrRec->id.g_low = rmp_s ;
      ptrRec->id.g_high = rmp_f ;
      ptrRec->id.type = i1 ;
      ptrRec->id.Np = i2 ;
      ptrRec->id.kx = freq_low ;
      ptrRec->id.period = freq_high ;
      ptrRec->id.kyh = d_max ;
    break ; 
    case septum_type: 
      ptrRec->in.angle = freq_high ;
    case kicker_type:
      ptrRec->in.V_low = I_low ;
      ptrRec->in.V_high = I_high ;
      ptrRec->in.t_peak = rmp_s ;
      ptrRec->in.t_full = rmp_f ;
      ptrRec->in.length = freq_low ;
      ptrRec->in.d_max = d_max ;
    break ;  
    case scraper_type:
      ptrRec->sc.b_low = I_low ;
      ptrRec->sc.b_high = I_high ;
      ptrRec->sc.s_low = rmp_s ;
      ptrRec->sc.s_high = rmp_f ;
      ptrRec->sc.type = i1 ;
      ptrRec->sc.N_low = freq_low ;
      ptrRec->sc.N_high = freq_high ;
    break ;  
    case cavity_type:
      ptrRec->cav.temp_low = I_low ;
      ptrRec->cav.temp_high = I_high ;
      ptrRec->cav.voltage_low = rmp_s ;
      ptrRec->cav.voltage_high = rmp_f ;
    break ;  
    case RF_type:
      ptrRec->rf.V_low = I_low ;
      ptrRec->rf.V_high = I_high ;
      ptrRec->rf.p_low = rmp_s ;
      ptrRec->rf.p_high = rmp_f ;
      ptrRec->rf.f_low = freq_low ;
      ptrRec->rf.f_high = freq_high ;
    break ;  
    default:
      printf ("This type not supported in find_parameter_record: %i\n", type );
      printf ("Exit at identifier %s\n", ident ) ;
      exit (0) ;
  }  
  return ( ptrRec ) ;

} /* find_parameter_record */


static access_record *find_access_record (type, ident)

int 	type ;
str13 	ident ;

{
  access_record	*ptrRec ;
  char 		*_Ptr ;
  str13 	fam_mem, ident1, Pident, Cident1, Cident2 ;
  str32 	host ;
  int 		port, i,j,n ;

   
  /* read ORACLE, key=ident */
   
  read_access_file ( ident, ident1, host, &port, Pident, Cident1, Cident2) ;

  /* if ident1 found already in access_records, return pointer to record */
  if ( NULL != ( ptrRec = (access_record *)
             find_link (ident1, (unsigned int)sizeof(*ptrRec), &access_chain) ) )
    return ( ptrRec ) ;

ptrRec = (access_record *) access_chain->ptr;
  
  

  /* take fam_mem from Cident2 for special elements !! */
  if ( type == RF_type || type == special_type ) strcpy (fam_mem,Cident2) ;
  else strcpy (fam_mem, ident1) ;	/* leave ident untouched ! */
   


  /*now break ident1 into family and member strings */
  if ( NULL == ( _Ptr = strchr ( fam_mem, '_' ) ) )
  {
    printf ("Wrong family_member in access file: %s\n", ident1) ;
    exit (0) ;
  }
  *(_Ptr++) = '\0' ; 
  
  if ( strlen(fam_mem) > AL_str6 || strlen(_Ptr) > AL_str6 )
  {
    *(--_Ptr) = '_' ;
    printf ("Wrong family_member in access file: %s\n", ident1) ;
    exit (0) ;
  }
  
  

  strcpy ( ptrRec->family, fam_mem ) ;
  strcpy ( ptrRec->member, _Ptr ) ;


  ptrRec->pHost = find_host_record ( host, port ) ;
  ptrRec->pPar = NULL ;
  ptrRec->pCal1 = NULL ;
  ptrRec->pCal2 = NULL ;

  switch (type)
  {
    case bend_type:
      /* only for bending magnets: get calibration pointer to
         the records, which allows energy scaling of the PS current */
      ptrRec->pCal1 = find_calibration_record (Cident1) ;
    case quad_type:
    case sextupole_type: 
      ptrRec->pPar = find_parameter_record (type, Pident) ;
      /* create family record and store pointer for add_magnet_record */
      access_chain->pFam = (family_record *)alloc_record(sizeof(family_record));
      access_chain->pFam->pAcc = ptrRec ;
      access_chain->pFam->N_member = 0 ;
      /* add family pointer in fa[] array */
      if ( N_family >= AL_fa ) {
        printf ( "Access file: too many families with element:%s\n", ident1 ) ;
        exit(0) ; 
      }
      fa[N_family++] = access_chain->pFam ;
    break ;      
    case ID_type:
      ptrRec->pPar = find_parameter_record (type, Pident) ;
      ptrRec->pCal1 = find_calibration_record (Cident1) ;
      ptrRec->pCal2 = find_calibration_record (Cident2) ;
    break ;  
    case cavity_type:
      ptrRec->pPar = find_parameter_record (type, Pident) ;
      ptrRec->pCal1 = find_calibration_record (Cident1) ;
      ptrRec->pCal2 = find_calibration_record (Cident2) ;
    break ;  
    case septum_type:
    case corrector_type:
    case chicane_type:
    case kicker_type:
      ptrRec->pPar = find_parameter_record (type, Pident) ;
      ptrRec->pCal1 = find_calibration_record (Cident1) ;
    break ;  
    case scraper_type:
      ptrRec->pPar = find_parameter_record (type, Pident) ;
    break ;  

    case FS_type:
    case SRPM_type:
    case CM_type:
    break ;

    case BPM_type:
      ptrRec->pCal1 = find_calibration_record (Cident1) ;
    break ;  
      
    case RF_type:
      ptrRec->pPar = find_parameter_record (type, Pident) ;
    break ;  

    case special_type:	/* all special elements must have one calibration */
      ptrRec->pCal1 = find_calibration_record (Cident1) ;
    break ;  
      
    default:
      printf ("This type not supported in find_access_record: %i\n", type ) ;
      printf ("Exit at identifier %s\n", ident ) ;
      exit (0) ;
  }  
        
  return ( ptrRec ) ;

} /* find_access_record */


/* the add_..._record functions which allocate records of the first layer
   and fill the global pointer arrays of hlsvar.h */


static void add_general_record ( name, sect, type, length, index, ptr ) 

str13 	name ;
int 	sect ;
int 	type ;
double 	length ;
int 	index ; 
void 	*ptr ; /* it is assumed that the generic type is allowed */

{
  int 	plane, i,j ;

  if ( N_general >= AL_ge )
     { printf ( "too many general elements with element:%s\n", name );exit(0);}

  /* first allocate a new general record */
  ge[N_general] = (general_record *) alloc_record ( sizeof(general_record) ) ;

  /* fill general record data */
  strcpy ( ge[N_general]->name, name ) ;
  ge[N_general]->section = sect ;
  ge[N_general]->type = type ;
  ge[N_general]->position = total_length ;
  ge[N_general]->index = index ;
  ge[N_general]->p.generic = ptr ;

  /* allocate transfer matrix and twiss record to thick elements */
  if ( type < thin_type ) 
  {
    /* allocate transfer matrix of thick element and set it to drift space */
    ge[N_general]->transfer = (matrix_2D *) alloc_record ( sizeof(matrix_2D) ) ;
    for (plane=horizontal;plane<=vertical;plane++)
    {
      for(i=0;i<3;i++) for(j=0;j<3;j++) 
        if (i==j) ge[N_general]->transfer->m[plane][i][j] = 1.0 ;
        else ge[N_general]->transfer->m[plane][i][j] = 0.0 ;
      ge[N_general]->transfer->m[plane][0][1] = length ;
    }
    /* allocate twiss record of thick element */
    ge[N_general]->twiss = (twiss_record *) alloc_record(sizeof(twiss_record)) ;
    /* point all previous adjacent thin element pointers to this twiss_record */
    for (i=N_general-1;i>=0;i--)
    {
      if (ge[i]->type < thin_type ) break ;
      ge[i]->twiss = ge[N_general]->twiss ;
    }
  }
  else 
  { 
    /* point zero length elements to unit matrix (defined in hlsvar.h) */
    ge[N_general]->transfer = &unit_matrix ;
    /* point twiss pointer to beginning of structure, in case this is the
       last element of the structure - otherwise it will be overwritten anyway*/
    ge[N_general]->twiss = ge[0]->twiss ;
  }

  /* calculate length of the whole structure */
  total_length += length ;

  /* assuming that the sections are numbered from 1 on. */
  if ( sect==1 ) section_length += length ;
  if ( sect > N_section ) N_section = sect ; 

  N_general++ ;

} /* add_general_record */


static void add_drift_record ( name, sect, length ) 

str13 	name ;
int 	sect ;
double 	length ;

{
  if ( N_drift >= AL_dr )
     { printf ( "too many drifts with element:%s\n", name ) ; exit(0) ; }

  /* first allocate a new drift record */
  dr[N_drift] = (drift_record *) alloc_record ( sizeof(drift_record) ) ;

  /* fill drift record data */
  dr[N_drift]->length = length ;

  /* create general record */
  add_general_record ( name, sect, drift_type,length, N_drift, dr[N_drift] ) ;
  dr[N_drift]->back_index = N_general - 1 ; /* counting starts from 0 */
  dr[N_drift]->pGen = ge[N_general-1] ;

  N_drift++ ;

} /* add_drift_record */


static void add_magnet_record ( type,name,sect,pAcc,pCalB,pCalL,sub_type ) 

int		   type ;
str13		   name ;
int		   sect ;
access_record	   *pAcc;
calibration_record *pCalB, *pCalL;
int		   sub_type ;

{
  double           length ;
  magnet_record    *ptr ;
  int              *pN ;
  record_counter   *chain = access_chain ; /* initialize to start of chain */

  /* first allocate a new magnet record */
  ptr = (magnet_record *) alloc_record ( sizeof(magnet_record) ) ;

  /* find family record*/
  /* search access_chain for same access pointer (chain->ptr) */
  while ( pAcc != chain->ptr ) chain = chain->next ;
  ptr->pFam = chain->pFam ;
  if ( ptr->pFam->N_member >= AL_member) 
     {printf( "too many elements per family with element:%s\n", name);exit(0);}

  ptr->pFam->pMember[ptr->pFam->N_member] = ptr ;
  (ptr->pFam->N_member)++ ;

  /* fill the type independent data */
  ptr->pPar = pAcc->pPar ;
  ptr->pCalB = pCalB ;
  ptr->pCalL = pCalL ;
  ptr->sub_type = sub_type ;

  length = pAcc->pPar->ps.length ;

  /* initialize magnetic_length to length (sufficient for most purposes) */
  ptr->magnetic_length = length ;


  /* fill type specific data */
  switch (type) 
  {
    case bend_type: 
      if ( N_bend >= AL_be )
         {printf ( "too many bends with element:%s\n", name ); exit(0); }
      pN = &N_bend ; 
      be[*pN] = ptr ;
      /* consider that edges have zero length, despite being thick type ! */
      switch (sub_type%reversed_polarity)
      {
        case hor_rectangle: case ver_rectangle: case com_rectangle:
        case hor_sector:    case ver_sector:    case com_sector:
        break ;
        case hor_edge:
        case ver_edge:
          length = 0.0 ; /* set length to zero for edge elements */
        break ;
        default:
          printf ("This sub_type not supported for bends: %i\n", sub_type ) ;
          printf ("Exit at element %s\n", name ) ;
          exit (0) ;
      }
    break ;
    case quad_type: 
      if ( N_quad >= AL_qu )
         { printf ( "too many quads with element:%s\n", name ); exit(0); }
      pN = &N_quad ; 
      qu[*pN] = ptr ; 
    break ;
    case sextupole_type: 
      if ( N_sextupole >= AL_se )
         { printf ( "too many sextupoles with element:%s\n", name ); exit(0); }
      pN = &N_sextupole ; 
      se[*pN] = ptr ;
      length = 0.0 ; /* set length to zero for sextupoles ! */ 
    break ;
  }
  
  /* create general record */
  add_general_record ( name, sect, type, length, *pN, ptr ) ;
  ptr->back_index = N_general - 1 ; /* counting starts from 0 */
  ptr->pGen = ge[N_general-1] ;

  (*pN)++ ;

} /* add_magnet_record */


static void add_ID_record ( name, sect, pAcc, pAcch ) 

str13 		name ;
int 		sect ;
access_record 	*pAcc, *pAcch ;

{
  char	fullName[N80] ;
  FILE 	*K_file;
  char 	line[N80] ;
  char 	*chPtr ;
  double    k, kx, ky, kxh, kyh;
  double    pip;
  int Nc;
  
  if ( N_ID >= AL_id )
     { printf ( "too many IDs with element:%s\n", name ) ; exit(0) ; }

  /* first allocate a new ID record */
  id[N_ID] = (ID_record *) alloc_record(sizeof(ID_record));
  id[N_ID]->pKal = NULL;
  
  /* fill ID specific data */
  id[N_ID]->pAcc[0] = pAcc ;
  id[N_ID]->pAcc[1] = NULL ;
  id[N_ID]->pPar = pAcc->pPar ;
  id[N_ID]->closed = 0 ;	/* ID initially assummed to be open ! */
  id[N_ID]->gap = pAcc->pPar->id.g_high ;	/* set to max gap */
  id[N_ID]->gap_h = pAcc->pPar->id.gh_high ;	/* set to max gap */
  id[N_ID]->phase = 0.0 ;	/* ID initially assummed to be at 0 phase ! */
  id[N_ID]->k_x = 0.0;
  id[N_ID]->k_y = 0.0;
  id[N_ID]->kh_x = 0.0;
  id[N_ID]->kh_y = 0.0;
  
  switch (pAcc->pPar->id.type)
  {
    case EEW:
      id[N_ID]->pAcc[1] = pAcch ;

      id[N_ID]->gap = 0.0 ;	                  
      id[N_ID]->gap_h = 0.0 ;	
    break;
    
    case modulator:
      id[N_ID]->gap = 0.0 ;	
      id[N_ID]->gap_h = 0.0 ;	
    break;

    case sazaki:
      strcpy ( fullName, path ) ;
      strcat ( fullName, name ) ;
      id[N_ID]->pKal = (kid_calibration_record *) alloc_record(sizeof(kid_calibration_record));
      id[N_ID]->pKal->Nc=0;
      if ( NULL == ( K_file = fopen (fullName,"r") ) ) 
      {
         printf ( "roll off calibration file not found in %s\n", fullName ) ;
         exit ( -1 ) ;
      }
      while ( (chPtr = get_line(line, N80, K_file)) != NULL )
      {
        Nc=id[N_ID]->pKal->Nc;
        sscanf ( line, "%lf%lf%lf%lf%lf%lf", &id[N_ID]->pKal->gap[Nc], 
        &id[N_ID]->pKal->phase[Nc], &id[N_ID]->pKal->kx1[Nc], &id[N_ID]->pKal->ky1[Nc],
        &id[N_ID]->pKal->kx[Nc], &id[N_ID]->pKal->ky[Nc]) ;

        id[N_ID]->pKal->kx1[Nc]=-id[N_ID]->pKal->kx1[Nc];
        id[N_ID]->pKal->ky1[Nc]=-id[N_ID]->pKal->ky1[Nc];
        id[N_ID]->pKal->kx[Nc]=-id[N_ID]->pKal->kx[Nc];
        id[N_ID]->pKal->ky[Nc]=-id[N_ID]->pKal->ky[Nc];

        id[N_ID]->pKal->Nc++;
        if ( id[N_ID]->pKal->Nc == AL_kid_calibration )
        {
          printf("Too many lines in %s!\n", fullName);
          printf("Increase variable AL_kid_calibration.\n");
          exit(-1);
        }
      } ;
      
      fclose(K_file);  
    break;
    
    default:
    break;
  }
    pip = acos(-1.0); 
    k = 2*pip/pAcc->pPar->id.period ;
    id[N_ID]->k_x = pAcc->pPar->id.kx ;	
    id[N_ID]->kh_y = pAcc->pPar->id.kyh ;
    kx=id[N_ID]->k_x;
    kyh=id[N_ID]->kh_y;
    
    if ( (kx <= 0.0) && (fabs(kx) <= k) ) { ky = -sqrt( k*k - kx*kx );  }
    if ( (kx <= 0.0) && (fabs(kx) > k) ) { ky = sqrt( kx*kx - k*k);  }
    if (( kx > 0.0 ) && (kx <= k) ) { ky = -sqrt( k*k + kx*kx ); }
    if ( kx > k ) { ky = -sqrt( k*k + kx*kx ); }
    id[N_ID]->k_y = ky ;	

    if ( (kyh <= 0.0) && (fabs(kyh) <= k) ) { kxh = -sqrt( k*k - kyh*kyh );  }
    if ( (kyh <= 0.0) && (fabs(kyh) > k) ) { kxh = sqrt( kyh*kyh - k*k);  }
    if (( kyh > 0.0 ) && (kyh <= k) ) { kxh = -sqrt( k*k + kyh*kyh ); }
    if ( kyh > k ) { kxh = -sqrt( k*k + kyh*kyh ); }
    id[N_ID]->kh_x = kxh;
        
    id[N_ID]->curvature = 0.0;	
    id[N_ID]->curvature_h = 0.0;	
  /* create general record */
  add_general_record ( name, sect, ID_type,
             pAcc->pPar->id.period*pAcc->pPar->id.Np, N_ID, id[N_ID] ) ;
  id[N_ID]->back_index = N_general - 1 ; /* counting starts from 0 */
  id[N_ID]->pGen = ge[N_general-1] ;

  N_ID++ ;

} /* add_ID_record */

static void add_cavity_record ( name, sect, pAcc, pAccV) 

str13 		name ;
int 		sect ;
access_record 	*pAcc, *pAccV ;

{
  if ( N_cavity >= AL_cav )
     { printf ( "too many cavities with element:%s\n", name ) ; exit(0) ; }

  /* first allocate a new ID record */
  cav[N_cavity] = (cavity_record *) alloc_record(sizeof(cavity_record));

  /* fill cavity specific data */
  cav[N_cavity]->pAcc[0] = pAcc ;
  cav[N_cavity]->pAcc[1] = pAccV ;
  cav[N_cavity]->pPar = pAcc->pPar ;
  cav[N_cavity]->temperature=pAcc->pPar->cav.temp_low;
  cav[N_cavity]->voltage=pAcc->pPar->cav.voltage_low;
  /* create general record */
  add_general_record ( name, sect, cavity_type,
             0.0, N_cavity, cav[N_cavity] ) ;
  cav[N_cavity]->back_index = N_general - 1 ; /* counting starts from 0 */
  cav[N_cavity]->pGen = ge[N_general-1] ;

  N_cavity++ ;

} /* add_cavity_record */

void add_corrector_record ( type, name, sect, pAccH, pAccV, plane ) 

str13 		name ;
int 		type, sect ;
access_record 	*pAccH, *pAccV ;
int 		plane ;

{

  switch (type)
  {
  case corrector_type:
    if ( N_corrector >= AL_co )
      { printf ( "too many correctors with element:%s\n", name ) ; exit(0) ; }

    /* first allocate a new corrector record */
    co[N_corrector] = (corrector_record *) alloc_record(sizeof(corrector_record));

    /* fill corrector record data */
    co[N_corrector]->pAcc[horizontal] = pAccH ;
    co[N_corrector]->pAcc[vertical] = pAccV ;
    co[N_corrector]->plane = plane ;

    /* create general record */
    add_general_record (name,sect,type,0.0,N_corrector,co[N_corrector]);
    co[N_corrector]->back_index = N_general - 1 ; /* counting starts from 0 */
    co[N_corrector]->pGen = ge[N_general-1] ;

    N_corrector++ ;
    break;
  
  case chicane_type:
    if ( N_chicane >= AL_chi )
      { printf ( "too many chicanes with element:%s\n", name ) ; exit(0) ; }

    /* first allocate a new chicane record */
    chi[N_chicane] = (corrector_record *) alloc_record(sizeof(corrector_record));

    /* fill corrector record data */
    chi[N_chicane]->pAcc[horizontal] = pAccH ;
    chi[N_chicane]->pAcc[vertical] = pAccV ;
    chi[N_chicane]->plane = plane ;

    /* create general record */
    add_general_record (name,sect,type,0.0,N_chicane,chi[N_chicane]);
    chi[N_chicane]->back_index = N_general - 1 ; /* counting starts from 0 */
    chi[N_chicane]->pGen = ge[N_general-1] ;

    N_chicane++ ;
    break;
    
  default:
    break;
  
  }
} /* add_corrector_record */


void add_simple_record ( type, name, sect, pAcc, length, size, 
						AL_type, N_type, ptr ) 

/* this routine fills all simple records: septum, kicker, scraper, monitor,
   CM and tag  and cavities for now*/

str13 		name ;
int 		type, sect, AL_type, *N_type ;
unsigned int 	size ;
access_record 	*pAcc ;
double 		length ;
struct 		{
		    int			back_index ;
		    general_record	*pGen ;
		    access_record	*pAcc ;
		} **ptr ; 	/* a dirty trick: all simple records start 
				with these three variables so ptr can be 
				used as a generic pointer to all simple 
				records ! */

{
  if ( *N_type >= AL_type )
     { printf ( "too many elements with element:%s\n", name ) ; exit(0) ; }

  /* first allocate a new record */
  *ptr = alloc_record ( size ) ;

  /* fill access pointer */
  (*ptr)->pAcc = pAcc ;

  /* create general record */
  add_general_record (name, sect, type, length, *N_type, *ptr) ;
  (*ptr)->back_index = N_general - 1 ; /* counting starts from 0 */
  (*ptr)->pGen = ge[N_general-1] ;

  (*N_type)++ ;
} /* add_simple_record */


/**************************************************************************/
/* The three input functions which are accessible to the outside world    
   and are declared in hlsfun.h:					  */
/**************************************************************************/


void input_open_files () 

{
/* first initialize the energy to 0.0, to allow testing whether the
   energy has been obtained from the field or file or not */
   
  energy.E = 0.0 ;

/* read ORACLE data into dynamic memory, where it can be searched faster */
  copy_RA_files() ;

/* here read extra data: energy SR & TL and BPM SR access file data */
/* the string constants must be compatible with the type str13 ! */

  pAccRF = find_access_record ( RF_type, "RF" ) ;

  pAccTune = find_access_record ( special_type, "TUNE" ) ;
  pAccDCCT = find_access_record ( special_type, "DCCT" ) ;
  pAccBPM_SR = find_access_record ( special_type, "BPM_SR" ) ;
  energy.pAcc[0] = find_access_record ( special_type, "ENERGY_TL" ) ;
  energy.pAcc[SR-TL] = find_access_record ( special_type, "ENERGY_SR" ) ;
  
  /* new 21.11.95 */
  pAccInfo = find_access_record ( special_type, "BEAM_INFO" ) ;
}


int input_structure  ( file_name ) 

str32	file_name ;

{
access_record *pAcc, *pAccV ;
calibration_record *pCalB, *pCalL ;
str13 name,Aident,Cident_B,Cident_L ;
int section,type,sub_type ;
double length ;
int	err;
/* printf("\n I am in input_structure \n"); */

/* open ORACLE structure from outside routine to simplify porting */
/* return if error occured during opening */
if ( ( err = open_structure_file ( file_name ) ) != 0 ) return ( err ) ;

/* here ORACLE structure file database is read from beginning to end */

while ( read_structure_file( name, &section, &type, Aident,
 					Cident_B, Cident_L, &sub_type) )
/* printf("%s %i %i %s %s %s %i\n", name, section, type, 
			Aident,Cident_B, Cident_L, sub_type) ; */
{
  switch (type)
  {
  
    case drift_type: 
      sscanf ( Aident, "%lf", &length ) ;
      add_drift_record ( name, section, length ) ;
    break ;
            
    case bend_type:
    case quad_type:
      pAcc = find_access_record (type, Aident) ;
      pCalB = find_calibration_record (Cident_B) ;
      pCalL = find_calibration_record (Cident_L) ;
      add_magnet_record (type, name, section, pAcc, pCalB, pCalL, sub_type) ;
    break ;

    case ID_type:
      pAcc = find_access_record (type, Aident) ;
      pAccV = find_access_record (type, Cident_B) ;
      add_ID_record (name, section, pAcc, pAccV) ;
    break ;
                               
    case septum_type:
      pAcc = find_access_record (type, Aident) ;
      add_simple_record (type, name, section, pAcc, pAcc->pPar->in.length,
      			   sizeof(*sp[0]), AL_sp, &N_septum, &sp[N_septum] ) ;
    break ;
                               
    case sextupole_type: 
      pAcc = find_access_record (type, Aident) ;
      pCalB = find_calibration_record (Cident_B) ;
      add_magnet_record (type, name, section, pAcc, pCalB, NULL, 0) ;
    break ;

    case corrector_type: 
    case chicane_type: 
      switch (sub_type )
      {
        case horizontal:
          pAcc = find_access_record (type, Aident) ;
          pAccV = NULL ;
        break ;
        case vertical:
    	  pAccV = find_access_record (type, Cident_B) ;
          pAcc = NULL ;
        break;
        case lumped:
          pAcc = find_access_record (type, Aident) ;
    	  pAccV = find_access_record (type, Cident_B) ;
        break ;
      }
      add_corrector_record (type, name, section, pAcc, pAccV, sub_type) ;
    break ;

    case kicker_type: 
      pAcc = find_access_record (type, Aident) ;
      add_simple_record (type, name, section, pAcc, 0.0,
      			   sizeof(*ki[0]), AL_ki, &N_kicker, &ki[N_kicker] ) ;
    break ;

    case scraper_type: 
      pAcc = find_access_record (type, Aident) ;
      add_simple_record (type, name, section, pAcc, 0.0,
                             sizeof(*sc[0]), AL_sc, &N_scraper, &sc[N_scraper]);
      break ;
                           
      case FS_type:
        pAcc = find_access_record (type, Aident) ;
        add_simple_record (type, name, section, pAcc, 0.0,
                             sizeof(*fs[0]), AL_fs, &N_FS, &fs[N_FS] ) ;
      break ;

      case SRPM_type: 
        pAcc = find_access_record (type, Aident) ;
        add_simple_record (type, name, section, pAcc, 0.0,
                             sizeof(*pr[0]), AL_pr, &N_SRPM, &pr[N_SRPM] ) ;
      break ;

      case BPM_type: 
        pAcc = find_access_record (type, Aident) ;
        add_simple_record (type, name, section, pAcc, 0.0,
                             sizeof(*po[0]), AL_po, &N_BPM, &po[N_BPM] ) ;
      break ;
              
      case CM_type: 
        pAcc = find_access_record (type, Aident) ;
        add_simple_record (type, name, section, pAcc, 0.0,
                             sizeof(*cm[0]), AL_cm, &N_CM, &cm[N_CM] ) ;
      break ;
      
      case cavity_type: 
        pAcc = find_access_record (type, Aident) ;
    	  pAccV = find_access_record (type, Cident_B) ;
        add_cavity_record (name, section, pAcc, pAccV) ;
      break ;
             
      case tag_type: 
        add_simple_record (type, name, section, NULL, 0.0,
                             sizeof(*ta[0]), AL_ta, &N_tag, &ta[N_tag] ) ;
      break ;
             
    default:
      printf ("This type not supported in the structure file: %i\n", type ) ;
      printf ("Exit at element %s, section %i\n", name, section ) ;
      exit (0) ;

    }  /* switch */

  }  /* while not eof */

  close_structure_file () ;
  
  return ( 0 ) ;

} /* input_structure */


void input_special ( type, key, ppAcc ) 

int		type ;
str13		key ;
access_record	**ppAcc ;

{
  *ppAcc = find_access_record ( type, key ) ;
}


void input_close_files()

{
  /* this must be done after all reading !! If two structure files are read */

  clear_RA_files () ; /* empties heap from random access file copies */

  free_chain (host_chain) ;
  free_chain (access_chain) ;
  free_chain (parameter_chain) ;
  free_chain (calibration_chain) ;
}


void	input_all ( file_name ) 

str32	file_name ;

{
  input_open_files() ;
  if ( 0 > input_structure ( file_name ) )
  {
    printf ( "cannot open structure file\n" ) ;
    exit ( 0 ) ;
  }
  input_close_files() ;
}


/**************************************************************************/
/* Finally the function to cleanup any RPC related data. It must be called
   after any RPC's are done.					          */
/**************************************************************************/

void	cleanup_rpc ()

{
  int i ;
  
  for ( i=0; i<N_channel; i++ )
    if ( channel[i]->sid >= 0 ) {
/*      switch ( channel[i]->service ) 
      {
        case service_lpc:  free_lpc ( channel[i]->sid ) ; break ;
        case service_cyc:  free_cyc ( channel[i]->sid ) ; break ;
        case service_bpm:  free_bpm ( channel[i]->sid ) ; break ;
        case service_bump: free_bump( channel[i]->sid ) ; break ;
        case service_inj:  free_inj ( channel[i]->sid ) ; break ;
        case service_sare: free_sare( channel[i]->sid ) ; break ;
        case service_flsc: free_flsc( channel[i]->sid ) ; break ;
        case service_info: free_beam_info( channel[i]->sid ) ; break ;
      } LT switch */
      
      channel[i]->sid = -2 ; /* set to -2 so do_rpc can ignore further calls */
    } /* if */
}


/*==================================================================
 ************************* end of input.c *************************
==================================================================*/
