/*********************************************************************
*   File : Plot_Twiss.c
**********************************************************************
mod. 22/12/92  imposed subtype for ID = fnom after label toolong:
mod. 05/01/93  transfer matrices become 3x3 instead of 2x2
mod. 23/02/93  added all types of bending magnets & septum
**********************************************************************
created : Dec. 14th 1992      *       Last modified : Feb. 23rd 1993
*********************************************************************
The function Plot_Twiss :

int plot_twiss( ptrTwiss0, io, so, il, sl, fnom, n, ptr_bh, ptr_bv,
               ptr_dh,ptr_dv)

returns four arrays containing respectively 'n' horizontal beta values in
'ptr_bh' , 'n' vertical ones in 'ptr_bv' , 'n' horizontal dispersion values in
'ptr_dh & 'n' vertical ones in 'ptr_dv' for the plot widget in the uniform mode.

The values are computed from the distance 'so' from the start of 
the element 'io' of the general_record in the data bush to the distance 'sl'
from the start of the element 'il'.

'ptrTwiss0' is a pointer to a twiss_record containing ( please NOTE ) the 
initial twiss parameters at the BEGINNING of element 'io'.

'fnom' is a boolean flag which set to the value TRUE ( alias 1 in C )
will indicate that the optics are the nominal ones, i.e. the insertion
devices are treated as drifts. If 'fnom' is set to the value FALSE, the
linear effects of the insertion devices will be considered.

Please note that the computation is progressive, i.e. it goes from 'io' to
'il', which may be less than 'io'.

At the present moment, only the optics for the Storage Ring may be computed,
for certain elements are missing.
The ones for the Transfer Line will be implemented as soon as the insertion
device compensation program is finished, unless a strongly motivated request
comes from one of the HLS implementors.

The program has been debugged using the structure file of the SR and it can be
said that it is 90% bug-free. However it would be highly appreciated if whoever
begins using it could check if the values obtained are correct. 

                                           Lidia
*/                                           


/*
 *  Include files 
*/
#include <stdio.h>
#include <math.h>
#include "hlsvar.h"
#include "hlsfun.h"
#define PI acos(-1.0)


/*  Internal Functions  */


void PlT_Optics(double [2], double [2],double [2], double [2],
double [2], double [2], double [2], double [2], double [3][3],
            double [3][3]);
void PlT_Transfer_Matrix(int, int, double, double [3][3], double [3][3]);
void PlT_Construct_Matrix(double [3][3], double [3][3], double [4][4]); 
void PlT_Mul_Mat (int, double [3][3], double [3][3], double [3][3]);      


int plot_twiss( ptrTwiss0, io, so, il, sl, fnom, n, ptr_bh, ptr_bv, ptr_dh,ptr_dv)
  twiss_record *ptrTwiss0;
  int io, il, n, fnom;
  double so, sl;
  double ptr_bh[], ptr_bv[];
  double ptr_dh[], ptr_dv[];
  {
    double bn[2], di[2], dn[2];
    double bo[2], ao[2], an[2];
    double d1o[2], d1n[2];
    int i, j, k, lim, type, subtype;
    double mh[3][3], mv[3][3];
    double total_s, s, partial_s, ds;
    double length,length1,strength,rho,id_k;
    char c;

/*  checking errors  */    
    if ( so<0.0 || sl<0.0 ) return(1);
    if ( ge[io]->type == drift_type) partial_s=ge[io]->p.dr->length;
    if ( ge[io]->type == quad_type) partial_s=ge[io]->p.qu->magnetic_length;
    if ( ge[io]->type == bend_type) partial_s=ge[io]->p.be->magnetic_length;
    if ( ge[io]->type == septum_type)
                           partial_s=ge[io]->p.sp->pAcc->pPar->in.length;
    if ( ge[io]->type == ID_type) partial_s=ge[io]->p.id->pPar->id.Np*
                                             ge[io]->p.id->pPar->id.period;
    if ( ge[io]->type > thin_type) partial_s=0.0;
    if ( so > partial_s ) return(2);    

    if ( ge[il]->type == drift_type) partial_s=ge[il]->p.dr->length;
    if ( ge[il]->type == quad_type) partial_s=ge[il]->p.qu->magnetic_length;
    if ( ge[il]->type == bend_type) partial_s=ge[il]->p.be->magnetic_length;
    if ( ge[il]->type == ID_type) partial_s=ge[il]->p.id->pPar->id.Np*
                                             ge[il]->p.id->pPar->id.period;
    if ( ge[il]->type == septum_type)
                           partial_s=ge[il]->p.sp->pAcc->pPar->in.length;
    if ( ge[il]->type > thin_type) partial_s=0.0;
    if ( sl > partial_s ) return(2);    
    
/* get total distance  */
    total_s=0.0;
    lim=(N_general-io+il)%N_general;
    
    for ( i=0; i<lim; i++) 
    {
      int i1;
      
      i1=(i+io)%N_general;
      if ( ge[i1]->type == drift_type) partial_s=ge[i1]->p.dr->length;
      if ( ge[i1]->type == quad_type) partial_s=ge[i1]->p.qu->magnetic_length;
      if ( ge[i1]->type == bend_type) partial_s=ge[i1]->p.be->magnetic_length;
      if ( ge[i1]->type == ID_type) partial_s=ge[i1]->p.id->pPar->id.Np*
                                             ge[i1]->p.id->pPar->id.period;
      if ( ge[i1]->type == septum_type)
                           partial_s=ge[i1]->p.sp->pAcc->pPar->in.length;
      if ( ge[i1]->type > thin_type) partial_s=0.0;
      
      
      total_s = total_s + partial_s;
    }
    total_s = total_s - so + sl;
    
    ds = total_s/(n-1);

/*  initialize twiss parameters and transfer matrices */    
    for ( i=0; i<2; i++)
    {
      bo[i] = ptrTwiss0->beta[i];
      ao[i] = ptrTwiss0->alpha[i];
      di[i] = ptrTwiss0->D[i];
      d1o[i] = ptrTwiss0->Dprime[i];
    }
    for ( i=0; i<3; i++)
    {
      for ( j=0; j<3; j++)
      {
        mh[i][j]=0.0;
        mv[i][j]=0.0;
        if ( i == j )
        {
          mh[i][j]=1.0;
          mv[i][j]=1.0;
        }
      }
    }
      s=0.0;
      subtype=0;
      if ( ge[io]->type == drift_type )
        {
          subtype=0;
          s=ge[io]->p.dr->length;
        }
      if ( ge[io]->type == quad_type )
        {
          subtype=0;
          s=ge[io]->p.qu->magnetic_length;
        }
      if ( ge[io]->type == bend_type )
        {
          subtype = -ge[io]->p.be->sub_type;
          if ( !subtype ) subtype = -100;
          s=ge[io]->p.be->magnetic_length;
        }
      if ( ge[io]->type == ID_type )
        {
          subtype = fnom;
          s=ge[io]->p.id->pPar->id.Np*ge[io]->p.id->pPar->id.period;
        }
      if ( ge[io]->type == septum_type)
       {
          s=ge[io]->p.sp->pAcc->pPar->in.length;
          subtype = 0;
       }
        length=so;
        PlT_Transfer_Matrix(io, subtype, length, mh, mv);
      PlT_Optics( bo, ao, di, d1o, bn, an, dn, d1n, mh, mv);
      ptr_bh[0] = bn[0];
      ptr_bv[0] = bn[1];
      ptr_dh[0] = dn[0];
      ptr_dv[0] = dn[1];
      k=io;
      partial_s = so;

/*  loop over all points to get beta values  */
      for ( i=1; i< n; i++ )
      {
      
        length=length+ds;
        partial_s=partial_s+ds;
        if ( length > s )
        {
 toolong:
          if ( ge[k]->type == bend_type ) subtype = ge[k]->p.be->sub_type;
          if ( ge[k]->type == ID_type ) subtype = fnom;
          PlT_Transfer_Matrix( k, subtype, s, mh, mv);
          PlT_Optics( bo, ao, di, d1o, bn, an, dn, d1n, mh, mv);
          bo[0]=bn[0];
          bo[1]=bn[1];
          ao[0]=an[0];
          ao[1]=an[1];
          di[0]=dn[0];
          di[1]=dn[1];
          d1o[0]=d1n[0];
          d1o[1]=d1n[1];

          k=(k+1)%N_general;
          length=length-s;
      
          if ( ge[k]->type > thin_type ) s=0.0;
          if ( ge[k]->type < thin_type )
          {
            if ( ge[k]->type == drift_type )
            {
              subtype=0;
              s=ge[k]->p.dr->length;
            }
            if ( ge[k]->type == quad_type )
            {
              subtype=0;
              s=ge[k]->p.qu->magnetic_length;
            }
            if ( ge[k]->type == bend_type )
            {
              subtype = -ge[k]->p.be->sub_type;
              if ( !subtype ) subtype = -100;
              s=ge[k]->p.be->magnetic_length;
            }
           if ( ge[k]->type == ID_type )
           {
              subtype = fnom;
              s=ge[k]->p.id->pPar->id.Np*ge[k]->p.id->pPar->id.period;
           }
           if ( ge[k]->type == septum_type)
           {
              s=ge[k]->p.sp->pAcc->pPar->in.length;
              subtype = 0;
           }
         }
         if ( length > s ) goto toolong;
         
      }   
      
      PlT_Transfer_Matrix(k, subtype, length, mh, mv);
      
      PlT_Optics( bo, ao, di, d1o, bn, an, dn, d1n, mh, mv);
      ptr_bh[i] = bn[0];
      ptr_bv[i] = bn[1];
      ptr_dh[i] = dn[0];
      ptr_dv[i] = dn[1];
    }   
    return(0);
  }
    
    


void PlT_Transfer_Matrix(i, sub_type, length, mh, mv)            
  int i, sub_type;
  double mh[3][3], mv[3][3];
  double length;
  {  
    int type, j, k;
    double length1,strength,rho,id_k;
    
    for ( j=0; j<3; j++)
    {
      for ( k=0; k<3; k++)
      {
        mh[j][k]=mv[j][k]=0.0;
        if ( j==k ) mh[j][k]=mv[j][k]=1.0;
      }
    }
    
    type = ge[i]->type;
    if ( type > thin_type )
    {
       mh[0][0] = 1.0;
       mh[0][1] = 0.0;        
       mh[1][0] = 0.0;
       mh[1][1] = 1.0;
       mv[0][0] = 1.0;
       mv[0][1] = 0.0;        
       mv[1][0] = 0.0;
       mv[1][1] = 1.0;
     }
    if ( type == drift_type )
    {
       mh[0][0] = 1.0;
       mh[0][1] = length;        
       mh[1][0] = 0.0;
       mh[1][1] = 1.0;
       mv[0][0] = 1.0;
       mv[0][1] = length;        
       mv[1][0] = 0.0;
       mv[1][1] = 1.0;
     }
    if ( type == quad_type )
    {
       strength=ge[i]->p.qu->strength;
       if ( strength < 0.0 )
       {
         mh[0][0] = cos( sqrt( fabs( strength)) * length);
         mh[0][1] = sin( sqrt( fabs( strength)) * length);
         mh[1][0] = mh[0][1];
         mh[0][1] = mh[0][1]/sqrt( fabs( strength));
         mh[1][0] = -mh[1][0]*sqrt( fabs( strength));   
         mh[1][1] = mh[0][0];
         
         mv[0][0] = cosh( sqrt( fabs( strength)) * length);
         mv[0][1] = sinh( sqrt( fabs( strength)) * length);
         mv[1][0] = mv[0][1];
         mv[0][1] = mv[0][1]/sqrt( fabs( strength));
         mv[1][0] = mv[1][0]*sqrt( fabs( strength));
         mv[1][1] = mv[0][0];
       }
       if ( strength > 0.0 )
       {
         mh[0][0] = cosh( sqrt( fabs( strength)) * length);
         mh[0][1] = sinh( sqrt( fabs( strength)) * length);
         mh[1][0] = mh[0][1];
         mh[0][1] = mh[0][1]/sqrt( fabs( strength));
         mh[1][0] = mh[1][0]*sqrt( fabs( strength));   
         mh[1][1] = mh[0][0];
         
         mv[0][0] = cos( sqrt( fabs( strength)) * length);
         mv[0][1] = sin( sqrt( fabs( strength)) * length);
         mv[1][0] = mv[0][1];
         mv[0][1] = mv[0][1]/sqrt( fabs( strength));
         mv[1][0] = -mv[1][0]*sqrt( fabs( strength));
         mv[1][1] = mv[0][0];
       }
       if ( strength == 0.0 )
       {
         mh[0][0] = 1.0;
         mh[0][1] = length;        
         mh[1][0] = 0.0;
         mh[1][1] = 1.0;
         mv[0][0] = 1.0;
         mv[0][1] = length;        
         mv[1][0] = 0.0;
         mv[1][1] = 1.0;
       }
     }
    if ( type == bend_type )
    {
      double angle;
      int sub;
      
      angle = PI * ge[i]->p.be->pPar->ps.angle /180.0;
      sub = ge[i]->p.be->sub_type;
      if ( sub >= reversed_polarity )
        angle = -angle;
      sub = sub%reversed_polarity;
      if ( angle == 0.0 ) 
      {     
        mh[0][0] = 1.0;
        mh[0][1] = length;        
        mh[1][0] = 0.0;
        mh[1][1] = 1.0;
        mv[0][0] = 1.0;
        mv[0][1] = length;        
        mv[1][0] = 0.0;
        mv[1][1] = 1.0;
        sub = 1000;
      }
      rho =  angle / ge[i]->p.be->magnetic_length;
      
      if ( sub == com_rectangle || sub == com_sector )
      {
        double kay;
        double  kayx;
        double m1[3][3], m2[3][3], m3[3][3];
        int l1,l2;
        char c;
        
        for ( l1=0; l1<3; l1++ )
        {
          for ( l2=0; l2<3; l2++ )
          {
            m1[l1][l2] = 0.0;
            m2[l1][l2] = 0.0;
            m3[l1][l2] = 0.0;
            if ( l1==l2 )
            {
              m1[l1][l2]=1.0;
              m2[l1][l2]=1.0;
              m3[l1][l2]=1.0;
            }
          }
        }
          
          length1 = ge[i]->p.be->magnetic_length;
          strength = ge[i]->p.be->strength * rho*rho;
        
        kay=strength;
        kayx = strength-rho*rho;
        
        m1[0][0] = 1.0;
        m1[0][1] = 0.0;
        if ( sub == com_rectangle )
          m1[1][0] = rho*tan(length1*rho/2.0);
        m1[1][1] = m1[0][0]; 

        if ( kayx < 0.0 )
        {        
          m2[0][0] = cos( sqrt( fabs( kayx)) * length);
          m2[0][1] = sin( sqrt( fabs( kayx)) * length);
          m2[1][0] = m2[0][1];
          m2[0][1] = m2[0][1]/sqrt( fabs( kayx));
          m2[1][0] = -m2[1][0]*sqrt( fabs( kayx));
          m2[1][1] = m2[0][0];
          m2[0][2] = rho*(1.0-cos( sqrt( fabs( kayx)) * length))/ fabs( kayx);
          m2[1][2] = rho*sin( sqrt( fabs( kayx)) * length)/sqrt( fabs( kayx));
        
          PlT_Mul_Mat(3, m1, m2, m3);
          if ( sub_type < 0 )
          {     
            for ( l1=0; l1<3; l1++ )
            {
              for ( l2=0; l2<3; l2++ )
              {
                mh[l1][l2]=m3[l1][l2];
              }
            }
          }
          else
          {
            PlT_Mul_Mat(3, m3, m1, m2);
            for ( l1=0; l1<3; l1++ )
            {
              for ( l2=0; l2<3; l2++ )
              {
                mh[l1][l2]=m2[l1][l2];
              }
            }
          }
        }
        if ( kayx > 0.0 )
        {        
          m2[0][0] = cosh( sqrt( fabs( kayx)) * length);
          m2[0][1] = sinh( sqrt( fabs( kayx)) * length);
          m2[1][0] = m2[0][1];
          m2[0][1] = m2[0][1]/sqrt( fabs( kayx));
          m2[1][0] = m2[1][0]*sqrt( fabs( kayx));
          m2[1][1] = m2[0][0];
          m2[0][2] = -rho*(1.0-cosh( sqrt( fabs( kayx)) * length))/ fabs( kayx);
          m2[1][2] = rho*sinh( sqrt( fabs( kayx)) * length)/sqrt( fabs( kayx));
        
          PlT_Mul_Mat(3, m1, m2, m3);
          if ( sub_type < 0 )
          {
            for ( l1=0; l1<3; l1++ )
            {
              for ( l2=0; l2<3; l2++ )
              {
                mh[l1][l2]=m3[l1][l2];
              }
            }
          }
          else
          {
          PlT_Mul_Mat(3, m3, m1, m2);
            for ( l1=0; l1<3; l1++ )
            {
              for ( l2=0; l2<3; l2++ )
              {
                mh[l1][l2]=m2[l1][l2];
              }
            }
          }
        }
        m1[1][0] = -m1[1][0];
        
        if ( kay <= 0.0 )
        {        
          m2[0][0] = cosh( sqrt( fabs( kay)) * length);
          m2[0][1] = sinh( sqrt( fabs( kay)) * length);
          m2[1][0] = m2[0][1];
          if ( kay != 0.0 )
            m2[0][1] = m2[0][1]/sqrt( fabs( kay));
          else
            m2[0][1] = length;
          m2[1][0] = m2[1][0]*sqrt( fabs( kay));
          m2[1][1] = m2[0][0];
          m2[0][2] = 0.0;
          m2[1][2] = 0.0;
        
          PlT_Mul_Mat(3, m1, m2, m3);
          if ( sub_type < 0 )
          {
            for ( l1=0; l1<3; l1++ )
            {
              for ( l2=0; l2<3; l2++ )
              {
              mv[l1][l2]=m3[l1][l2];
              }
            }
          }
          else
          {
           PlT_Mul_Mat(3, m3, m1, m2);
            for ( l1=0; l1<3; l1++ )
            {
              for ( l2=0; l2<3; l2++ )
              {
              mv[l1][l2]=m2[l1][l2];
              }
            }
          }
        }

        if ( kay > 0.0 )
        {        
          m2[0][0] = cos( sqrt( fabs( kay)) * length);
          m2[0][1] = sin( sqrt( fabs( kay)) * length);
          m2[1][0] = m2[0][1];
          m2[0][1] = m2[0][1]/sqrt( fabs( kay));
          m2[1][0] = -m2[1][0]*sqrt( fabs( kay));
          m2[1][1] = m2[0][0];
          m2[0][2] = 0.0;
          m2[1][2] = 0.0;
        
          PlT_Mul_Mat(3, m1, m2, m3);
          if ( sub_type < 0 )
          {
            for ( l1=0; l1<3; l1++ )
            {
              for ( l2=0; l2<3; l2++ )
              {
                mv[l1][l2]=m3[l1][l2];
              }
            }
          }
          else
          {
            PlT_Mul_Mat(3, m3, m1, m2);
            for ( l1=0; l1<3; l1++ )
            {
              for ( l2=0; l2<3; l2++ )
              {
                mv[l1][l2]=m2[l1][l2];
              }
            }
          }
        }
        
      }
      if ( sub == hor_rectangle )
      {
        double tn;
        
        angle = length*rho;
        
        tn = tan(angle/2.0);
        mh[0][1] = sin(angle)/rho;
        mh[0][2] = (1.0 - cos(angle))/rho;
        mh[1][2] = 2.0 * tn;
        mv[0][0] = 1.0 - angle*tn;
        mv[0][1] = length;
        mv[1][0] = -tn*rho*(2.0 - angle*tn);
        mv[1][1] = mv[0][0];
      }
      if ( sub == hor_sector )
      {
        double cs, sn;
        
        angle = length*rho;
        
        cs = cos(angle);
        sn = sin(angle);
        mh[0][0] = cs;
        mh[0][1] = sn/rho;
        mh[0][2] = (1.0 - cs)/rho;
        mh[1][0] = -sn*rho;
        mh[1][1] = mh[0][0];
        mh[1][2] = sn;
        mv[0][1] = length;
      }
      if ( sub == ver_rectangle )
      {
        double cs, sn, tn;
        
        angle = length*rho;
        
        cs = cos(angle);
        sn = sin(angle);
        tn = tan(angle/2.0);
        mh[0][0] = 1.0 - angle*tn;
        mh[0][1] = length;
        mh[1][0] = -tn*rho*(2.0 - angle*tn);
        mh[1][1] = mh[0][0];
        mv[0][1] = sn/rho;
        mv[0][2] = (1.0 - cs)/rho;
        mv[1][2] = 2.0*tn;
      }
      if ( sub == ver_sector )
      {
        double cs, sn;
        
        angle = length*rho;
        
        cs = cos(angle);
        sn = sin(angle);
        mh[0][1] = length;
        mv[0][0] = cs;
        mv[0][1] = sn/rho;
        mv[0][2] = (1.0 - cs)/rho;
        mv[1][0] = -sn*rho;
        mv[1][1] = cs;
        mv[1][2] = sn;
      }
      if ( sub == hor_edge )
      {
        mh[1][0] = tan(angle/2.0)*rho;;
        mv[1][0] = -mh[1][0];
      }
      if ( sub == ver_edge )
      {
        mv[1][0] = tan(angle/2.0)*rho;;
        mh[1][0] = -mh[1][0];
      }
      
    }
    if ( type == septum_type )
    {
      double angle, cs, sn, tn;
      
      if ( ge[i]->p.sp->curvature == 0.0 )
      {
        mh[0][1] = length;
        mv[0][1] = length;
      }
      else
      {
        rho = 1.0/ge[i]->p.sp->curvature;
        angle = length/rho;
        cs = cos(angle);
        sn = sin(angle);
        tn = tan(angle/2.0);
        mh[0][1] = rho*sn;
        mh[0][2] = (1.0 - cs)*rho;
        mh[1][2] = 2.0*tn;
        mv[0][0] = 1.0-angle*tn;
        mv[0][1] = length;
        mv[1][0] = -tn/rho*(2.0-angle*tn);
        mv[1][1] = 1.0-angle*tn;
      }
        
    }
    if ( type == ID_type )
    {
      int kkx, kky, kkxh, kkyh;
      double kx, ky, kxh, kyh, kayx, kayy, rhoh, k, kay;  /* ID variables */
      
      ky = ge[i]->p.id->k_y;
      kx = ge[i]->p.id->k_x;
      k = 2*PI/ge[i]->p.id->pPar->id.period;
      rho = ge[i]->p.id->curvature;
      
       kxh = ge[i]->p.id->kh_x;
        kyh = ge[i]->p.id->kh_y;
        rhoh = ge[i]->p.id->curvature_h;
        
        if ( (kx <= 0.0) && (ky <= 0.0) ) {  kkx=-1; kky=-1; }
        if ( (kx <= 0.0) && (ky > 0.0) ) {  kkx=-1; kky=1; }
        if (( kx > 0.0 ) && (ky <= 0.0) ) {  kkx=1; kky=-1;}
        if (( kx > 0.0 ) && (ky > 0.0) ) {  kkx=1; kky=1;}

        if ( (kyh <= 0.0) && (kxh <= 0.0) ) { kkyh=-1; kkxh=-1; }
        if ( (kyh <= 0.0) && (kxh > 0.0) ) {  kkyh=-1; kkxh=1; }
        if ( (kyh > 0.0 ) && (kxh <= 0.0) ) { kkyh=1; kkxh=-1;}
        if ( (kyh > 0.0 ) && (kxh > 0.0) ) {  kkyh=1; kkxh=1;}
        
        kayx = kkx*kx*kx*rho*rho + kkxh*kxh*kxh*rhoh*rhoh;
        if ( kayx <=0.0 ) kayx=-sqrt(fabs(kayx));
        if ( kayx > 0.0 ) kayx=sqrt(fabs(kayx));
        
        kayy = kky*ky*ky*rho*rho + kkyh*kyh*kyh*rhoh*rhoh;
        if ( kayy <=0.0 ) kayy=-sqrt(fabs(kayy));
        if ( kayy > 0.0 ) kayy=sqrt(fabs(kayy));
      
      
      if ( sub_type )
      {
        kayx=0.0;
        kayy=0.0;
      }
      if ( kayx == 0.0 )
      {
        mh[0][0] = 1.0;
        mh[0][1] = length;        
        mh[1][0] = 0.0;
        mh[1][1] = 1.0;
      }
      if ( kayx < 0.0 )
      {
        kay = fabs( kayx/(sqrt(2.0)*k) );
        
        mh[0][0] = cos( kay * length );
        mh[0][1] = sin( kay * length);
        mh[1][0] = mh[0][1];
        mh[0][1] = mh[0][1]/kay;
        mh[1][0] = -mh[1][0]*kay;   
        mh[1][1] = mh[0][0];
      }
      if ( kayx > 0.0 )
      {
        kay = kayx/(sqrt(2.0)*k);
        
        mh[0][0] = cosh( kay * length );
        mh[0][1] = sinh( kay * length);
        mh[1][0] = mh[0][1];
        mh[0][1] = mh[0][1]/kay;
        mh[1][0] = mh[1][0]*kay;   
        mh[1][1] = mh[0][0];
      }
      
      if ( kayy < 0.0 )
      {
       kay = fabs(kayy)/(sqrt(2.0)*k);

       mv[0][0] = cos( kay * length );
       mv[0][1] = sin( kay * length );
       mv[1][0] = mv[0][1];
       mv[0][1] = mv[0][1]/kay;
       mv[1][0] = -mv[1][0]*kay;   
       mv[1][1] = mv[0][0];
      }
      if ( kayy > 0.0 )
      {
       kay = fabs(kayy)/(sqrt(2.0)*k);

       mv[0][0] = cosh( kay * length );
       mv[0][1] = sinh( kay * length );
       mv[1][0] = mv[0][1];
       mv[0][1] = mv[0][1]/kay;
       mv[1][0] = mv[1][0]*kay;   
       mv[1][1] = mv[0][0];
      }
       
      if ( kayy == 0.0 )
      {
        mv[0][0] = 1.0;
        mv[0][1] = length;        
        mv[1][0] = 0.0;
        mv[1][1] = 1.0;
      }
     }
   }


void PlT_Optics( bo, ao, di, d1o, bn, an, dn, d1n, mh, mv)
  double bo[2], ao[2], di[2],d1o[2], bn[2], an[2], dn[2], d1n[2];
  double mh[3][3],mv[3][3];
  {
    double m[4][4];
    double go;
    double c, s, c1, s1;
    int j, k;
    
    PlT_Construct_Matrix( mh, mv, m );
    
    k=0;   
    for ( j=0; j<2; j++ )
    {
      go = (1.0+ao[j]*ao[j])/bo[j];
      
      c = m[j+k][j+k];
      s = m[j+k][j+k+1];
      c1 = m[j+k+1][j+k];
      s1 = m[j+k+1][j+k+1];
       
      bn[j] = c*c*bo[j]-2*ao[j]*c*s+s*s*go;
      an[j] = -c*c1*bo[j]+(c1*s+c*s1)*ao[j]-go*s*s1;
      if (!k)
      {
        dn[j] = c*di[j]+s*d1o[j]+ mh[0][2] ; 
        d1n[j] = c1*di[j]+s1*d1o[j]+ mh[1][2]; 
      } 
      if (k)
      {
        dn[j] = c*di[j]+s*d1o[j]+ mv[0][2] ; 
        d1n[j] = c1*di[j]+s1*d1o[j]+ mv[1][2]; 
      } 
      k=1;
    }
  }
  
  
void PlT_Construct_Matrix( t1, t2, t3 )
  double t1[3][3], t2[3][3], t3[4][4];
  {
    int i, j;
    
    for ( i=0; i<4; i++ )
      {
       for ( j=0; j<4; j++ )   
         t3[i][j] = 0.;
      }
      
    
    for ( i=0; i<2; i++ )
      {
       for ( j=0; j<2; j++ )   
       {
         t3[i][j] = t1[i][j];
         t3[i+2][j+2] = t2[i][j];
       }
      }
  }


void PlT_Mul_Mat (n, t1, t2, t3)
  int n;
  double t1[3][3],t2[3][3],t3[3][3];
  {
  
    int i, j, k;
    for ( i=0; i<n; i++)
    {
      for ( j=0; (j<n); j++)
      {
        t3[i][j] = 0.0;
        for ( k=0; k<n; k++)
          t3[i][j] = t3[i][j] + t1[i][k] * t2[k][j];
       }
     }
   }  
  
