/*==================================================================
                     	+---------------+
 			|     rpc.c     |
                     	+---------------+

 The set of routines which deal with RPCs directly. Those routines
 should be called instead of direct RPC routines. New errors in
 do_rpc: -5 means that the service number does not correspond to
 an existent service.; -6 means that the service requested differs 
 from the service that imported the channel.
 
 First version 1.0. 			(Mark Plesko 11.5.93)
 
 Ignoring the services bpm and bump, as they are not yet supported
 by the control group.
 1.1					(M.P. 19.5.93)
 
 A routine which returns a string with rpc errtype/number and string
 is added (rpc_error_message).
 1.2					(M.P. 22.5.93)
 
 Modified the call to bpm_all, due to new specs from M.Lonza (29.6.93)
 and changed the bpm.x_s and bpm.y_s variables to bpm.s_x and bpm.s_y
 (accordingly in hlsrpc.h).
 inj and bpm service supported, bump service still not supported. The
 bump service is excluded by having dummy routines which exit the
 program immediately.
 1.3					(M.P. 11.8.93)

 Fixed an error in calling the injection service routines. The mode
 was taken to be the mode passed to do_rpc, instead of the proper
 mode as defined in Lonza's note (i.e. "RAR", "WTIPOL", etc.).
 1.4					(M.P. 22.9.93)
 
 Added a few lpc routines for CJB: WS, WR, RS, RRI.
 1.5					(M.P. 28.9.93)
 
 Added lpc routine lpc_WM, which is needed to read the anular electrode.
 1.6					(M.P. 7.10.93)
 
 Added handling of the save/restore server (sare) for the Machine
 File and other tasks. The bump service is now made available.
 THe cyc server has now WTIPOL (uppercase) for the mode.
 1.7					(M.P. 26.10.93)
 
 The cyc server uses now the correct parameters: family,member,datum.
 The note by M.Lonza has the wrong description of parameters !
 Therefore the mode for the cycling has been changed back to lowercase
 (wtipol).
 1.8					(M.P. 2.11.93)
 
 Fixed an error for service_inj, where pList->lpc was used instead
 of pList->inj ! Considers that inj routines return also positive
 error messages.
 1.9					(M.P. 23.2.94)
 
 Added modes WARI and RFI to lpc service.
 1.10					(M.P. 10.3.94)
 
 Added mode WAIR (which is a misspelling of WAIR) as an alternative
 to the WARI mode.
 1.11					(M.P. 7.11.94)
 
 Added mode RBI, which is needed to read the RF mode (sync or async).
 1.12					(M.P. 19.12.94)
 
 Added mode WPI to set the brakes of the IDs.
 1.13					(M.P. 16.1.95)

 Added mode WBI to set the loops of the IDs.
 1.14					(M.P. 6.2.95)
 
 Added new service flsc to read the FS and SRPM image.
 1.15					(M.P. 21.3.95)
 
 Added new service info to get the beam current, lifetime, inj. rate
 and Energy from the new server of Surace.
 1.16					(M.P. 21.11.95)
 
 Added security check and clearance for processes running on control
 room machines, displays and accounts. Simulation input is recognized
 by its sid (== INT_MAX) and not check for security.
 2.0					(M.P. 29.11.95)
 
 Removed the check for the right display, as DISPLAY sometimes is not
 defined on the control room xterms. Also dino removed from list of 
 authorized hosts and displays.
 2.1					(M.P. 11.12.95)
 
 Removed complete security check, on request of CJB and LT
 2.2					(M.P. 26.2.96)
 ==================================================================*/

#include <stdio.h>
#include <limits.h>
#include "hlsvar.h"
#include "hlsrpc.h"

/* global variable only in this file: stores the sid of the last bind,
   to avoid binding the same sid in two consecutive calls */
   
static	int	previous_sid = -1 ;
static	int	security_cleared = 0 ;

/* added 16-jan-2017 by F. Iazzourene */
static	int	previous_sid_fi = -1 ;
/* end added*/

/* strings which identify the service */
str6	service_name[] = {"DUMMY","lpc","cyc","bpm","bump",
			"inj","sare", "flsc", "info" } ; 


/*========================================================================*/
/*security routines, performing checks and clearance                      */
/*========================================================================*/

int	security_alert ()
{
  typedef char names[32];

  names	hosts[] = { "barney", "wilma", "fred", "betty" } ;
  names	users[] = { "machine", "elettra" } ;
  names	displays[] = { "barney", "wilma", "fred", "betty" } ;

  names	host, user, display ;
  char	*ch ;
  int	i ;
  int	wrong_host, wrong_user, wrong_display ;
  
  strcpy ( host, getenv ("HOST") ) ;
  strcpy ( user, getenv ("USER") ) ;
  strcpy ( display, getenv ("DISPLAY") ) ;
  
  /* convert to device name, without :0.0 appendix */
  ch = display ;
  while ( *ch ) if ( *(++ch) == ':' ) *ch = '\0' ;
    
  /* perform checks */

  wrong_host = 1 ;
  for (i=0; i<sizeof hosts/sizeof(names); i++) 
    if ( !strcmp ( hosts[i], host ) ) { wrong_host = 0 ; break ; }
  
  wrong_user = 1 ;
  for (i=0; i<sizeof users/sizeof(names); i++) 
    if ( !strcmp ( users[i], user ) ) { wrong_user = 0 ; break ; }
  
  wrong_display = 1 ;
  for (i=0; i<sizeof displays/sizeof(names); i++) 
    if ( !strcmp ( displays[i], display ) ) { wrong_display = 0 ; break ; }
  
  if ( wrong_host ) printf ("Unauthorized host: %s\n", host ) ;
  if ( wrong_user ) printf ("Unauthorized user: %s\n", user ) ;

/***************************************************************************
outcommented on request of L.T. (11.12.95)
--------------------------------
  if ( wrong_display ) printf ("Unauthorized display: %s\n", display ) ;
  
  if ( wrong_host || wrong_user || wrong_display ) return ( 1 ) ;
*****************************************************************************/ 
 
  if ( wrong_host || wrong_user ) return ( 1 ) ;

  else
  {
    security_cleared = 1 ;
    return ( 0 ) ;
  }
   
} /* security_alert */


int	request_clearance ()

{
  char	password[80] ;
  
  printf ("You're about to set the machine from an unauthorized place\n") ;
  printf ("If you want to really set, then say that you want to\n") ;
  printf ("otherwise press RETURN to return a RPC error.\n") ;
  printf ("All your actions will be monitored and logged !\n" ) ;
  printf ("Password:") ;
  fflush ( stdout ) ;
  fgets ( password, 79, stdin ) ;
  if ( !strcmp ( password,"that you want to\n" ) )
  {
    printf ("Password accepted, this and further setting is allowed.\n") ;
    security_cleared = 1 ;
    return ( 1 ) ;
  }
  else
  {
    printf ("A wise decision - machine will not be set.\n") ;
    return ( 0 ) ;
  }
} /* request_clearance */


int	check_security ( int service, char *mode, RPC_list *pList )

{
  char	*ch ;
  
  switch ( service ) 
  {
    case service_lpc: 
      if ( *mode == 'W' ) if ( security_alert() )
      {
        printf ("Security check " ) ; 
        printf("for lpc %s_%s_%s_%s\n",pList->lpc.family,pList->lpc.member,pList->lpc.action,mode);
        return ( request_clearance() ) ;
      }
      return ( 1 ) ;
    break;
    
    case service_cyc: 
      if ( security_alert() )
      {
        printf ("Security check " ) ; 
        printf("for cyc %s_%s_%s\n",pList->cyc.family,pList->cyc.member,mode);
        return ( request_clearance() ) ;
      }
      return ( 1 ) ;
    break;
    
    case service_bpm:
      if ( !strcmp ( mode, "wn" ) ) if ( security_alert() )
      {
        printf ("Security check " ) ; 
        printf("for bpm_%s\n",mode);
        return ( request_clearance() ) ;
      }
      return ( 1 ) ;
    break;
    
    case service_bump:
      if ( !strcmp ( mode, "check" ) ) return ( 1 ) ;
      if ( !strcmp ( mode, "read" ) ) return ( 1 ) ;
      if ( security_alert() )
      {
        printf ("Security check " ) ; 
        printf("for bump_%s\n",mode);
        return ( request_clearance() ) ;
      }
      return ( 1 ) ;
    break;
    
    case service_inj:  
    case service_sare: 
      ch = mode ;
      while ( *ch ) 
        if ( *(ch++) == '_' ) if ( *ch == 'w' ) if ( security_alert() )
        {
          printf ("Security check " ) ; 
          if (service==service_sare) printf("for sare_%s\n",mode);
          else printf("for inj%s %s_%s_%s\n",mode,pList->inj.family,pList->inj.member,pList->inj.action);
          return ( request_clearance() ) ;
        }
      return ( 1 ) ;
    break;
    
    case service_flsc:
    case service_info: 
      return ( 1 ) ; /* no check necessary */
    break;
    
    default:
      /* for forgotten modes always check security */
      if ( security_alert() )
      {
        printf ("Security check " ) ; 
        printf("for service#%i, mode=%s\n", service, mode);
        return ( request_clearance() ) ;
      }
      return ( 1 ) ;
      
  } /* switch */
} /* check_security */


/*========================================================================*/
/*rpc routines, calling the real rpc's (for hackers only !)               */
/*========================================================================*/

int do_rpc_err(host_record *pHost, int service, char *mode, RPC_list *pList )
/*-----------------------------------------------------------------------------\
|                                                                              |
|   The routine which allows interactively to repeat a RPC request thus        |
|   enabling the operator to fix some problems during program execution        |
|   and not having to restart the program. Works in terminal mode only,        |
|   but can be used also while having other panels with widgets.               |
|                                                                              |
|   In case of an RPC failure, the routine requests to either Retry the RPC,   |
|   Ignore the error and just continue, or Abort the whole program. The        |
|   operator types one of the letters R, I or A (upper or lowercase).          |
|                                                                              |
|   V1.0                                 (Mark Plesko, 11.2.93)                |
|                                                                              |
\-----------------------------------------------------------------------------*/
{
int c,c1, err ;

  if ( (err=do_rpc(pHost,service,mode,pList)) >= 0 ) return ( err ) ;
  
  /* go into retry/ignore/abort loop if first try was unsuccessful */
  while ( 1 ) 
  {
    printf ( "\n(R)etry, (I)gnore, (A)bort ?" ) ;
    c = getchar() ; 
    /*flush input buffer */
    if ( c != '\n' ) while ( (c1=getchar()) != '\n' ) ; 
    switch ( c ) 
    {
      case 'a': case 'A': exit ( -1 ) ;    break ;
      case 'i': case 'I': return ( err ) ; break ;
      case 'r': case 'R':
        if ( (err=do_rpc(pHost,service,mode,pList)) >= 0 ) return ( err );
      break ;
    }
  }
} 
        
                                                            
int do_rpc ( host_record *pHost, int service, char *mode, RPC_list *pList )
/*-----------------------------------------------------------------------------\
|                                                                              |
|   General routine to read/write to the field for the injection               |
|   panel and associated routines. The values passed to the routine            |
|   are:                                                                       |
|     The host pointer "pHost"                                                 |
|     The "service" and "mode" of the rpc call                                 |
|     The "pList" which is defined as a pointer to a parameter list            |
|                                                                              |
|   Note strncmp == 0 if the two strings are equal, hence the ! in the if.     |
|   I assume that mode is NULL terminated.                                     |
|                                                                              |
|   Obtained from CJB 1.2.93                                                   |
|                                                                              |
|   Modified for use with the functions for the HLS by Mark Plesko. Error      |
|   handling added: returns -2 if import_xxx was unsuccessful, -1 if mode      |
|   not handled by do_rpc. If sid not set, then first import is performed.     |
|   The option to write to a machine file via channel was deleted. The mode    |
|   "RI" was added to the list.                                                |
|   V1.0                                                (M.P. 11.2.93)         |
|                                                                              |
|   All newer changes are documented in the beginning of this file.            |
|                                                                              |
|                                                                              |
\-----------------------------------------------------------------------------*/

{
int res;

 previous_sid_fi = pHost->sid;
 
 /*printf("0) previous_sid_fi = %i service = %i\n", previous_sid_fi,service) ;*/
 
 
  /* no more RPCs after cleanup_rpc() has been called ! */
 /*  if (pHost->sid == -2)  */
   if(previous_sid_fi == -2)
  {
    printf ("@RPC after cleanup_rpc not allowed !\nCall ignored.\n" ) ;
    return ( -4 ) ;
  }
  
  /* check whether there exists a sid value for the requested channel
     and import it if it doesn't */
  /* if ( pHost->sid < 0 ) */
  if(previous_sid_fi < 0)
  {
    switch ( service ) 
    {
      case service_lpc:  pHost->sid = import_lpc (pHost->host,pHost->port);break;
      case service_cyc:  pHost->sid = import_cyc (pHost->host,pHost->port);break;
      case service_bpm:  pHost->sid = import_bpm (pHost->host,pHost->port);break;
      case service_bump: pHost->sid = import_bump(pHost->host,pHost->port);nc_calltmo(pHost->sid,30000);break;
      case service_inj:  pHost->sid = import_inj (pHost->host,pHost->port);break;
      case service_sare: pHost->sid = import_sare(pHost->host,pHost->port);break;
      case service_flsc: pHost->sid = import_flsc(pHost->host,pHost->port);break;
      /* no port number necessary for the beam_info service !!! */
      case service_info: pHost->sid = import_beam_info(pHost->host); printf("sid beam_info %i\n",pHost->sid);break;
default:
        printf ("@Nonexistent service #%i - host:%s port#%i\n", 
               service, pHost->host, pHost->port ) ;
        return ( -5 ) ;
    } 
    if ( pHost->sid < 0 ) 
    {
      printf ("@error %i with service %s at host:%s port#%i\n", pHost->sid, 
               service_name[service], pHost->host, pHost->port ) ;
      nc_perror ( "import_xxx()" ) ;
      return ( -2 ) ;
    }
    else previous_sid = pHost->sid ; /*correct import gives new sid*/
    pHost->service = service ; /* identify service */
  }
  
  else /* channel has been imported previously */
  {
    /* check if the service corresponds to the service first imported */
    if ( service != pHost->service )
    {
      printf("@requested service #%i differs from service # at import: %i\n",
       service, pHost->service ) ; 
      return ( -6 ) ;
    }
  }
  
  /* bind the sid only if previous RPC had a different one - previous_sid
     is declared globally, but only in this file. */
  if ( pHost->sid != previous_sid )
  {
    previous_sid = pHost->sid ;
    switch ( service ) 
    {
      case service_lpc:  bind_lpc  ( previous_sid ) ; break ;
      case service_cyc:  bind_cyc  ( previous_sid ) ; break ;
      case service_bpm:  bind_bpm  ( previous_sid ) ; break ;
      case service_bump: bind_bump ( previous_sid ) ; break ;
      case service_inj:  bind_inj  ( previous_sid ) ; break ;
      case service_sare: bind_sare ( previous_sid ) ; break ;
      case service_flsc: bind_flsc ( previous_sid ) ; break ;
      case service_info: bind_beam_info ( previous_sid ) ; break ;
    }
  }
/*   printf("\n bla \n"); */
 
  /* skip security check on request of LT and CJB */
  security_cleared = 1 ;
  
  /* previous_sid is equal to INT_MAX only when in simulation mode ! */
  if ( previous_sid != INT_MAX ) 
    /* check for security violation */
    if ( !security_cleared ) 
      if ( !check_security (service, mode, pList) ) return ( -7 ) ;
      
 /*      printf("\n bla bla\n"); */

/* Monster switch between services */

  switch ( service ) 
  {
    case service_lpc: 
      /* Select the mode and access the field. */
      if(!strcmp("WR",mode))       res = lpc_WR (pList->lpc.family,pList->lpc.member,pList->lpc.action,mode,pList->lpc.datum);
      else if(!strcmp("WRI",mode)) res = lpc_WRI(pList->lpc.family,pList->lpc.member,pList->lpc.action,mode,pList->lpc.datum);
      else if(!strcmp("WP",mode))  res = lpc_WP (pList->lpc.family,pList->lpc.member,pList->lpc.action,mode,pList->lpc.datum);
      else if(!strcmp("WPI",mode)) res = lpc_WPI(pList->lpc.family,pList->lpc.member,pList->lpc.action,mode,pList->lpc.datum);
      else if(!strcmp("WAR",mode)) res = lpc_WAR(pList->lpc.family,pList->lpc.member,pList->lpc.action,mode,pList->lpc.datum);
      /* the following lpc_WAIR is NOT a typing error !!! */
      else if(!strcmp("WARI",mode))res = lpc_WAIR(pList->lpc.family,pList->lpc.member,pList->lpc.action,mode,pList->lpc.datum);
      else if(!strcmp("WAIR",mode))res = lpc_WAIR(pList->lpc.family,pList->lpc.member,pList->lpc.action,mode,pList->lpc.datum);
      else if(!strcmp("WB",mode))  res = lpc_WB (pList->lpc.family,pList->lpc.member,pList->lpc.action,mode,pList->lpc.datum);
      else if(!strcmp("RE",mode))  res = lpc_RE (pList->lpc.family,pList->lpc.member,pList->lpc.action,mode,pList->lpc.datum);
      else if(!strcmp("WE",mode))  res = lpc_WE (pList->lpc.family,pList->lpc.member,pList->lpc.action,mode,pList->lpc.datum);
      else if(!strcmp("WAS",mode)) res = lpc_WAS(pList->lpc.family,pList->lpc.member,pList->lpc.action,mode,pList->lpc.datum);
      else if(!strcmp("RI",mode))  res = lpc_RI (pList->lpc.family,pList->lpc.member,pList->lpc.action,mode,pList->lpc.datum);
      else if(!strcmp("RR",mode))  res = lpc_RR (pList->lpc.family,pList->lpc.member,pList->lpc.action,mode,pList->lpc.datum);
      else if(!strcmp("RAR",mode)) res = lpc_RAR(pList->lpc.family,pList->lpc.member,pList->lpc.action,mode,pList->lpc.datum);
      else if(!strcmp("RB",mode))  res = lpc_RB (pList->lpc.family,pList->lpc.member,pList->lpc.action,mode,pList->lpc.datum);
      else if(!strcmp("RF",mode))  res = lpc_RF (pList->lpc.family,pList->lpc.member,pList->lpc.action,mode,pList->lpc.datum);
      else if(!strcmp("RFI",mode)) res = lpc_RFI(pList->lpc.family,pList->lpc.member,pList->lpc.action,mode,pList->lpc.datum);
      else if(!strcmp("RRI",mode)) res = lpc_RRI(pList->lpc.family,pList->lpc.member,pList->lpc.action,mode,pList->lpc.datum);
      else if(!strcmp("RBI",mode)) res = lpc_RBI(pList->lpc.family,pList->lpc.member,pList->lpc.action,mode,pList->lpc.datum);
      else if(!strcmp("WBI",mode)) res = lpc_WBI(pList->lpc.family,pList->lpc.member,pList->lpc.action,mode,pList->lpc.datum);
      else if(!strcmp("RS",mode))  res = lpc_RS (pList->lpc.family,pList->lpc.member,pList->lpc.action,mode,pList->lpc.datum);
      else if(!strcmp("WS",mode))  res = lpc_WS (pList->lpc.family,pList->lpc.member,pList->lpc.action,mode,pList->lpc.datum);
      else if(!strcmp("RM",mode))  res = lpc_RM (pList->lpc.family,pList->lpc.member,pList->lpc.action,mode,pList->lpc.datum);
      else if(!strcmp("WM",mode))  res = lpc_WM (pList->lpc.family,pList->lpc.member,pList->lpc.action,mode,pList->lpc.datum);
      else if(!strcmp("WTIPOL",mode))  res = lpc_WTIPOL(pList->lpc.family,pList->lpc.member,pList->lpc.action,mode,pList->lpc.datum);
      else {
        printf("@service lpc, %s_%s_%s: Mode %s not in program list.\n", 
               pList->lpc.family,pList->lpc.member,pList->lpc.action,mode);  
        return (-3); 
      }
      if( res <0) printf("@rpc error #%i for lpc %s_%s_%s_%s\n",res,pList->lpc.family,pList->lpc.member,pList->lpc.action,mode);
    break ;
    
    case service_cyc: 
      if(!strcmp("wtipol",mode))  res = cyc_wtipol(pList->cyc.family,pList->cyc.member,pList->cyc.datum);
      else {
        printf("@service cyc, %s_%s: Mode %s not in program list.\n", 
               pList->cyc.family,pList->cyc.member,mode);  
        return (-3); 
      }
      if( res <0) printf("@rpc error #%i for cyc %s_%s_%s\n",res,pList->cyc.family,pList->cyc.member,mode);
    break ;

    case service_bpm: 
      if(!strcmp("x",mode))        res = bpm_x (pList->bpm.s_x,pList->bpm.errn);
      else if(!strcmp("y",mode))   res = bpm_y (pList->bpm.s_y,pList->bpm.errn);
      else if(!strcmp("all",mode)) res = bpm_all (pList->bpm.x,pList->bpm.xrms,pList->bpm.y,pList->bpm.yrms,pList->bpm.s,pList->bpm.errn);
      else if(!strcmp("wn",mode))  res = bpm_wn (pList->bpm.n_sample,pList->bpm.errn);
      else if(!strcmp("rn",mode))  res = bpm_rn (pList->bpm.n_sample);
      else if(!strcmp("time",mode))res = bpm_time (pList->bpm.time);
      else {
        printf("@service bpm: Mode %s not in program list.\n",mode);  
        return (-3); 
      }
      if( res <0) printf("@rpc error #%i for bpm_%s\n",res,mode);
    break ;
    
    case service_bump: 
      if(!strcmp("start",mode))      res = bump_start (pList->bump.family,pList->bump.member,pList->bump.curr,pList->bump.steps);
      else if(!strcmp("stop",mode))  res = bump_stop ();
      else if(!strcmp("go",mode))    res = bump_go ();
      else if(!strcmp("abort",mode)) res = bump_abort ();
      else if(!strcmp("step",mode))  res = bump_step ();
      else if(!strcmp("check",mode)) res = bump_check (pList->bump.busy,pList->bump.stop,pList->bump.step,pList->bump.message);
      else if(!strcmp("read",mode))  res = bump_read (pList->bump.family,pList->bump.member,pList->bump.curr,pList->bump.errn);
      else {
        printf("@service bump: Mode %s not in program list.\n",mode);  
        return (-3); 
      }
      if( res <0) printf("@rpc error #%i for bump_%s\n",res,mode);
    break ;
    
    case service_inj: 
      if(!strcmp("k_war",mode))      res = injk_war(pList->inj.family,pList->inj.member,pList->inj.action,"WAR",pList->inj.datum,pList->inj.errn,pList->inj.errt);
      else if(!strcmp("k_rar",mode)) res = injk_rar(pList->inj.family,pList->inj.member,pList->inj.action,"RAR",pList->inj.datum,pList->inj.errn,pList->inj.errt);
      else if(!strcmp("s_war",mode)) res = injs_war(pList->inj.family,pList->inj.member,pList->inj.action,"WAR",pList->inj.datum,pList->inj.errn,pList->inj.errt);
      else if(!strcmp("s_rar",mode)) res = injs_rar(pList->inj.family,pList->inj.member,pList->inj.action,"RAR",pList->inj.datum,pList->inj.errn,pList->inj.errt);
      else if(!strcmp("src_wt",mode))res = injsrc_wtipol(pList->inj.family,pList->inj.member,pList->inj.action,"WTIPOL",pList->inj.datum,pList->inj.errn,pList->inj.errt);
      else if(!strcmp("src_ra",mode))res = injsrc_rar(pList->inj.family,pList->inj.member,pList->inj.action,"RAR",pList->inj.datum,pList->inj.errn,pList->inj.errt);
      else if(!strcmp("tlc_wt",mode))res = injtlc_wtipol(pList->inj.family,pList->inj.member,pList->inj.action,"WTIPOL",pList->inj.datum,pList->inj.errn,pList->inj.errt);
      else if(!strcmp("tlc_ra",mode))res = injtlc_rar(pList->inj.family,pList->inj.member,pList->inj.action,"RAR",pList->inj.datum,pList->inj.errn,pList->inj.errt);
      else {
        printf("@service inj, %s_%s_%s: Mode %s not in program list.\n", 
               pList->inj.family,pList->inj.member,pList->inj.action,mode);  
        return (-3); 
      }
      if( res !=0) printf("@rpc error #%i for inj%s %s_%s_%s\n",res,mode,pList->inj.family,pList->inj.member,pList->inj.action);
    break ;
    
    case service_sare: 
      if(!strcmp("bqt_w",mode))      res = sare_bqt_w (pList->sare.curr,pList->sare.flag,pList->sare.errn,pList->sare.max_time);
      else if(!strcmp("ct_w",mode))  res = sare_ct_w  (pList->sare.curr,pList->sare.flag,pList->sare.errn,pList->sare.max_time);
      else if(!strcmp("bqss_w",mode))res = sare_bqss_w(pList->sare.curr,pList->sare.flag,pList->sare.errn,pList->sare.max_time);
      else if(!strcmp("cs_w",mode))  res = sare_cs_w  (pList->sare.curr,pList->sare.flag,pList->sare.errn,pList->sare.max_time);
      else if(!strcmp("bqt_r",mode)) res = sare_bqt_r (pList->sare.curr,pList->sare.errcurr,pList->sare.curset,pList->sare.errcurset,pList->sare.stat,pList->sare.errstat,pList->sare.flag);
      else if(!strcmp("ct_r",mode))  res = sare_ct_r  (pList->sare.curr,pList->sare.errcurr,pList->sare.curset,pList->sare.errcurset,pList->sare.stat,pList->sare.errstat,pList->sare.flag);
      else if(!strcmp("bqss_r",mode))res = sare_bqss_r(pList->sare.curr,pList->sare.errcurr,pList->sare.curset,pList->sare.errcurset,pList->sare.stat,pList->sare.errstat,pList->sare.flag);
      else if(!strcmp("cs_r",mode))  res = sare_cs_r  (pList->sare.curr,pList->sare.errcurr,pList->sare.curset,pList->sare.errcurset,pList->sare.stat,pList->sare.errstat,pList->sare.flag);
      else {
        printf("@service sare: Mode %s not in program list.\n",mode);  
        return (-3); 
      }
      if( res <0) printf("@rpc error #%i for sare_%s\n",res,mode);
    break ;
    
    case service_flsc: 
      if(!strcmp("flsc",mode))      res=flsc_get_image(pList->flsc.image);
      else {
        printf("@service flsc: Mode %s not in program list.\n",mode);  
        return (-3); 
      }
      if( res <0) printf("@rpc error #%i for flsc_%s\n",res,mode);
    break ;
    
    case service_info: 
	    if(!strcmp("info",mode))   res=   get_beam_info(pList->info.curr,pList->info.cflag,pList->info.life,pList->info.lflag,pList->info.injrate,pList->info.injflag,pList->info.intcurr,pList->info.intflag,pList->info.energy,pList->info.engflag);
      else {
        printf("@service info: Mode %s not in program list.\n",mode);  
        return (-3); 
      }
      if( res <0) printf("@rpc error #%i for info_%s\n",res,mode);
    break ;
    
    default:
      printf ("@Nonexistent service #%i - host:%s port#%i\n", 
             service, pHost->host, pHost->port ) ;
      return ( -5 ) ;
  } /* switch between services */

  /* print general RPC error variables if error occurred */
  if( res <0)
  {
    if(errstring) printf("%s\n",errstring); /* C.S. added check for null pointer */
    printf("error type: %d with error number: %d\n\n",errtype,errno);
  }
  
  return ( res ) ;
}


char *rpc_error_message ( str ) 
char str[] ;
{
  sprintf ( str, "Errtype/no. %d/%d, %s", errtype, errno, errstring ) ;
  return ( str ) ;
} 


/*==================================================================
  ************************* end of rpc.c *************************
==================================================================*/
