#include <math.h>
#include <stdlib.h>
#include "hlsvar.h"

#define hor	0
#define ver	1
 
/*****************************************************************************
*                                                                            *
*    update_twiss.c: by Ryutaro Nagaoka                                      *
*               ... translation of Update_Twiss_Functions.f  [26/11/1992]    *
*                                                                            *
*  <<HLS>>                          Last Modified:  November 1992            *
*                                                                            *
* changed to incorporated the data bush strucuture (M.P. 15.12.92)           *
*                                                                            *
* errors returned:      0          all is OK                                 *
*                      -1          no closed solution                        *
*                      -2          type of element unknown                   *
*                      -3          sub_type of bending magnet unknown        *
*                                                                            *
* in case of errors, one line is written to stdout and the error code        *
* returned.                                                                  *
*                                                                            *
* In this module also Get_Tmatrix is defined and can be used by programmers. *
* It returns the same errors as update_twiss_functions, except -1.           *
*                                                                            *
* A minor bug in Get_Tmatrix fixed. Rectangular magnets now treated          *
* properly, sector and edge magnets added.                                   *
* V 1.1						(M.P. 14.1.93)               *
*                                                                            *
* A bug in closed mode fixed. Transfer matrices now copied correctly         *
* from the data bush.                                                        *
* V 1.2						(M.P. 27.1.93)               *
*                                                                            *
* model_tune[] and ID matrix calculation added. ID treated as drift if       *
* status is open (i.e. closed = 0).                                          *
* V 1.3                                         (M.P. 11.2.93)               *
*                                                                            *
* An error fixed in the treatment of ID as a drift: the length is now        *
* calculated correctly.                                                      *
* V 1.4                                         (M.P.  9.3.94)               *
******************************************************************************/

/*************************************************************************** 
 *  first two utility routines (are declared static to avoid conflicts)    * 
 ***************************************************************************/
 
static c_Twiss_Transform(twa_0, twa, am)
double twa_0[3], twa[3], am[3][3];
{

  twa[0] =          am[0][0]*am[0][0] *twa_0[0]
         -       2.*am[0][0]*am[0][1] *twa_0[1]
         +          am[0][1]*am[0][1] *twa_0[2];
         
  twa[1] =         -am[1][0]*am[0][0] *twa_0[0]
         + (1. + 2.*am[0][1]*am[1][0])*twa_0[1]
         -          am[0][1]*am[1][1] *twa_0[2];
         
  twa[2] =          am[1][0]*am[1][0] *twa_0[0]
         -       2.*am[1][1]*am[1][0] *twa_0[1]
         +          am[1][1]*am[1][1] *twa_0[2];
}
 

static c_Matrix_Multiplication(amh,amv,bmh,bmv,cmh,cmv)
double amh[3][3],amv[3][3],bmh[3][3],bmv[3][3],cmh[3][3],cmv[3][3];
{  
  int i, j, k;
  double sumH, sumV;

   for(i=0; i<3; i++) for(j=0; j<3; j++) {
     sumH = 0.0;
     sumV = 0.0;
     for(k=0; k<3; k++) {
        sumH = sumH + bmh[i][k]*amh[k][j];
        sumV = sumV + bmv[i][k]*amv[k][j];
     }
     cmh[i][j] = sumH;   
     cmv[i][j] = sumV;  
  }
}


/*************************************************************************** 
 *  Now the update_twiss_functions and Get_Tmatrix are defined             *
 ***************************************************************************/
 

int update_twiss_functions(system, twiss)

int 		system;
twiss_record 	*twiss;

{
   int i, j, ib;
   int err;
   int debug = 0;   
   double pi, twopi;
   double beta_I[2], alfa_I[2], gamma_I[2], phase_I[2], disp_I[2], dspp_I[2];
   double ah0[3][3], av0[3][3], amh[3][3], amv[3][3], cmh[3][3], cmv[3][3];       
   double denom[2], denomD[2], sumF[2], argF[2];
   double twaH_0[3], twaV_0[3], twaH[3], twaV[3],
          dphas[2], phas[2], disp_0[2],  dspp_0[2], disp_N[2],  dspp_N[2]; 
               
/* ------------------------------------------------------------------------- */

   pi = 4.0*atan(1.0);  twopi = 2.0*pi;
   
         
   if(system == TL ) for (i=hor;i<=ver;i++) {
      beta_I[i] = twiss->beta[i];  
      alfa_I[i] = twiss->alpha[i];   
      disp_I[i] = twiss->D[i];   
      dspp_I[i] = twiss->Dprime[i];   
   }


/*---- Derivation of Initial Optics Values for the Closed System ----------- */
   
   if(system == SR ) {  
   
/*   Initialization of Transfer Matrices:                                    */
/*   ------------------------------------                                    */
     for(i=0; i<3; i++) {
        for(j=0; j<3; j++) {
           ah0[i][j] = 0.0;
           av0[i][j] = 0.0;
           if(j == i) {
              ah0[i][j] = 1.0;
              av0[i][j] = 1.0;
           }
        }
     }         
           
/*   Iteration Loop to Obtain Transfer Matrices of the Ring:                 */
/*   -------------------------------------------------------                 */
     for(ib=0; ib<N_general; ib++) 
       if ( ge[ib]->type < thin_type ) {
     
        if ( (err = get_transfer_matrix(ge[ib], amh, amv)) != 0 ) return(err) ;
        c_Matrix_Multiplication(ah0,av0, amh,amv, cmh,cmv);

        for(i=0; i<3; i++) {
           for(j=0; j<3; j++) {
              ah0[i][j] = cmh[i][j];
              av0[i][j] = cmv[i][j];
           }
        }
     }            
           
/*   Optics Functions at the Initial Position:                               */
/*   -----------------------------------------                               */
     sumF[hor]   = ah0[0][0] + ah0[1][1];
     sumF[ver]   = av0[0][0] + av0[1][1];
     denom[hor]  = 1.0 - sumF[hor]*sumF[hor]/4.0;
     denom[ver]  = 1.0 - sumF[ver]*sumF[ver]/4.0;
     denomD[hor] = 2.0 - sumF[hor];
     denomD[ver] = 2.0 - sumF[ver];
     
     if(denom[hor] < 0.0 || denom[ver] < 0.0 ) {
        printf("\n\n  +++ NO OPTICAL SOLUTION +++ \n");
        return(-1);
     }
     
     denom[hor] = sqrt(denom[hor]);   
     denom[ver] = sqrt(denom[ver]);   
     
     beta_I[hor] = fabs( ah0[0][1] )/denom[hor];
     beta_I[ver] = fabs( av0[0][1] )/denom[ver];
     alfa_I[hor] = (ah0[0][0] - ah0[1][1])*ah0[0][1]
                  /(2.*fabs( ah0[0][1] )*denom[hor]);
     alfa_I[ver] = (av0[0][0] - av0[1][1])*av0[0][1]
                  /(2.*fabs( av0[0][1] )*denom[ver]);
     disp_I[hor] = ((1.0 - ah0[1][1])*ah0[0][2] + ah0[0][1]*ah0[1][2])
                  /denomD[hor];             
     disp_I[ver] = ((1.0 - av0[1][1])*av0[0][2] + av0[0][1]*av0[1][2])
                  /denomD[ver];             
     dspp_I[hor] = ((1.0 - ah0[0][0])*ah0[1][2] + ah0[1][0]*ah0[0][2])
                  /denomD[hor];             
     dspp_I[ver] = ((1.0 - av0[0][0])*av0[1][2] + av0[1][0]*av0[0][2])
                  /denomD[ver];     
                          
   for (i=hor;i<=ver;i++) {
      twiss->beta[i] = beta_I[i] ;  
      twiss->alpha[i] = alfa_I[i] ;   
      twiss->D[i] = disp_I[i] ;   
      twiss->Dprime[i] = dspp_I[i] ;   
   }

   }  
      
   gamma_I[hor] = (1.0 + alfa_I[hor]*alfa_I[hor])/beta_I[hor];
   gamma_I[ver] = (1.0 + alfa_I[ver]*alfa_I[ver])/beta_I[ver];
   phase_I[hor] = 0.0;
   phase_I[ver] = 0.0;
   
   if(debug) {
      printf("\n Initial Optics Values:");
      for(i=0; i<2; i++) printf("\n   %10.6f  %10.6f  %10.6f  %10.6f",
      beta_I[i], alfa_I[i], disp_I[i], dspp_I[i]);  printf("\n");      
   }   

/* ------------------------------------------------------------------------- */
           
/*   Iteration Loop to Obtain Optics at each Individual Elements:            */
/*   ------------------------------------------------------------            */

   twaH_0[0] =  beta_I[hor];
   twaH_0[1] =  alfa_I[hor];
   twaH_0[2] = gamma_I[hor];
   twaV_0[0] =  beta_I[ver];
   twaV_0[1] =  alfa_I[ver];
   twaV_0[2] = gamma_I[ver];
   
   phas[hor] = phase_I[hor];
   phas[ver] = phase_I[ver];
   
   disp_0[hor] = disp_I[hor];
   dspp_0[hor] = dspp_I[hor];
   disp_0[ver] = disp_I[ver];
   dspp_0[ver] = dspp_I[ver];
   
   for(ib=0; ib<N_general; ib++) 
     if ( ge[ib]->type < thin_type ) {
      
      ge[ib]->twiss-> beta[hor]  = twaH_0[0];
      ge[ib]->twiss-> beta[ver]  = twaV_0[0];
      ge[ib]->twiss->alpha[hor]  = twaH_0[1];
      ge[ib]->twiss->alpha[ver]  = twaV_0[1];
/* skip gamma, unless new request
      ge[ib]->twiss->gamma[hor]  = twaH_0[2];
      ge[ib]->twiss->gamma[ver]  = twaV_0[2];
end of skip */
      ge[ib]->twiss->phase[hor]  = phas[hor]/twopi;
      ge[ib]->twiss->phase[ver]  = phas[ver]/twopi;
      ge[ib]->twiss->    D[hor]  = disp_0[hor];
      ge[ib]->twiss->    D[ver]  = disp_0[ver];
      ge[ib]->twiss->Dprime[hor]  = dspp_0[hor];
      ge[ib]->twiss->Dprime[ver]  = dspp_0[ver];

      if(system == TL ) {
        if ( (err = get_transfer_matrix(ge[ib], amh, amv)) != 0 ) return(err) ;
      } else /* copy already created transfer matrices from global structure */
      {
        for(i=0; i<3; i++) {
           for(j=0; j<3; j++) {
              amh[i][j] = ge[ib]->transfer->m[horizontal][i][j] ;
              amv[i][j] = ge[ib]->transfer->m[vertical][i][j] ;
           }
        }
      } /* else */
      
      c_Twiss_Transform(twaH_0, twaH, amh);
      c_Twiss_Transform(twaV_0, twaV, amv);
      
      argF[hor]  = amh[0][1]/(amh[0][0]*twaH_0[0] - amh[0][1]*twaH_0[1]);
      argF[ver]  = amv[0][1]/(amv[0][0]*twaV_0[0] - amv[0][1]*twaV_0[1]);
      dphas[hor] = atan( argF[hor] );
      dphas[ver] = atan( argF[ver] );
      if(dphas[hor] < 0.0)  dphas[hor] = dphas[hor] + pi;
      if(dphas[ver] < 0.0)  dphas[ver] = dphas[ver] + pi;
      phas[hor] = phas[hor] + dphas[hor];
      phas[ver] = phas[ver] + dphas[ver];

      disp_N[hor] = amh[0][0]*disp_0[hor] + amh[0][1]*dspp_0[hor] + amh[0][2];
      dspp_N[hor] = amh[1][0]*disp_0[hor] + amh[1][1]*dspp_0[hor] + amh[1][2];
      disp_N[ver] = amv[0][0]*disp_0[ver] + amv[0][1]*dspp_0[ver] + amv[0][2];
      dspp_N[ver] = amv[1][0]*disp_0[ver] + amv[1][1]*dspp_0[ver] + amv[1][2];
            
      for(i=0; i<3; i++) {
         twaH_0[i] = twaH[i];
         twaV_0[i] = twaV[i];
         if(i <= 1) {
            disp_0[i] = disp_N[i];
            dspp_0[i] = dspp_N[i];
         }
      }
   }         
   
   model_tune[hor] = phas[hor] / twopi ;
   model_tune[ver] = phas[ver] / twopi ;
   
   return ( 0 ) ;
 
/* ------------------------------------------------------------------------- */

}


/***************************************************************************
*                                                                          *
*  Get_Tmatrix.c: by Ryutaro Nagaoka                                       * 
*               ... translation of Get_Tmatrix.f     [25/11/1992]          *
*                                                                          *  
*  <<HLS>>                          Last Modified:  November 1992          *
*                                                                          *       
* septum type added as rectangular magnet.                                 *       
* modified to include global data bush           (M.P. 15.12.1992)         *       
*                                                                          *       
****************************************************************************/

int get_transfer_matrix(ptr, amh, amv)

general_record	*ptr;
double amh[3][3], amv[3][3];

{    
  int i, j, sub, kkx, kky, kkxh, kkyh;
  double pi;
  double bmh[3][3], bmv[3][3], cmh[3][3], cmv[3][3];
  double sL, fK, fK0, rfK, rfK0, arg, arg0, rho, theta,
         cs, sn, tn, ch, sh, epp, epn,
         fKEL;
  double kx, ky, kxh, kyh, kayx, kayy, rhoh, k, kay, length;  /* ID variables */
  
/* ------------------------------------------------------------------------- */

   pi = 4.0*atan(1.0); 

/*  Initialization of Transfer Matrices:                                     */
/*  ------------------------------------                                     */
   for(i=0; i<3; i++) {
      for(j=0; j<3; j++) {
         amh[i][j] = 0.0; amv[i][j] = 0.0; 
         bmh[i][j] = 0.0; bmv[i][j] = 0.0; 
         if(j == i) {
            amh[i][j] = 1.0; amv[i][j] = 1.0;
            bmh[i][j] = 1.0; bmv[i][j] = 1.0;
         }
      }
   }
   
   if(ptr->type > thin_type ) return;         
    
/* ------------------------------------------------------------------------- */

   switch ( ptr->type ) {

/*  Element type = Drift:                                                    */
/*  ---------------------                                                    */
   case drift_type:       
      sL = ptr->p.dr -> length;
      amh[0][1] = sL;
      amv[0][1] = sL;
   break ;   

/*  Element type = Quadrupole:                                               */
/*  --------------------------                                               */
   case quad_type:       
      sL = ptr->p.qu -> magnetic_length;
      fK = ptr->p.qu -> strength;
      
      if(fK == 0.0 ) {
         amh[0][1] = sL;
         amv[0][1] = sL;
         break ;   
      }
      
      rfK = sqrt(fabs(fK));
      arg = rfK*sL;
      cs  = cos(arg);  sn  = sin(arg);
      epp = exp(arg);  epn = 1.0/epp;
      ch  = (epp + epn)/2.0;
      sh  = (epp - epn)/2.0;
      
      if(fK < 0.0) {      
         amh[0][0] =  cs;       amh[0][1] = sn/rfK;
         amh[1][0] = -sn*rfK;   amh[1][1] = cs;
         
         amv[0][0] =  ch;       amv[0][1] = sh/rfK;
         amv[1][0] =  sh*rfK;   amv[1][1] = ch;
      }   
      else {           
         amh[0][0] =  ch;       amh[0][1] = sh/rfK;
         amh[1][0] =  sh*rfK;   amh[1][1] = ch;
         
         amv[0][0] =  cs;       amv[0][1] = sn/rfK;
         amv[1][0] = -sn*rfK;   amv[1][1] = cs;
      }      
   break ;   

/*  Bend type = Bending Magnet:                                             */
/*  ------------------------------------------------------                  */
   case bend_type:       
      sL    = ptr->p.be -> magnetic_length;
      theta = ptr->p.be->pPar -> ps.angle*pi/180.0;
      
      if(theta == 0.0) {
         amh[0][1] = sL;
         amv[0][1] = sL;
         break;
      }   
      /* sub_type larger than reversed_polarity means opposite bending !! */
      if ( ptr->p.be->sub_type >= reversed_polarity ) theta = - theta ;
      
      rho   = sL/theta;
      
      /* subtype specific calculations follow: */
      switch ( sub = ptr->p.be->sub_type % reversed_polarity ) {
      
/*  Bend sub_type = Combined Function Horizontal Dipole:                    */
/*  -----------------------------------------------------------             */
      case com_rectangle:
      case com_sector:

        fK0   = ptr->p.be->strength / (rho*rho);
        fK    = fK0 - 1./(rho*rho);
        rfK0  = sqrt(fabs(fK0));
        rfK   = sqrt(fabs(fK ));
        arg0  = rfK0*sL;
        arg   = rfK *sL;
        
        if(fK0 == 0.0 ) {   
           amv[0][1] = sL;        
           cs = cos(arg);  sn = sin(arg);
           amh[0][0] = cs;      amh[0][1] = sn/rfK;  amh[0][2] = (1. - cs)*rho;
           amh[1][0] =-sn*rfK;  amh[1][1] = cs;      amh[1][2] = sn ;
        }   
        else   {
           cs = cos(arg0);  sn = sin(arg0);
           amv[0][0] =  cs;        amv[0][1] = sn/rfK0;
           amv[1][0] = -sn*rfK0;   amv[1][1] = cs;
           if(fK == 0.0 ) {
              amh[0][1] = sL;
              amh[0][2] = sL*theta/2.0;
              amh[1][2] = theta;
           }   
           else if(fK < 0.0) {
              cs = cos(arg);  sn = sin(arg);
              amh[0][0] = cs;      amh[0][1] = sn/rfK;  amh[0][2] = -(1. - cs)/(rho*fK);
              amh[1][0] =-sn*rfK;  amh[1][1] = cs;      amh[1][2] = sn/(rho*rfK);
           }   
           else  {
              epp = exp(arg);  epn = 1.0/epp;
              ch  = (epp + epn)/2.0;
              sh  = (epp - epn)/2.0;
              amh[0][0] = ch;      amh[0][1] = sh/rfK;  amh[0][2] = -(1. - ch)/(rho*fK);
              amh[1][0] = sh*rfK;  amh[1][1] = ch;      amh[1][2] = sh/(rho*rfK);
           }
        }
        
        if ( sub == com_sector ) break ;

  /*    Edge Matrices for rectagular magnet:                                 */ 
  /*    -------------------------------------                                */
        fKEL = tan(theta/2.0)/rho;
        bmh[1][0] = fKEL;     
        bmv[1][0] =-fKEL;  
        
        c_Matrix_Multiplication(bmh,bmv,amh,amv,cmh,cmv);      
        c_Matrix_Multiplication(cmh,cmv,bmh,bmv,amh,amv);    
        
      break;

/*  Bend sub_type =  Horizontal Dipole:                                     */
/*  -------------------------------------------------                       */
      case hor_rectangle:

        cs = cos(theta);  sn = sin(theta);  tn = tan(theta/2.0);
        
        amh[0][1] = rho*sn;
        amh[0][2] = rho*(1.0 - cs);
        amh[1][2] = 2.0*tn;
        amv[0][0] = 1.0 - theta*tn;
        amv[0][1] = sL;
        amv[1][0] =-tn/rho*(2.0 - theta*tn);
        amv[1][1] = 1.0 - theta*tn;
                
      break;

      case hor_sector:

        cs = cos(theta);  sn = sin(theta);
        
        amh[0][0] = cs;
        amh[0][1] = rho*sn;
        amh[0][2] = rho*(1.0 - cs);
        amh[1][0] = -sn/rho;
        amh[1][1] = cs;
        amh[1][2] = sn;
        amv[0][1] = sL;
                
      break;

/*  Bend sub_type = Vertical Dipole:                                        */
/*  ------------------------------------------------                        */
      case ver_rectangle:

        cs = cos(theta);  sn = sin(theta);  tn = tan(theta/2.0);
        
        amh[0][0] = 1.0 - theta*tn;
        amh[0][1] = sL;
        amh[1][0] =-tn/rho*(2.0 - theta*tn);
        amh[1][1] = 1.0 - theta*tn;
        amv[0][1] = rho*sn;
        amv[0][2] = rho*(1.0 - cs);
        amv[1][2] = 2.0*tn;
        
      break ;  
        
      case ver_sector:

        cs = cos(theta);  sn = sin(theta);
        
        amh[0][1] = sL;
        amv[0][0] = cs;
        amv[0][1] = rho*sn;
        amv[0][2] = rho*(1.0 - cs);
        amv[1][0] = -sn/rho;
        amv[1][1] = cs;
        amv[1][2] = sn;
                
      break;

/*  Bend sub_type = Edge Focussing:                                         */
/*  ------------------------------------------------                        */
      case hor_edge:

        fKEL = tan(theta/2.0)/rho;
        amh[1][0] = fKEL;     
        amv[1][0] =-fKEL; 
        
      break;
        
      case ver_edge:

        fKEL = tan(theta/2.0)/rho;
        amh[1][0] =-fKEL;     
        amv[1][0] = fKEL; 
        
      break;
        
/*  Undefined subtypes:                                                     */
/*  ------------------------------------------------                        */
      default:
        printf ( "sub_type #%i not supported at general el#%i\n",
        		ptr->p.be->sub_type, ptr->p.be->back_index ) ;
        return ( -3 ) ;
        
      }  /* switch ( ptr->p.be->sub_type ) */

    break ; /* type == bend_type */
    
/*  Element type =  Septum Magnet:                                          */
/*  -------------------------------------------                             */
    case septum_type:
          
      sL    = ptr->p.sp->pAcc->pPar->in.length;
      
      if(ptr->p.sp->curvature == 0.0) {
         amh[0][1] = sL;
         amv[0][1] = sL;
         break;
      }   
      
      rho   = 1.0 / ptr->p.sp->curvature; 
      theta = sL / rho ;
      cs = cos(theta);  sn = sin(theta);  tn = tan(theta/2.0);
      
      amh[0][1] = rho*sn;
      amh[0][2] = rho*(1.0 - cs);
      amh[1][2] = 2.0*tn;
      amv[0][0] = 1.0 - theta*tn;
      amv[0][1] = sL;
      amv[1][0] =-tn/rho*(2.0 - theta*tn);
      amv[1][1] = 1.0 - theta*tn;
      
    break;

/*  Element type =  Insertion Device:         Added by Lidia Tosi           */
/*  -------------------------------------------                             */
    case ID_type:

      /* calculate length, as it is used in any case */
      length = ptr->p.id->pPar->id.period * ptr->p.id->pPar->id.Np ;
            

      if ( ptr->p.id->closed == 1 )
      {
        k = 2*pi/ptr->p.id->pPar->id.period ;
        kx = ptr->p.id->k_x;
        ky = ptr->p.id->k_y;
        rho = ptr->p.id->curvature;
        
        kxh = ptr->p.id->kh_x;
        kyh = ptr->p.id->kh_y;
        rhoh = ptr->p.id->curvature_h;
        
        if ( (kx <= 0.0) && (ky <= 0.0) ) {  kkx=-1; kky=-1; }
        if ( (kx <= 0.0) && (ky > 0.0) ) {  kkx=-1; kky=1; }
        if (( kx > 0.0 ) && (ky <= 0.0) ) {  kkx=1; kky=-1;}
        if (( kx > 0.0 ) && (ky > 0.0) ) {  kkx=1; kky=1;}

        if ( (kyh <= 0.0) && (kxh <= 0.0) ) { kkyh=-1; kkxh=-1; }
        if ( (kyh <= 0.0) && (kxh > 0.0) ) {  kkyh=-1; kkxh=1; }
        if ( (kyh > 0.0 ) && (kxh <= 0.0) ) { kkyh=1; kkxh=-1;}
        if ( (kyh > 0.0 ) && (kxh > 0.0) ) {  kkyh=1; kkxh=1;}
        
        kayx = kkx*kx*kx*rho*rho + kkxh*kxh*kxh*rhoh*rhoh;
        if ( kayx <=0.0 ) kayx=-sqrt(fabs(kayx));
        if ( kayx > 0.0 ) kayx=sqrt(fabs(kayx));
        
        kayy = kky*ky*ky*rho*rho + kkyh*kyh*kyh*rhoh*rhoh;
        if ( kayy <=0.0 ) kayy=-sqrt(fabs(kayy));
        if ( kayy > 0.0 ) kayy=sqrt(fabs(kayy));
         
         
      }
      else { kayx = 0.0; kayy = 0.0 ; } /* open ID will be treated as drift */
      
      kay = fabs( kayx/(sqrt(2.0)*k) );
      if ( kayx == 0.0 )
      {
        amh[0][0] = 1.0;
        amh[0][1] = length;        
        amh[1][0] = 0.0;
        amh[1][1] = 1.0;
      }
      if ( kayx < 0.0 )
      {
        
        amh[0][0] = cos( kay * length );
        amh[0][1] = sin( kay * length);
        amh[1][0] = amh[0][1];
        amh[0][1] = amh[0][1]/kay;
        amh[1][0] = -amh[1][0]*kay;   
        amh[1][1] = amh[0][0];
      }
      if (kayx > 0.0 )
      {
        amh[0][0] = cosh( kay * length );
        amh[0][1] = sinh( kay * length);
        amh[1][0] = amh[0][1];
        amh[0][1] = amh[0][1]/kay;
        amh[1][0] = amh[1][0]*kay;   
        amh[1][1] = amh[0][0];
      }
      
      kay = fabs(kayy)/(sqrt(2.0)*k);
      if ( kayy == 0.0 )
      {
        amv[0][0] = 1.0;
        amv[0][1] = length;        
        amv[1][0] = 0.0;
        amv[1][1] = 1.0;
      }
      if ( kayy < 0.0 )
      {
        amv[0][0] = cos( kay * length );
        amv[0][1] = sin( kay * length );
        amv[1][0] = amv[0][1];
        amv[0][1] = amv[0][1]/kay;
        amv[1][0] = -amv[1][0]*kay;   
        amv[1][1] = amv[0][0];
      }
      if (kayy > 0.0 )
      {
        amv[0][0] = cosh( kay * length );
        amv[0][1] = sinh( kay * length);
        amv[1][0] = amv[0][1];
        amv[0][1] = amv[0][1]/kay;
        amv[1][0] = amv[1][0]*kay;   
        amv[1][1] = amv[0][0];
      }
    
    break;

    default:
      /* a little dirty: type is unknown, still the element is considered 
         a bend, to get the back_index */
      printf ( "type #%i not supported at general el#%i\n",
                      ptr->type, ptr->p.be->back_index ) ;
      return ( -2 ) ;
      
  }  /* switch ( ptr->type ) */   
   
  /* copy calculated transfer matrix into data bush: */
  for(i=0; i<3; i++) {
    for(j=0; j<3; j++) {
      ptr->transfer->m[horizontal][i][j] = amh[i][j] ;
      ptr->transfer->m[vertical][i][j] = amv[i][j] ;
    }
  }
  return ( 0 ) ;
/* ------------------------------------------------------------------------- */

}
